;#pragma usePoFile 0
; Scripts for TeamTalk Classic
;
; Author:  Doug Lee
/*
Copyright (c) 2010-2018 Doug Lee.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* The names of the copyright holders and contributors may not be used to
  endorse or promote products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/
include "hjconst.jsh"
include "MSAAConst.jsh"
include "TeamTalk5Classic.jsh"
include "TeamTalk5Classic.jsm"

use "tt_jcpdict.jsb"

const
; Script revision number, spoken by JAWSKey+Q.
	Script_Revision = "228"

script scriptFileName()
scriptAndAppNames(formatString(msgScriptName, Script_Revision))
endScript

void function autoStartEvent()
settings__default("chatMessageTimes", True)
settings__default("AnalyzeTreeIcons", True)
chat4__registerAllChatWindows()
toolbar__watchForChanges(10)
endFunction

script refreshScreen()
roomtree__init(2)
performScript refreshScreen()
endScript

script focusChatEdit()
; Move focus to the input edit box for the current channel or private message window.
_focusChatPart(1)
endScript

script focusChatHistory()
; Move focus to the message history area for the current channel or private message window.
_focusChatPart(0)
endScript

void function _focusChatPart(int wantInput)
; Focus one of the two edit windows in a channel or private chat window.
; Announces the new focus on success and beeps on failure.
var handle hwnd = findWindow(getTopLevelWindow(getFocus()), "RichEdit20W", "")
if !hwnd
	hwnd = findWindow(getTopLevelWindow(getFocus()), "RichEdit20A", "")
endIf
if !hwnd
	beep()
	return
endIf
var int cnt = wantInput +1
while cnt && hwnd
	if stringContains(stringLower(getWindowClass(hwnd)), "edit")
		cnt = cnt -1
	endIf
	if cnt
		hwnd = getNextWindow(hwnd)
	endIf
endWhile
if !hwnd
	beep()
	return
endIf
if hwnd == getFocus()
	sayObjectTypeAndText()
else
	setFocus(hwnd)
endIf
endFunction

script focusChannelTree()
; Move focus to the channel tree in the current window.
var handle hwnd = findWindow(getTopLevelWindow(getFocus()), "SysTreeView32", "")
if !hwnd
	beep()
	return
endIf
if hwnd == getFocus()
	sayObjectTypeAndText()
else
	setFocus(hwnd)
endIf
endScript

int function ReadMsg(string which, int isReviewSystem, optional int alwaysIncludeTime)
; Which should be +n (n forward), -n (n backward), n (message #n), "L" (last), or "C" (current).
var string buf
var collection rec = chat4__record()
var int msgno = rec.reviewMessageNumber
var int messageCount = chat4__messageCount()
var string ch
; Adjustment for reversed windows.
if rec.reversed
	ch = stringLeft(which, 1)
	var string rest = stringChopLeft(which, 1)
	if ch == "+"
		which = formatString("-%1", rest)
	elif ch == "-"
		which = formatString("+%1", rest)
	elif ch == "l"
		which = "1"
	elif ch == "c"
		; No change.
	else  ; 0-9, absolute position.
		which = intToString( messageCount -stringToInt(which) +1)
	endIf
endIf

; First deal with room switches.
if messageCount < 0
	sayMessage(OT_Error, msgMessageError)
	return False
endIf
if !messageCount
	msgno = 0
	sayMessage(OT_User_Requested_Information, msgNoMessages)
	if isReviewSystem
		rec.reviewMessageNumber = msgno
	endIf
	return False
elif msgno < 1
	msgno = 1
elif msgno > messageCount
	msgno = messageCount
endIf

ch = stringLeft(which, 1)
if ch == "L"
	msgno = messageCount
elif ch == "C"
	; Do nothing; leave msgno as is.
elif stringContains("+-", stringLeft(which, 1))
	msgno = msgno +stringToInt(which)
else
	msgno = stringToInt(which)
endIf

; Get the message after possible timestamp info removal.
buf = chat4__messageText(msgno)
if !alwaysIncludeTime
	applyChatTimePreference(buf)
endIf

if ch == "c"
	var string buf1
	; For reversed windows we have to flip the message number.
	var int mn1 = rec.reviewMessageNumber
	if rec.reversed
		mn1 = messageCount -mn1 +1
	endIf
	buf1 = formatString(msgMessagePos, intToString(mn1), intToString(messageCount))
	buf = formatString("%1: %2", buf1, buf)
endIf
sayMessage(OT_User_Requested_Information, buf)

if isReviewSystem
	if msgno < 1
		msgno = 1
	elif msgno > messageCount
		msgno = messageCount
	endIf
	rec.reviewMessageNumber = msgno
endIf
return True
endFunction

int function applyChatTimePreference(string byRef buf)
; Strip out the message time if requested by the user option for this.
if opt__get("chatMessageTimes")
	return False
endIf
var int pos = stringContains(buf, " ")
var string maybeTime = stringLeft(buf, pos-1)
if stringContains(maybeTime, ":")
&& stringContains("0123456789", stringLeft(maybeTime, 1))
&& stringContains("0123456789", stringRight(maybeTime, 1))
	buf = stringChopLeft(buf, pos)
	return True
endIf
return False
endFunction

Script ReadFirstMessage()
ReadMsg("1", True)
EndScript
Script ReadLastMessage()
ReadMsg("L", True)
EndScript
Script ReadNextMessage()
ReadMsg("+1", True)
EndScript
Script ReadPriorMessage()
ReadMsg("-1", True)
EndScript

Script ReadCurrentMessage()
if !isSameScript()
	ReadMsg("c", True, True)
	return
endIf
var collection rec = chat4__record()
var string buf = chat4__messageText(rec.reviewMessageNumber)
applyChatTimePreference(buf)
spellString(buf)
EndScript

script altEnterKey()
if util__isSpecialFocus()
	performScript altEnterKey()
	return
endIf
performScript readCurrentMessage()
endScript

script readAbsoluteMessage(int idx)
if idx < 0
	idx = chat4__messageCount() +idx +1
endIf
readMsg(intToString(idx), False)
endScript

script topOfFile()
if !util__isSpecialFocus() && isPCCursor() && getFocus() && getFocus() == roomtree__currentTreeWindow()
	roomtree__move("first", False)
	return
endIf
performScript topOfFile()
endScript

script selectFromTop()
if !util__isSpecialFocus() && isPCCursor() && getFocus() && getFocus() == roomtree__currentTreeWindow()
	roomtree__move("first", True)
	return
endIf
performScript selectFromTop()
endScript

script bottomOfFile()
if !util__isSpecialFocus() && isPCCursor() && getFocus() && getFocus() == roomtree__currentTreeWindow()
	roomtree__move("last", False)
	return
endIf
performScript bottomOfFile()
endScript

script selectToBottom()
if !util__isSpecialFocus() && isPCCursor() && getFocus() && getFocus() == roomtree__currentTreeWindow()
	roomtree__move("last", True)
	return
endIf
performScript selectToBottom()
endScript

script ControlUpArrow()
if !util__isSpecialFocus() && isPCCursor() && getFocus() && getFocus() == roomtree__currentTreeWindow()
	roomtree__move("prev", False)
	return
endIf
performScript ControlUpArrow()
endScript

script prevByAge()
if !util__isSpecialFocus() && isPCCursor() && getFocus() && getFocus() == roomtree__currentTreeWindow()
	roomtree__move("prev", True)
	return
endIf
sayCurrentScriptKeyLabel()
typeCurrentScriptKey()
endScript

script controlDownArrow()
if !util__isSpecialFocus() && isPCCursor() && getFocus() && getFocus() == roomtree__currentTreeWindow()
	roomtree__move("next", False)
	return
endIf
performScript controlDownArrow()
endScript

script nextByAge()
if !util__isSpecialFocus() && isPCCursor() && getFocus() && getFocus() == roomtree__currentTreeWindow()
	roomtree__move("next", True)
	return
endIf
sayCurrentScriptKeyLabel()
typeCurrentScriptKey()
endScript

script nextChat()
; Move to the next TeamTalk window containing a visible chat.
if !chat4__focusNext(1)
	beep()
endIf
endScript

script prevChat()
; Move to the previous TeamTalk window containing a visible chat.
if !chat4__focusNext(-1)
	beep()
endIf
endScript

script sayStatuses()
; Speak the statuses of toolbar icons.
; On one press, provide an abbreviated indication.
; On a double press, list all statuses completely.
toolbar__sayStatuses(isSameScript())
endScript

script DoEnter()
; Makes Enter join a room if focus is on a room line or on someone within one.
if util__isSpecialFocus()
	typeCurrentScriptKey()
	sayCurrentScriptKeyLabel()
	return
endIf
var handle hwnd = getFocus()
if getWindowClass(hwnd) == "SysTreeview32"
	if _DoubleClick(hwnd)
		return
	endIf
	beep()
endIf
typeCurrentScriptKey()
sayCurrentScriptKeyLabel()
endScript

int function _DoubleClick(handle hwnd)
; Double-click the focused tree item.
; For a room node, this joins the room if "Double click enters a room" is set in TeamTalk.
; For a member, this opens a chat window.
; Returns True on success and False on error.
var object o = MSAA__objectFromWindow(hwnd)
if !o
	return False
endIf
var int childID = o.accFocus +0
if !childID
	; Even if this means focus is just on the tree itself,
	; that's not a room to join or a member to talk to.
	return False
endIf
; Even though we're using winclick to avoid having to mess with the mouse pointer,
; we probably have to have the node actually on screen.
; TODO: Could improve this to allow for partial visibility.
var int accState = o.accState(childID)
if (accState & (State_System_Invisible | State_System_OffScreen))
	return False
endIf
; Click the center of the node.
var int left, int right, int top, int bottom
if !MSAA__rect(o, childID, left, right, top, bottom)
	return False
endIf
var int x = (left +right) /2
var int y = (top +bottom) /2
winclick__post(hwnd, "LeftDblClick", x, y)
; Send seems to return 0 even on success.
return True
endFunction

int function SayCellUnit(int UnitMovement)
; Ctrl+Alt+arrows change volumes in this application, so we pass them through.
typeCurrentScriptKey()
return True
endFunction

script sayBottomLineOfWindow()
; Make JAWSKey+PgDn work even when the window is not visible.
if !util__isSpecialFocus()
	var string txt = ttutil__statusText(getFocus())
	if txt
		sayMessage(OT_Status, txt)
		return
	endIf
endIf
performScript sayBottomLineOfWindow()
endScript

void function jcp__setCustomProps(variant byRef props, string sOrigin)
; Set any custom properties required per situation.
; props is empty on entry, can be modified, and must be taken by reference.
if isVirtualPCCursor()
|| util__isSpecialFocus()
	return
endIf
var collection loc = jcp__getLocation(sOrigin, "w")
if ttutil__isMainWindow(getTopLevelWindow(loc.hwnd))
	_labelMain(props, loc.hwnd)
endIf
endFunction

int function _labelMain(variant byRef props, handle hwnd)
; Label the sliders and other items in the main window.
var string wclass = getWindowClass(hwnd)
var string name
if wclass == "Edit"
	jcp__set(props, "Name", msgChannelMessageInput)
	return True
elif wclass == "RichEdit20W"
	jcp__set(props, "Name", msgChannelMessageHistory)
	return True
elif wclass == "SysTreeView32"
	name = getObjectValue()
	; That includes flags because of a getObjectValue() override in this file.
	if !name
		name = " "
	endIf
	jcp__set(props, "Value", name)
	return True
endIf
if getWindowTypeCode(hwnd) != WT_Slider
	return False
endIf
name = MSAA__windowName(hwnd)
if name
	; TT Classic 5.1.3 names the sliders natively.
	return False
endIf
var handle h1 = getFirstWindow(hwnd)
var int i = 0
var string names = msgSliderNames
while h1
	if getWindowTypeCode(h1) == WT_Slider
		i = i +1
	endIf
	if h1 == hwnd
		name = stringSegment(names, "|", i)
		jcp__set(props, "Name", name)
		return True
	endIf
	h1 = getNextWindow(h1)
endWhile
return False
endFunction

string function getObjectValue(/*?*/ int alwaysUseMSAA, int nLevel)
var string val = getObjectValue(alwaysUseMSAA, nLevel)
if nLevel != 0
	return val
endIf
if !util__isSpecialFocus() && isPCCursor()
&& getWindowClass(getFocus()) == "SysTreeView32"
	var string flags = roomtree__iconStatus()
	var string buf = ""
	if stringContains(flags, "s")
		buf = buf +" " +msgSending
	endIf
	if stringContains(flags, "i")
		buf = buf +" " +msgIdle
	endIf
	if stringContains(flags, "o")
		buf = buf +" " +msgOnline
	endIf
	val = stringTrimTrailingBlanks(val)
	if val && !getObjectName(True)
		val = ""
	endIf
	buf = stringChopLeft(buf, 1)
	val = string__join(val, " - ", buf)
endIf
return val
endFunction

globals int _sayingTitle
script sayWindowTitle()
_sayingTitle = True
performScript sayWindowTitle()
_sayingTitle = False
endScript

string function getWindowName(handle hwnd)
var string name = getWindowName(hwnd)
if _sayingTitle
&& hwnd == getAppMainWindow(getFocus())
	var string server = roomtree__serverName(hwnd)
	if server
		name = server +", " +name
	endIf
endIf
return name
endFunction

globals int _insideSAV
script sayAppVersion()
_insideSAV = True
performScript sayAppVersion()
_insideSAV = False
endScript

string function getVersionInfoString(string path, string field)
if _insideSAV
&& field == "ProductVersion"
	path = stringReplaceSubstrings(path, "Classic.exe", ".dll")
endIf
var string sVer =  getVersionInfoString(path, field)
sVer = stringReplaceSubstrings(sVer, ", ", ".")
return sVer
endFunction

int function caretVisible()
; Override to avoid occasional duplicate speech in the channel treeView caused by this natively returning True sometimes.
if getWindowClass(getFocus()) == "SysTreeView32"
	return False
endIf
return caretVisible()
endFunction

script sayFrameAtCursor()
stats__show()
endScript


;================================= Module chat4 ================================
; ChatOutput window interface for TeamTalk Classic
; This is an iTextDocument interface.
;
; Interface:
;	chat4__record(): Get the data record for the current or a specific chat.
;	chat4__messageIndex() and chat4__messageCount(): Position in and count of messages.
;	chat4__messageText([index]): Text of current message or message with given 1-based index.
;	chat4__setPos(index):  Set the "current" message.
;	chat4__move(delta):  Move delta (plus or minus) messages from current position.
; Attempting to move outside the valid message range causes a beep and stays at the border encountered..
;
; Author:  Doug Lee


globals
; Records for chat windows, by top-level window handle. Fields:
;	hTop: The top-level window (same as the key for this record).
;	hwnd, oDoc: The actual chat window and its RichEdit document object.
;	type: The type of the window: c=channel, p=private.
;	reversed: Present and True if this is a reversed window,
;		meaning the latest chat message is at the top.
;	unit: The TOM_* text unit for moving among messages.
;	firstStart, firstIdx: The starting position and index of the first "real" message.
;	curStart: The starting position of the range for the current text unit.
;	range: A Range object set by chat4___messageIndex() and reused by chat4__messageText() to get message text.
; Note: The caller of this module manages a reviewMessageNumber field as well.
; TODO: That should be handled in here.
	collection chat4___chats

collection function chat4__record(/*?*/ handle hwnd)
; Get or make the record for the current or given chat.
; Allows fields to be filled in that were missed at first.
if !hwnd || hwnd == 0
	hwnd = getFocus()
endIf
var handle hTop = ttutil__mainWindow(hwnd, True)
if !hTop || !isWindowVisible(hTop)
	return
endIf
var collection rec = chat4__winrecs__record(chat4___chats, hTop)
if !rec
	return
endIf

; Build or finish the record.

; Top-level window and navigation unit.
if !collectionItemExists(rec, "hTop")
|| !getWindowStyleBits(rec.hTop)
	rec.hTop = hTop
	; Public messages appear to be considered paragraphs.
	; In a private message, the time/author header is a bolded paragraph,
	; and the message text is one or more normal paragraphs.
	rec.unit = 4  ; TOMParagraph
endIf

; Chat window and chat type.
if !collectionItemExists(rec, "hwnd")
|| !getWindowStyleBits(rec.hwnd)
	hwnd = findWindow(hTop, "RichEdit", "")
	if !hwnd
		chat4__winrecs__remove(chat4___chats, hTop)
		return
	endIf
	rec.hwnd = hwnd
	if findWindow(hTop, "SysTreeView32", "")
		rec.type = "c"
	elif findWindow(hTop, "Button", chat4___scSend)
		rec.type = "p"
		rec.reversed = True
	else
		chat4__winrecs__remove(chat4___chats, hTop)
		return
	endIf
endIf

; Chat window's RichEdit interface.
var object o
if !collectionItemExists(rec, "oDoc")
|| !rec.oDoc.storyCount
	o = getRichEditDocument(rec.hwnd)
	if !o
		stats__inc("chatNoRichEditDocObjectForWindow" +intToString(rec.hwnd))
		chat4__winrecs__remove(chat4___chats, hTop)
		return
	endIf
	rec.oDoc = o
else
	o = rec.oDoc
endIf

if !collectionItemExists(rec, "firstStart")
	rec.firststart = 0
	rec.curstart = 0
endIf

; Set rec.firststart if necessary and possible.
if !rec.firststart
	if chat4___size(rec.hwnd) < 10
		return
	endIf
	var object oRange = o.Range(0, 0)
	var int safety = 0
	while safety < 10 && oRange.char <= 13
		safety = safety +1
		oRange.start = oRange.start +1
	endWhile
	if safety >= 10
		rec.firststart = 0
		rec.firstidx = 0
		;sayInteger(chat4___size(rec.hwnd))
	else
		rec.firststart = oRange.Start
		rec.firstidx = oRange.getIndex(rec.unit)
	endIf
endIf
return rec
endFunction

int function chat4__clean(/*?*/ int clearAll)
; Clean out invalid or all chat windows.
; Returns the number of windows removed
var int nremoved = chat4__winrecs__clean(chat4___chats, clearAll)
return nremoved
endFunction

int function chat4___end(collection rec)
; Return the current endpoint of the indicated chat's text range.
; This is used in speaking new text on arrival.
var object o = rec.oDoc
if !o
	return 0
endIf
var object oRange = o.range(0, 0x7FFFFFFFL)
return oRange.end +0
endFunction

int function chat4___size(handle hwnd)
; Return the "size" of the material in the given message area window.
; This is used to detect changes.
return sendMessage(hwnd, 0x0E, 0, 0)  ; WM_GetTextLength
endFunction

int function chat4___messageIndex(int start, /*?*/ variant which)
; Return the 1-based index of the message whose start position is given, or 0 on failure..
; A start position of -1 means the last message.
; Blank lines at the top and bottom of the buffer are avoided.
; Which can bbe a window handle, a record collection, or omitted.
; This function is also used internally to set the rec.range object.
var collection rec
if !which
	; Record for currently focused chat.
	rec = chat4__record()
elif intToString(which)
	; Record for given window handle.
	rec = chat4__record(which)
else
	; Record was passed directly.
	rec = which
endIf
if !rec
	return 0
endIf

var object oDoc = rec.oDoc
var int reversed = rec.reversed
var object oRange
if start == -1
	; Find the end of the buffer.
	oRange = oDoc.Range(0, 0)
	oRange.Expand(6)  ; 6 is TOMStory
	; A blank line tends to remain at the end of the buffer.
	; -2 here avoids it.
	start = oRange.End -2
endIf
; This avoids the blank line before the first message.
if start < rec.firststart
	start = rec.firststart
endIf
oRange = oDoc.Range(start, start)
oRange.Expand(rec.unit)
rec.range = oRange
return oRange.getIndex(rec.unit) -rec.firstidx +1
endFunction

int function chat4__messageIndex(/*?*/ handle hwnd)
; Returns the 1-based index of the last-read message.
var collection rec = chat4__record(hwnd)
return chat4___messageIndex(rec.curstart, rec)
endFunction

int function chat4__messageCount(/*?*/ handle hwnd)
; Returns the number of messages in the buffer.
; A return of -1 indicates an error.
; TODO: Error checking is not here yet.
var collection rec = chat4__record(hwnd)
return chat4___messageIndex(-1, rec)
endFunction

string function chat4__messageText(/*?*/ int idx, handle hwnd)
; Return the text of the current or indicated message.
; Negative numbers index from the end.
; The current message position is not changed.
var collection rec = chat4__record(hwnd)
var string txt
if idx
	if idx < 0
		chat4___messageIndex(-1, rec)
		; sets rec.range to the last message.
		if idx != -1
			chat4___move(idx+1, rec)
		endIf
	else  ; positive idx.
		chat4___messageIndex(rec.firststart, rec)
		; sets rec.range to the first message.
		if idx != 1
			chat4___move(idx-1, rec)
		endIf
	endIf
else
	; Current message wanted.
	chat4___messageIndex(rec.curstart, rec)
	; rec.range set and we don't have to move anywhere from here.
endIf
txt = rec.range.Text
return stringTrimLeadingBlanks(stringTrimTrailingBlanks(txt))
endFunction

int function chat4___move(int delta, /*?*/ collection rec)
; Move delta messages from the current position of rec.range.
; rec.range must be set first.
; Note that this is from rec.range, not rec.curstart.
; Returns the distance moved.
if !delta
	return 0
endIf
if !rec
	rec = chat4__record()
	if !rec
		return 0
	endIf
endIf
var int curidx = chat4___messageIndex(rec.range.Start, rec)
rec.range.Collapse(-1)  ; -1 is TOMTrue, for collapse to start.
if chat4___trimDelta(delta, curidx, rec)
	beep()
endIf
rec.range.Move(rec.unit, delta)
rec.range.Expand(rec.unit)
var int result =
util__abs(curidx -chat4___messageIndex(rec.range.Start))
return result
endFunction

int function chat4___trimDelta(int byRef delta, int curidx, collection rec)
; Forces delta not to move outside the valid messages.
; Returns True if delta is changed and False if not.
if delta < 0
	if curidx+delta < 1
		delta = 1-curidx
		return True
	endIf
else
	var object oRange = rec.range.duplicate()
	var int lastidx = chat4___messageIndex(-1, rec)
	rec.range = oRange
	if curidx +delta > lastidx
		delta = lastidx -curidx
		return True
	endIf
endIf
return False
endFunction

int function chat4__move(int delta, /*?*/ handle hwnd)
; Move delta messages from the current position.
; Returns the number (absolute value) of messages moved.
var int result
var collection rec = chat4__record(hwnd)
chat4___messageIndex(rec.curstart, rec)
; That sets rec.range to reflect the current position.
result = chat4___move(delta, rec)
rec.curstart = rec.range.Start
return result
endFunction

int function chat4__setPos(int idx, /*?*/ handle hwnd)
; Set the current message to the given index.
; Negative indices count from the end.
var collection rec = chat4__record(hwnd)
chat4__messageText(idx, rec.hTop)
rec.curstart = rec.range.Start
endFunction

globals handle chat4___giSNTHwnd, int chat4___giSNTStyle
string function chat4___title(handle hwnd, optional collection rec)
; Return a suitable title for the given window.
if !rec
	rec = chat4__record(hwnd)
endIf
var string title = ""
if !rec || rec.type == "c"
	; Main TT windows (visible or not; hence the !rec part).
	; Could use the window title...
	;title = getWindowName(ttutil__mainWindow(hwnd, True))
	;title = stringSegment(title, "-", 2)
	; but that is often a stale channel name and not a server name.
	; Use the top-level room tree node's name (possibly shortened).
	title = roomtree__serverName(hwnd)
else
	; Private chats are named in the title bar.
	title = getWindowName(getTopLevelWindow(hwnd))
	; ... after the word "Messages."
	title = stringSegment(title, "-", 2)
endIf
if stringIsBlank(title)
	; Catches private chats hidden with NUtils, for example.
	title = getWindowName(getTopLevelWindow(hwnd))
endIf
title = stringTrimLeadingBlanks(stringTrimTrailingBlanks(title))
return title
endFunction

int function chat4__focusNext(int delta)
; Focus the next or previous chat.
; Delta should be 1 for next and -1 for previous.
chat4__registerAllChatWindows()
var string hwnds = chat4__winrecs__list(chat4___chats, "n")
if !hwnds
	return False
endIf
var handle hTop = getTopLevelWindow(getFocus())
if !hTop
	return False
elif getParent(hTop)
	; For when a dialog is in focus over a TeamTalk main window.
	hTop = getParent(hTop)
endIf
var int idx = stringSegmentIndex(hwnds, "|", intToString(hTop), True)
var int n = stringSegmentCount(hwnds, "|")
if !idx
	if delta > 0
		idx = n +1
	else
		idx = 0
	endIf
endIf
var int tries = n
while tries
	tries = tries -1
	idx = idx +delta
	if idx < 1
		idx = idx +n
	elif idx > n
		idx = idx -n
	endIf
	var handle hwnd = stringToHandle(stringSegment(hwnds, "|", idx))
	if hwnd == hTop
		sayMessage(OT_Error, chat4___msgNoMoreWindows)
		setFocus(hTop)
		pause()
		sayObjectTypeAndText()
		return True
	endIf
	if chat4___focus(hwnd)
		return True
	endIf
	sayMessage(OT_Error, formatString(chat4___msgSkippingWindow, chat4___title(hwnd)))
endWhile
sayMessage(OT_Error, chat4___msgNoMoreWindows)
return False
endFunction

int function chat4___focus(handle hwnd)
; Focus a TT window.
if isWindowDisabled(hwnd)
	hwnd = chat4___getActiveDialog(hwnd)
	if !hwnd
		return False
	endIf
endIf
;util__activateWindow(hwnd)
setFocus(hwnd)
var int safety = 5
while safety
	if getTopLevelWindow(getFocus()) == getTopLevelWindow(hwnd)
	&& isWindowVisible(hwnd)
	&& !(MSAA__windowState(getForegroundWindow(), 0, ObjID_Window) & State_System_Invisible)
		return True
	elif !getFocus() && getForegroundWindow() == getTopLevelWindow(hwnd)
	&& MSAA__windowState(getForegroundWindow(), 0, ObjID_Window) & State_System_Invisible
	&& (getWindowStyleBits(getForegroundWindow()) & 0x20000000L)  ; minimized
		util__sysCommand(getForegroundWindow(), "restore")
		return True
	else
		pause()
	endIf
	safety = safety -1
endWhile
return False
endFunction

handle function chat4___getActiveDialog(handle hwnd0)
; Get the dialog that is disabling the given window.
hwnd0 = getTopLevelWindow(hwnd0)
var handle hwnd = getFirstWindow(hwnd0)
while hwnd
	if getParent(hwnd) == hwnd0
		return hwnd
	endIf
	hwnd = getNextWindow(hwnd)
endWhile
return 0
endFunction

void function chat4__registerAllChatWindows()
; Build or update the list of all valid chat windows.
var string thisOwner = getWindowOwner(getTopLevelWindow(getFocus()))
var handle hwnd = findTopLevelWindow("#32770", "")
while hwnd
	if isWindowVisible(hwnd)
	&& stringCompare(getWindowOwner(hwnd), thisOwner, True) == 0
	&& !getParent(hwnd)
		chat4__record(hwnd)
	endIf
	hwnd = getNextWindow(hwnd)
endWhile
chat4__clean()
endFunction


;-------------------------------- Module winrecs -------------------------------
; A manager for data structures (records) indexed by a window handle.
;
; Author:  Doug Lee of Level Access

collection function chat4__winrecs__record(collection byRef recs, handle hwnd, optional int noCreate)
; Return the record for the given window.
; Registers this window in the set if necessary.
; Also removes stale records for windows that no longer exist.
; The caller must handle record structure creation.
; Suggestion: Use an "hwnd" property so there's a uniform way to test for a new record.
; Pass noCreate if you need to avoid creating a new record.
chat4__winrecs__clean(recs)
if !hwnd
|| !getWindowStyleBits(hwnd)
	return
endIf
var string shwnd = intToString(hwnd)
var collection rec
if !recs
	recs = new collection
endIf
if !collectionItemExists(recs, shwnd)
	if noCreate
		return
	endIf
	; We have to make a new record for this one.
	rec = new collection
	recs[shwnd] = rec
	; Caller must detect that this record is empty and build it.
else
	rec = recs[shwnd]
endIf
return rec
endFunction

int function chat4__winrecs__remove(collection recs, handle hwnd)
; Remove a window's record.
var string shwnd = intToString(hwnd)
if !collectionItemExists(recs, shwnd)
	return False
endIf
collectionRemoveItem(recs, shwnd)
return True
endFunction

int function chat4__winrecs__clean(collection recs, /*?*/ int clearAll)
; Remove from recs any records for windows that no longer exist.
; If clearAll is True, removes all records.
; Returns the number of records removed.
; Note that this function is called by chat4__winrecs__record(),
; so manual use of it is optional.
var string shwnd
var int nremoved = 0
if clearAll
	var int n = 0
	forEach shwnd in recs
		n = n +1
	endForEach
	collectionRemoveAll(recs)
	return n
endIf
var handle hwnd
forEach shwnd in recs
	hwnd = stringToHandle(shwnd)
	if !hwnd || !getWindowStyleBits(hwnd)
		if chat4__winrecs__remove(recs, hwnd)
			nremoved = nremoved +1
		endIf
	endIf
endForEach
return nremoved
endFunction

string function chat4__winrecs__list(collection recs, /*?*/ string how)
; Return a |-delimited list of the registered window handles.
; By default, handles are returned in the order in which they were added to the set.
; If how is "n", they are returned in numeric order.
; If how is anything else, it is considered a function name or call.
; See util__compare() for details.
var string hwnds, string shwnd
forEach shwnd in recs
	hwnds = hwnds +"|" +shwnd
endForEach
hwnds = stringChopLeft(hwnds, 1)
if !how
	return hwnds
endIf
; Kludge: Make XList out of hwnds.
hwnds = "|" +hwnds +"|"
hwnds = chat4__winrecs__xlist__sort(hwnds, how, False)
; Remove the extra delimiters added above.
hwnds = stringChopLeft(stringChopRight(hwnds, 1), 1)
return hwnds
endFunction


;__________________________________ Module log _________________________________
; Allow logging to a text file.
;
; Interface:  chat4__winrecs__log__open, chat4__winrecs__log__isOpen, chat4__winrecs__log__close, chat4__winrecs__log__write, chat4__winrecs__log__setUserID.
; Basic usage: chat4__winrecs__log__open() from AutoStartEvent, chat4__winrecs__log__write("...") to log something.
; Usage:
;	- Call chat4__winrecs__log__open() from AutoStartEvent.
;	- Logging will occur by default only if the log file already exists (good for debugging during development).
;	- Call as chat4__winrecs__log__open("", True) if you want logging to occur even if the log file doesn't already exist.
;	- Pass a file name instead of "" for a custom file name (rarely needed).
;	- Call chat4__winrecs__log__write("Oops I saw an error") to log something.
;	- Use chat4__winrecs__log__write("Oops this error needs to speak too", True) to speak what is being logged.
;	- Call chat4__winrecs__log__close() to disable logging (rarely needed).
;
; Author:  Doug Lee of Level Access

; Constants for chat4__winrecs__log__write and related functions.
; Not localized because log text is not localized.

; Used as a user ID when none is found.
const chat4__winrecs__log___msgNoName = "<no name>"
; What "name" is used for the caller string when JAWS is too old for getScriptCallStack().
; %1 is the JAWS version.
const chat4__winrecs__log___msgJAWSTooOld = "(caller unknown; JAWS version (%1) too old)"
; Name used for the caller when it can't be determined for an unknown reason,
; probably a getScriptCallStack return value format surprise.
const chat4__winrecs__log___msgNoCaller = "(caller unknown)"

globals
; FileSystemObject and log stream object obtained from it.
	object chat4__winrecs__log___fso,
	object chat4__winrecs__log___logstream,
; Name of file.
	string chat4__winrecs__log___fileName,
	int chat4__winrecs__log___canCreate,
	string chat4__winrecs__log___userID,
	string chat4__winrecs__log___machineID

int function chat4__winrecs__log___setup()
; Internal function called to set things up when necessary before any log text is generated.
if !stringIsBlank(chat4__winrecs__log___userID)
	return True
endIf
chat4__winrecs__log___fso = createObjectEx("Scripting.FileSystemObject", False)
if !chat4__winrecs__log___fso
	return False
endIf
chat4__winrecs__log___machineID = ""
var object oNet = createObjectEx("WScript.Network", False)
if oNet
	chat4__winrecs__log___machineID = oNet.ComputerName
endIf
if stringIsBlank(chat4__winrecs__log___machineID)
	chat4__winrecs__log___machineID = chat4__winrecs__log___msgNoName
endIf
chat4__winrecs__log___userID = getJAWSUserName()
if stringIsBlank(chat4__winrecs__log___userID) && oNet
	chat4__winrecs__log___userID = oNet.UserName
endIf
if stringIsBlank(chat4__winrecs__log___userID)
	chat4__winrecs__log___userID = chat4__winrecs__log___msgNoName
endIf
return True
endFunction

int function chat4__winrecs__log__setUserID(string id)
; Optional means of setting a custom user ID.
; If not called, the value of getJAWSUserName() is used.
chat4__winrecs__log___userID = id
return True
endFunction

int function chat4__winrecs__log__close()
; Close the log file.
chat4__winrecs__log___close()
chat4__winrecs__log___fileName = ""
return True
endFunction

int function chat4__winrecs__log___setFileName(string name)
; Set the log file name.
; The name can be passed; otherwise <basename>.log is used.
; <basename> is getActiveConfiguration() or, failing that, the basename of getAppFileName().
; The log file is located in the JAWS user folder.
; Helper for chat4__winrecs__log__open().
var string basename
basename = name
if stringRight(basename,4) == ".log"
	basename = stringChopRight(basename, 4)
endIf
if !basename
	basename = getActiveConfiguration()
endIf
if !basename
	basename = getAppFileName()
	var int pos = stringContainsFromRight(basename, ".")
	if pos > 0
		basename = stringLeft(basename, pos-1)
	endIf
endIf
chat4__winrecs__log___fileName = getJAWSSettingsDirectory() +"\\" +basename
if stringRight(chat4__winrecs__log___fileName, 4) != ".log"
	chat4__winrecs__log___fileName = chat4__winrecs__log___fileName +".log"
endIf
return True
endFunction

int function chat4__winrecs__log__open(/*?*/ string name, int canCreate)
; Open and optionally create the log file.
; The name can be passed; otherwise <basename>.log is used.
; <basename> is getActiveConfiguration() or, failing that, the basename of getAppFileName().
; The log file is located in the JAWS user folder.
; Call from autoStartEvent, or logging will not occur.
; Returns True if the file is opened and False if not.
if !chat4__winrecs__log___setup()
	return False
endIf
chat4__winrecs__log___canCreate = canCreate
chat4__winrecs__log__close()
if !chat4__winrecs__log___setFileName(name) || !chat4__winrecs__log___open()
	chat4__winrecs__log___fileName = ""
	return False
endIf
return True
endFunction

int function chat4__winrecs__log___open()
; Raw open/create.  Sets chat4__winrecs__log___logstream.
; Returns True on success and False on failure.
if chat4__winrecs__log___logstream
	return True
endIf
if fileExists(chat4__winrecs__log___fileName)
	chat4__winrecs__log___logstream = chat4__winrecs__log___fso.openTextFile(chat4__winrecs__log___fileName, 8)  ; Append
elif chat4__winrecs__log___canCreate
	chat4__winrecs__log___logstream = chat4__winrecs__log___fso.CreateTextFile(chat4__winrecs__log___fileName)
endIf
if !chat4__winrecs__log___logstream
	return False
endIf
scheduleFunction("chat4__winrecs__log___close", 30)
return True
endFunction

int function chat4__winrecs__log___close()
; Raw cloase; unsets chat4__winrecs__log___logstream.
var variant null
if !chat4__winrecs__log___logstream
	return True
endIf
chat4__winrecs__log___logstream.Close()
chat4__winrecs__log___logstream = null
return True
endFunction

int function chat4__winrecs__log___isOpen()
; Returns True if the log file is actually open and False if not.
if chat4__winrecs__log___logstream
	return True
endIf
return False
endFunction

int function chat4__winrecs__log__isOpen()
; Returns True if the log file is open and False if not.
; "Open" here means can be written to but may not necessarily mean the file is physically opened at the moment.
if chat4__winrecs__log___fileName
	return True
endIf
return False
endFunction

string function chat4__winrecs__log___getStamp()
; Returns a formatted date/time stamp for the log file.
var string tm = sysGetTime("HH:mm:ss")
var string dt = sysGetDate("ddd MMM dd yyyy")
return formatString("%1 %2", tm, dt)
endFunction

int function chat4__winrecs__log__write(string s, /*?*/ int alsoSpeak)
; Write a string to the log file, and also optionally speak it.
if alsoSpeak
	sayMessage(OT_Error, s)
endIf
if !chat4__winrecs__log__isOpen()
	return False
endIf
if !chat4__winrecs__log___open()
	return False
endIf
var string buf = formatString(
	"*** %1 %2@%3 (%4)\n%5:\n",
	chat4__winrecs__log___getStamp(), chat4__winrecs__log___userID, chat4__winrecs__log___machineID, getAppFileName(),
	chat4__winrecs__log___getCaller()
) +s +"\n\n"
; Standardize any CR/LF characters in buf before writing.
buf = stringReplaceChars(buf, "\r\n", "\n")
buf = stringReplaceSubstrings(buf, "\n", "\r\n")
chat4__winrecs__log___logstream.write(buf)
return True
endFunction

string function chat4__winrecs__log___getCaller()
; Get and return the caller of this logging system.
; Helper for chat4__winrecs__log__write().
if getJFWVersion() < 1000000
	var string jver = stringSegment(util__getJAWSInfo(), "|", 1)
	return formatString(chat4__winrecs__log___msgJAWSTooOld, jver)
endIf
var string stack = util__callStackInfo(2, "s", 2)
if !stringIsBlank(stack)
	return stack
endIf
return formatString("%1\nStack begin:\n%2\nStack end.",
	chat4__winrecs__log___msgNoCaller,
	util__vcast(getScriptCallStack())
)
endFunction

string function chat4__winrecs__log___splitOff(string byRef s, string sep, /*?*/ int fromRight)
; Split what's before sep off and return it.
; If fromRight is True, split everything before the last sep off and return that.
; Returns by reference s with both that and sep removed.
var string s1
var int pos
if fromRight
	pos = stringContainsFromRight(s, sep)
else
	pos = stringContains(s, sep)
endIf
if pos
	s1 = stringLeft(s, pos-1)
	s = stringChopLeft(s, pos+stringLength(sep)-1)
else
	s1 = s
	s = ""
endIf
return s1
endFunction

;________________________________ End module log _______________________________


;_________________________________ Module xlist ________________________________
; XList, the arbitrary list (array and hash emulation) module.
;
; Terminology:
;	XList - An arbitrary list of items (an array, basically).
;	XLHash - An XList in which odd-numbered items are keys and even-numbered items are their values.
;
; Implementation Notes:
; XLists always begin and end with a delimiter; a null list is just a single delimiter character.
; This means it is NOT possible to use boolean (if sl) or length (if stringLength(sl)) to check for emptiness; use isEmpty or itemCount.
; Objects can not be stored in an XList directly, and attempting this will cause an error message to speak.
;
; Author:  Doug Lee
;
; Revision History:  See version control logs.

const
; Indexes used to request the character tables supported by this module.
	chat4__winrecs__xlist___CharANSI = 1,
; The corresponding character tables.
; The double quote (") and backslash characters must be escaped.
	chat4__winrecs__xlist___CharANSICharList = "\1\2\3\4\5\6\7\8\9\10\11\12\13\14\15\16\17\18\19\20\21\22\23\24\25\26\27\28\29\30\31 !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~\127\128\129\130\131\132\133\134\135\136\137\138\139\140\141\142\143\144\145\146\147\148\149\150\151\152\153\154\155\156\157\158\159\160\161\162\163\164\165\166\167\168\169\170\171\172\173\174\175\176\177\178\179\180\181\182\183\184\185\186\187\188\189\190\191\192\193\194\195\196\197\198\199\200\201\202\203\204\205\206\207\208\209\210\211\212\213\214\215\216\217\218\219\220\221\222\223\224\225\226\227\228\229\230\231\232\233\234\235\236\237\238\239\240\241\242\243\244\245\246\247\248\249\250\251\252\253\254\255"

string function chat4__winrecs__xlist__sort(string xl, string how, /*?*/ int sortDescending)
; Sorts xl and returns the result.
; How is as for util__compare().
; This is a stable sort; equal items are not swapped.
; TODO:  The algorithm is slow:  sorting 200 items took nine seconds on a Sony Vaio 1.2 GHz Centrino machine (Windows XP, JAWS 7.10, DGL, 2006-08-25).
var
	string xl1,
	string tmp, string firstVal, int firstIdx,
	int i, int n, int cmp
; Using the below trick instead of chat4__winrecs__xlist__new() preserves the separator from xl on return.
xl1 = stringLeft(xl, 1)
n = chat4__winrecs__xlist__itemCount(xl)
if !n
	return xl1
elif n == 1
	; The commented-out approach guarantees an XList structure return;
	; the faster plan, in effect, allows single-item strings to pass through unchecked.
	;chat4__winrecs__xlist__addItem(xl1, chat4__winrecs__xlist__getItem(xl, 1))
	;return xl1
	return xl
endIf

; We have two or more items.
while n > 1
	firstVal = chat4__winrecs__xlist__getItem(xl, 1)
	firstIdx = 1
	i = 2
	while i <= n
		tmp = chat4__winrecs__xlist__getItem(xl, i)
		cmp = util__compare(firstVal, tmp, how)
		if sortDescending
			cmp = cmp * (-1)
		endIf
		if cmp > 0
			firstVal = string__ref(tmp)
			firstIdx = i
		endIf
		i = i +1
	endWhile
	; firstVal and firstIdx are the value and index of the "first" value, respectively.
	; "First" means smallest for ascending and greatest for descending sorts.
	chat4__winrecs__xlist__addItem(xl1, firstVal)
	chat4__winrecs__xlist__delete(xl, firstIdx, 1)
	n = n -1
endWhile
; What's left is the max item.
chat4__winrecs__xlist__addItem(xl1, chat4__winrecs__xlist__getItem(xl, 1))
return xl1
endFunction

int function chat4__winrecs__xlist__delete(string byRef lst, int i, int n)
; Deletes n elements starting at element i.
var
	int pos1, int pos2
pos1 = chat4__winrecs__xlist___indexPos(lst, i)
if !pos1 || pos1 == stringLength(lst)+1
	return False
endIf
pos2 = chat4__winrecs__xlist___indexPos(lst, i+n)
if !pos2 || pos2 == stringLength(lst)+1
	lst = stringLeft(lst, pos1-1)
	return True
endIf
lst = string__cat(stringLeft(lst, pos1-1), stringChopLeft(lst, pos2-1))
return True
endFunction

int function chat4__winrecs__xlist__addItem(string byRef lst, variant item)
; Append item to lst.
; If lst is null, it will be created first; but if it is non-null and not an XList, a mess is likely.
; Returns True on success and False on failure.
; On failure, lst is not changed (except for maybe being initialized if null).
if !stringLength(item) && item
	sayMessage(OT_No_Disable, chat4__winrecs__xlist___msgNoObjects)
	return False
endIf
if !chat4__winrecs__xlist___settleDelim(lst, item, 1)
	return False
endIf
lst = string__cat(lst, string__cat(item, stringLeft(lst, 1)))
return True
endFunction

variant function chat4__winrecs__xlist__getItem(string lst, int i)
; Returns the ith element of lst.
; If i is greater than the item count, the null string is returned.
var string val
if i < 1 || !stringLength(lst)
	return ""
endIf
val = stringSegment(lst, stringLeft(lst, 1), i+1)
return val
endFunction

int function chat4__winrecs__xlist__itemCount(string lst)
; Returns the number of items in lst.
; If lst is not a proper XList, returns 0.
if !stringLength(lst)
	return 0
endIf
return string__segmentCount(lst, stringLeft(lst, 1)) -2
endFunction

int function chat4__winrecs__xlist___settleDelim(string byRef lst, string item, int dlen)
; Internal function to manage the changing of delimiters as needed.
; Must be called whenever material is being added to an XList.
; dlen is the length of lst's current delimiter (1 except when sent via fromJAWSList).
; Returns True if a valid delimiter is in place on exit and False if there is no safe delimiter available.
; A "valid" delimiter is a one-character delimiter not found in lst (except as the current delimiter) or item.
var
	string delim,
	int c0, int c, string ch,
	int done
if !stringLength(lst)
	lst = chat4__winrecs__xlist__new()
endIf
delim = stringLeft(lst, dlen)
if stringContains(item, delim) || dlen != 1
	c = chat4__winrecs__xlist__charToANSI(stringLeft(lst, 1))
	c0 = c
	done = False
	while !done
		c = c +1
		if c > 254
			; 255 may not write correctly to a file, and 0 is not possible in JAWS.
			c = 1
		elif c >= 8 && c <= 13
			; backspace, tab, LF, RF(?), FF, CR:
			; These characters are unfriendly when written to a text file, so they are avoided.
			c = 14
		elif c == 26 || c == 27
			; DOS/Windows EOF, ESC
			c = 28
		endIf
		ch = chat4__winrecs__xlist__charFromANSI(c)
		if c == c0
			; We've come all the way around; there is no safe delimiter for this situation.
			beep()
			sayMessage(OT_No_Disable, chat4__winrecs__xlist___msgOutOfDelims)
			return False
		elif !stringContains(lst, ch) && !stringContains(item, ch)
			; We have a winner!
			chat4__winrecs__xlist___setDelim(lst, ch, dlen)
			delim = ch
			done = True
		endIf
	endWhile
endIf
return True
endFunction

int function chat4__winrecs__xlist__charToANSI(string c)
; Returns the ANSI code for the first character in c, or 0 if c is null.
return chat4__winrecs__xlist___getCharCode(c, chat4__winrecs__xlist___CharANSI)
endFunction

string function chat4__winrecs__xlist__charFromANSI(int i)
; Returns the character corresponding to the given ANSI code.
return chat4__winrecs__xlist___getChar(i, chat4__winrecs__xlist___CharANSI)
endFunction

string function chat4__winrecs__xlist__new()
; Create and return a new XList.
return "\1"
endFunction

int function chat4__winrecs__xlist___indexPos(string lst, int i)
; Returns the character position in lst of the first character of the ith element.
; Returns 0 if there is no such element.
; Exception:  If i == itemCount+1, returns stringLength(lst)+1.
var
	string delim,
	int len,
	int pos,
	int result
if i < 1
	return 0
endIf
delim = stringLeft(lst, 1)
len = stringLength(lst)
if !len
	return 0
elif len == 1
	return 1
endIf
result = 0
while i
	pos = stringContains(lst, delim)
	if !pos
		return 0
	endIf
	result = result +pos
	lst = stringChopLeft(lst, pos)
	i = i -1
endWhile
return result +1
endFunction

int function chat4__winrecs__xlist___getCharCode(string s, int whichList)
; Converts the first character in s to the corresponding code based on the indicated character table.
; If s is null, 0 is returned.
; Can't handle \000 because this seems to end JAWS strings.
; In fact, including \000 anywhere in a string causes it to appear empty!
; [DGL, 02/28/03]
s = stringLeft(s, 1)  ; in case s contains more than one character
if !stringLength(s)
	return 0
endIf
return stringContains(chat4__winrecs__xlist___getCharList(whichList), s)
endFunction

string function chat4__winrecs__xlist___getChar(int i, int whichList)
; Converts the given code to the corresponding character from the indicated character table.
; If i is less than 1 or above the length of the table, the null string is returned.
var string charList
if i <= 0
	return ""
endIf
charList = chat4__winrecs__xlist___getCharList(whichList)
if i > stringLength(charList)
	return ""
endIf
return substring(charList, i, 1)
endFunction

string function chat4__winrecs__xlist___getCharList(int whichList)
; Returns the requested character list.  This is to avoid duplicating huge constants in the script binary.
if whichList == chat4__winrecs__xlist___CharANSI
	return chat4__winrecs__xlist___CharANSICharList
endIf
return ""
endFunction

int function chat4__winrecs__xlist___setDelim(string byRef lst, string ch, int dlen)
; Sets lst's delimiters to ch.
; dlen is the length of lst's current delimiter (1 except when sent via fromJAWSList).
var
	string delim
delim = stringLeft(lst, dlen)
if stringLength(ch) == 1 && dlen == 1
	lst = string__ref(stringReplaceChars(lst, delim, ch))
else
	; {To/From}JAWSList() can send a multi-character delimiter here, but no proper XList will use one.
	lst = string__ref(stringReplaceSubstrings(lst, delim, ch))
endIf
return True
endFunction

;_______________________________ End module xlist ______________________________

;------------------------------ End module winrecs -----------------------------

;=============================== End module chat4 ==============================


;=============================== Module roomtree ===============================
; Room tree interface for TeamTalk Classic.
;
; Author:  Doug Lee

/*
Design info:
The following apply to the SysTreeView32 control supported by this module,
and unless otherwise noted, this module depends on these features to
work properly:

Each node has a child ID; there are no full object descendants (this
is normal for a treeview).

Child IDs are numerically contiguous at first, though they may not
start at 1.  (Most treeviews do not allow any assumptions about childID
contiguousness.)  As people join and leave the server though, holes in
the childID contiguousness develop.

A childID is never reused, even when a user leaves one server and
joins another, which means childIDs increase monotonically for as long
as the user keeps TeamTalk open.  (This fact is not used here.)

Traversing by incrementing childID seems to go in breadth-first order
(not used here).

Traversing by AccNavigate(NavDir_Down,...) goes in depth-first order
(normal for treeviews).

Spatial navigation provides correct parent/child relationships (normal
for treeviews).

The tree consists of all nodes (channels and users) immediately on server
join.  (Normal trees can leave out subnodes until the user expands them.)

Channel nodes are denoted by a parenthesized number at the end of
accName, (0) for a channel with no members.

A channel can have both member and channel descendants, but the number
in its name represents only members in that channel.

A channel's Open state flag is not reset when the channel empties even
though there may no longer be descendants.
(This module makes no use of state flags; this is just written here
because it is unusual.)

This information comes from research by DGL on 2010-05-30,
supplemented on 2010-08-30.
*/

globals
; A list of tree window handles.
	string roomtree___hwnds,
; A collection of storage structures by server tree window handle.
; Fields:
;	hwnd, client:  The window handle and MSAA client object for the tree.
;	nodes:  The raw list of node names at last check.
	collection roomtree___trees,
; First and last childIDs in numerical order.
; Used for byAge navigation.
	int roomtree___firstID, int roomtree___lastID,
; Used to avoid very time-consuming getColorAtPoint calls on Win8
; and/or high-DPI laptops.
	collection roomtree___flagsByCRC,
	collection roomtree___flagInfo

int function roomtree__reportMemberships(int includeList, /*?*/ int startNode)
; Report who is where in the current TeamTalk server's channel tree.
; includeList: True to include names, False to show only counts per channel.
; TODO: This function is not fully written.
var collection tree = roomtree___treeRecord()
if !tree
	sayMessage(OT_Error, roomtree___msgNoTree)
	return False
endIf
var object o = tree.client
if !startNode
	; Start on the root node.
	startNode = 1
endIf
var string names = ""
var int childID = o.accNavigate(7, startNode)  ; 7 is NavDir_FirstChild
var int safety = 1000
var string name, int memberCount
while childID && safety
	roomtree___getNameAndCount(o.accName(childID), name, memberCount)
	if includeList
		if memberCount == -1
			; A member, not a channel.
		endIf
	else
	endIf
	childID = o.accNavigate(2, childID)  ; 2 is NavDir_Down
	safety = safety -1
endWhile
; Safety ran out; a bad thing.
beep()
return False
endFunction

collection function roomtree___treeRecord(/*?*/ handle hwnd)
; Returns the record for the given or currently focused TeamTalk's channel tree.
; Registers this tree in roomtree___trees if necessary.
; Also removes stale records for windows that no longer exist.
; Does not update the record itself from the tree.
roomtree___cleanOutStaleWindows()
var variant null
if hwnd
	hwnd = findWindow(hwnd, "SysTreeView32", "")
	if !hwnd
		return
	endIf
else
	hwnd = roomtree__currentTreeWindow()
	if !hwnd
		return
	endIf
endIf
var string shwnd = intToString(hwnd)
var collection tree, object o, int childID
if collectionItemExists(roomtree___trees, shwnd)
tree = roomtree___trees[shwnd]
endIf
if !tree || !tree.client.accRole(0)
	; We have to make a new record for this one.
	o = getObjectFromEvent(hwnd, -4, 0, childID)  ; -4 is ObjID_Client
	if !o
		return False
	endIf
	roomtree___hwnds = formatString("%1|%2", roomtree___hwnds, shwnd)
	tree = new collection
	if !roomtree___trees
		roomtree___trees = new collection
	endIf
	roomtree___trees[shwnd] = tree
	tree.hwnd = hwnd
	tree.client = o
	var collection nodes nodes = new collection
	tree.nodes = nodes
else
	tree = roomtree___trees[shwnd]
endIf
return tree
endFunction

int function roomtree___cleanOutStaleWindows()
; Remove from roomtree___trees any records for windows that no longer exist.
; Returns the number of records removed.
var int n = stringSegmentCount(roomtree___hwnds, "|")
var int nremoved = 0
var string shwnd
while n
	shwnd = stringSegment(roomtree___hwnds, "|", n)
	if getWindowClass(stringToHandle(shwnd)) != "SysTreeView32"
		nremoved = nremoved +1
		collectionRemoveItem(roomtree___trees, shwnd)
		if !string__segmentRemove(roomtree___hwnds, "|", shwnd)
			chat4__winrecs__log__write("Failed to remove handle " +shwnd +" from the tree list.")
		endIf
	endIf
	n = n -1
endWhile
return nremoved
endFunction

handle function roomtree__currentTreeWindow()
; Return the handle of the channel tree for the focused TeamTalk instance,
; or return null on error.
var handle hwnd = ttutil__mainWindow()
if !hwnd
	return 0
endIf
hwnd = findWindow(hwnd, "SysTreeView32", "")
if !hwnd
	return 0
endIf
return hwnd
endFunction

int function roomtree___getNameAndCount(string s, string byRef name, int byRef count)
; Return the name and count from the given channel tree node name text.
; If the text represents a user name and not a channel (no count included),
; the returned count will be -1.
s = stringTrimLeadingBlanks(stringTrimTrailingBlanks(s))
var int pos = stringContainsFromRight(s, " ")
if !pos || substring(s, pos+1, 1) != "("
|| stringRight(s, 1) != ")"
	count = -1
	name = s
	return True
endIf
name = stringLeft(s, pos-1)
var string tmp = stringChopRight(stringChopLeft(s, pos+1), 1)
var int n = stringLength(tmp)
var int i = 1
while i <= n
	if !stringContains("0123456789", substring(tmp, i, 1))
		name = s
		count = -1
		return True
	endIf
	i = i +1
endWhile
count = stringToInt(tmp)
return True
endFunction

int function roomtree__isMember(object o, int childID)
; Returns True if this is a member node, as opposed to a channel.
var string name, int count
if !o.accRole(childID)
	; This happens for a skipped childID in byAge navigation.
	return False
elif !roomtree___getNameAndCount(o.accName(childID), name, count)
	; Unrecognized nodes are called members here.
	beep()
	return True
endIf
return (count == -1)
endFunction

int function roomtree__getMember(int childID, string which, /*?*/ int byAge, int alsoSpeak)
; Get the child ID of a member based on which and, if passed, byAge.
; "Member" means person, as opposed to channel.
; If childID is 0 or invalid, starts as appropriate from an end of the list or the current position.
; If byAge is False or not passed, uses tree order.
; If byAge is True, attempts to order members chronologically by when they last joined a channel.
; If alsoSpeak is True, speaks error messages as appropriate.
; Values of which:
;	first, last: First or last member.
;	next, prev: Next or previous member from given position.
;	nextCyclic, prevCyclic: Same but looping around the member list.
; On success, returns the childID of the member sought.
; On error, returns 0.
var collection tree = roomtree___treeRecord()
if !tree
	if alsoSpeak
		sayMessage(OT_Error, roomtree___msgNoTree)
	endIf
	return False
endIf
var object o = tree.client
if byAge
	roomtree___setFirstAndLastChildIDs(o)
endIf

; Handle next/prev requests, or if appropriate, make first/last requests out of them.
if which == "next" || which == "prev"
	if !childID
		childID = o.accFocus
	endIf
	if childID
		childID = roomtree___navNext(o, childID, (which == "next"), byAge, True)
	endIf
	if !childID
		; No current position or no next one,
		; so these become first/last requests in cyclic mode.
		; They also do this if no node is focused yet,
		; which happens when a server connection is first opened.
		if !stringContains(stringLower(which), "cyclic") && o.accFocus
			if alsoSpeak
				sayMessage(OT_Error, roomtree___msgNoMoreMembers)
			endIf
			return 0
		endIf
		if which == "next"
			which = "first"
		else  ; prev
			which = "last"
		endIf
	endIf
endIf

; Handle first and last, both by tree and by age.
if which == "first"
	; The first childID is also the oldest.
	childID = o.accNavigate(NavDir_FirstChild, 0)
	childID = roomtree___navNext(o, childID, True, byAge, False)
elif which == "last"
	if byAge
		childID = roomtree___lastID
	else
		childID = o.accNavigate(NavDir_LastChild, 0)
	endIf
	childID = roomtree___navNext(o, childID, False, byAge, False)
endIf
if !childID
	if alsoSpeak
		sayMessage(OT_Error, roomtree___msgNoMembers)
	endIf
endIf
return childID
endFunction

int function roomtree___navNext(object o, int childID, int forward, int byAge, int skipThisOne)
; Return the next childID in the requested direction and sort order.
; If skipThisOne is True, the given node is not considered, otherwise it is.
; For byAge True, assumes roomtree___firstID and roomtree___lastID are set.
; Helper for roomtree__getMember().
var int dir
if byAge
	if forward
		dir = 1
	else
		dir = -1
	endIf
else
	if forward
		dir = NavDir_Next
	else
		dir = NavDir_Previous
	endIf
endIf

while childID
	if !skipThisOne && roomtree__isMember(o, childID)
		return childID
	endIf
	if byAge
		childID = childID +dir
		if childID < roomtree___firstID || childID > roomtree___lastID
			; We hit an edge of the valid childIDs.
			childID = 0
		endIf
	else
		childID = o.accNavigate(dir, childID)
	endIf
	skipThisOne = False
endWhile
return 0
endFunction

void function roomtree___setFirstAndLastChildIDs(object o)
; Sets roomtree___firstID and roomtree___lastID, used for byAge navigation.
; Helper for roomtree__getMember().
var int childID = o.accNavigate(NavDir_FirstChild, 0)
; This has to be the "oldest" childID because it represents the tree root,
; which is the first childID created.
roomtree___firstID = childID
; We know the last ID can't be less than this...
var int maxID = roomtree___firstID +o.accChildCount -1
; but without a loop, we can't know what the highest one really is.
while childID
	if childID > maxID
		maxID = childID
	endIf
	childID = o.accNavigate(NavDir_Next, childID)
endWhile
roomtree___lastID = maxID
endFunction

int function roomtree__move(string which, int byAge)
; Move to another member node.
; Returns True on success and False on failure.
var int childID = roomtree__getMember(0, which, byAge, True)
if !childID
	return False
endIf
var collection tree = roomtree___treeRecord()
if !tree
	return False
endIf
var object o = tree.client
var int focusID = o.accFocus +0
if childID == focusID
	beep()
	sayObjectActiveItem(False)
else
	; Leaving out SelFlag_TakeFocus avoids the following anomaly:
	; Focus remains, in MSAA and by window, on the right thing,
	; but Alt doesn't work, Alt+F4 destroys the tree leaving the app running,
	; and Tab and then Alt+F4 causes a busy loop terminated only by Alt+Tab.
	; Using TakeFocus+TakeSelection but calling AccSelect twice solves
	; the menu problem but causes the loop issue to occur on Alt+F4 without Tab.
	o.accSelect(SelFlag_TakeSelection, childID)
	if stringCompare(roomtree___treePath(o, childID), roomtree___treePath(o, focusID), True) != 0
		roomtree___sayNode()
	endIf
endIf
return True
endFunction

void function roomtree___sayNode()
; Say the path to the current tree node.
var collection tree = roomtree___treeRecord()
if !tree
	return
endIf
var object o = tree.client
var string path = roomtree___treePath(o, o.accFocus)
sayUsingVoice(VCTX_Message, path, OT_Message)
endFunction

string function roomtree___treePath(object o, int childID)
; Return the path from root to given tree node.
; Omits the given node and the top-level node.
var int safety = 50
var string path = ""
var string name, int cnt
while safety
	childID = o.accNavigate(NavDir_Left, childID)
	if childID && o.accNavigate(NavDir_Left, childID)
		if !roomtree___getNameAndCount(o.accName(childID), name, cnt)
			beep()
			return ""
		endIf
		if path
			path = ", " +path
		endIf
		path = name +path
	else
		safety = 1  ; exits loop
	endIf
	safety = safety -1
endWhile
return path
endFunction

void function roomtree__init(int level)
; Initialize structures used by this module.
; level: 0 ensure readiness, 1 clear on task switch, 2 full reset.
if !roomtree___flagInfo || level > 0
	roomtree___flagInfo = new collection
endIf
if !roomtree___flagsByCRC || level > 1
	roomtree___flagsByCRC = new collection
	roomtree___flagsByCRC.count = 0
	stats__set("iconStatusCRCCount", roomtree___flagsByCRC.count)
endIf
endFunction

string function roomtree__iconStatus()
; Indicates if the user at the active cursor position is transmitting sound and/or idle.
; Includes much caching to avoid two problems:
;	- getColorAtPoint can be slow (16 ms per call).
;		(Seems to apply on Win 8/8.1 and/or high-DPI laptops.)
;	- The OSM can position things wrongly on high-DPI laptops.
if !opt__get("AnalyzeTreeIcons")
	return ""
endIf
var int tc = getTickCount()

; First make sure all collections are ready for business.
roomtree__init(0)

; Now make sure the current focus position is ready for this.
var handle hwnd = getCurrentWindow()
; itemID is a treeView item ID, not an ordinal.
; We use it here to cache flags for the focused tree node.
var int itemID = tvGetSelection(hwnd)
var object o, int childID
o = getCurrentObject(childID)
if !o
	stats__inc("iconStatusNoMSAA")
	return ""
elif !childID
	; On top level of tree, no icons here.
	stats__inc("iconStatusNoChildID")
	return ""
	elif !o.accRole(childID)
		; Can happen when the tree is collapsed by a disconnect:
		; GetFocusObject() can return a stale childID.
		; [DGL, 2011-12-29, JAWS 13.0.704]
	stats__inc("iconStatusNoRole")
		return ""
endIf

; If we cached the flags for this node, return them.
; Scroll the tree as necessary to make sure we're looking at the right spot.
; The invisible cursor is also activated here.
var collection finfo = roomtree___flagInfo
roomtree___scrollTree(hwnd, o, childID, 20)
saveCursor() invisibleCursor() saveCursor()
if hwnd && itemID
&& hwnd == finfo.hwnd && itemID == finfo.itemID
	moveTo(finfo.x, finfo.y)
	if getGraphicID() == finfo.graphicID
		stats__inc("iconStatusInfoCacheHit")
		return finfo.flags
	endIf
	stats__inc("iconStatusInfoCacheMissStill")
else
	stats__inc("iconStatusInfoCacheMissOther")
endIf
var string flags = ""

; Get the area to scan for colors, and rule out non-member nodes.
; This sets x, y, and n.
if !roomtree__isMember(o, childID)
	; We don't have icon statuses for channel nodes.
	stats__inc("iconStatusNotMemberNode")
	return ""
endIf
var collection rect = roomtree___treeItemRect(o, childID)
if !rect
	stats__inc("iconStatusNoRect")
	return ""
endIf
var int x = rect.left-4
var int y = (rect.top+rect.bottom) /2
var int n = x -getWindowLeft(getCurrentWindow())
n = max(n, 0)
n = min(n, 15)

; GetColorAtPoint can be very slow, but getGraphicID() remains very fast.
; Find the graphic or fail, so we only check colors where they can be valid.
var int ptc = getTickCount()
var int gid = 0
while n && !gid
	moveTo(x, y)
	gid = getGraphicID()
	if !gid
		x = x -1
		n = n -1
	endIf
endWhile
if !gid
	return ""
endIf

; Since we have an id, we should soon get all we need;
; so store what we can in the finfo cache now.
finfo.hwnd = hwnd
finfo.itemID = itemID
finfo.x = x
finfo.y = y
finfo.graphicID = gid
; finfo.flags remains to be set.

; Now see if we already cached the flags for this graphic's CRC (ID).
var string cachedFlags = roomtree___flagsByCRC[intToString(gid)]
if cachedFlags
	flags = cachedFlags
	if flags == "none"
		flags = ""
	endIf
	stats__inc("iconStatusCRCCacheHit")
	finfo.flags = flags
	return flags
endIf
stats__inc("iconStatusCRCCacheMiss")

; We must determine the flags from the graphic.
var int clr
var int isSending = False
var int isIdle = False
var int foundIcon = False
while n && !flags
	clr = getColorAtPoint(x, y)
	if clr == 0x0000FFL
		isIdle = True
	elif clr == 0x00FF00L
		isSending = True
	elif (clr & 0xFF) == 228
		foundIcon = True
	endIf
	x = x -1
	n = n -1
endWhile
ptc = getTickCount() -ptc
stats__inc("iconStatusTickCount"+intToString(ptc))
if isIdle
	flags = flags +"i"
elif foundIcon
	flags = flags +"o"
endIf
if isSending
	flags = flags +"s"
endIf
if !flags
	flags = "none"
endIf
roomtree___flagsByCRC[intToString(gid)] = flags
roomtree___flagsByCRC.count = roomtree___flagsByCRC.count +1
stats__set("iconStatusCRCCount", roomtree___flagsByCRC.count)
stats__inc("iconStatusCRCCacheAdd")
if flags == "none"
	flags = ""
endIf
finfo.flags = flags
return flags
endFunction

int function roomtree___scrollTree(handle hwnd, object o, int childID, int reservedSpace)
; Scroll left/right the given tree as needed to make the following true
; regarding the MSAA rectangle's left edge (left):
;	- left is reservedSpace pixels or more right of the window's left edge.
;	- left is not right of the window's right edge.
var collection rect
var int wleft = max(getWindowLeft(hwnd), 1)
var int wright = min(getWindowRight(hwnd), screenGetWidth())
var int safety = 30
var int scrollLeft, int scrollRight
var int tryUIA = True
while safety
	rect = roomtree___treeItemRect(o, childID)
	if !rect
		return False
	endIf
	scrollLeft = (rect.left -wleft < reservedSpace)
	scrollRight = (wright -rect.left < 1)
	if scrollLeft && scrollRight
		return False
	endIf
	var int doScroll = (scrollLeft || scrollRight)
	if doScroll && tryUIA
		; UIA's scrollIntoView() is a better approach when available.
		; Only try once per scroll attempt though.
		tryUIA = False
		var object el = roomtree___UIAFromMSAA(o, childID)
		if el
			var object scrollPat = el.getScrollItemPattern()
			if scrollPat
				scrollPat.scrollIntoView()
				return True
			endIf
		endIf
	endIf
	if rect.left+rect.right+rect.top+rect.bottom <= 4
	|| (o.accState(childID) & State_System_Invisible)
		return False
	endIf
	if scrollLeft
		;sayString(formatString("Left %1", intToString(scrollLeft)))
		util__WMScroll(hwnd, "LineLeft")
	elif scrollRight
		;sayString(formatString("Right %1", intToString(scrollRight)))
		util__WMScroll(hwnd, "LineRight")
	else
		return True
	endIf
	safety = safety -1
endWhile
;sayString("Scroll fail")
return False
endFunction

string function roomtree__serverName(handle hwnd)
; Returns the server name given any window in its hierarchy.
; The server name is the top-level tree node's name.
var collection tree = roomtree___treeRecord(getTopLevelWindow(hwnd))
if !tree
	return ""
endIf
var object o = tree.client
if !o
	return ""
endIf
var int childID = o.accNavigate(NavDir_FirstChild, 0)
if !childID
	; The tree is empty.
	; This catches "messages" caused by failed server connects.
	return roomtree___serverNameFromStatus(getTopLevelWindow(hwnd))
endIf
var string name = o.accName(childID)
; Abbreviate this, some server names are long.
; This cuts off at the first listed punctuation mark.
name = stringSegment(name, ".,?/;:!()[]", 1)
return name
endFunction

string function roomtree___serverNameFromStatus(handle hwnd)
; Return the name of the server for the indicated TeamTalk instance from its status line if possible.
; This applies during server connect attempts.
var string txt = ttutil__statusText(hwnd)
if !txt
	return ""
endIf
var int pos = stringContains(txt, "\"")
if !pos
	return ""
endIf
txt = stringChopLeft(txt, pos)
pos = stringContains(txt, "\"")
if !pos
	return ""
endIf
txt = stringLeft(txt, pos-1)
return txt
endFunction

collection function roomtree___treeItemRect(object o, int childID)
; Attempt to get the correct bounding rectangle for the indicated MSAA item.
; Uses UIA in JAWS 14+ and MSAA in older versions,
; because UIA translates for high-DPI machines.
; Returns null on failure and a collection on success.
var collection rect
if getJFWVersion() < 1400000
	var int left, int right, int top, int bottom
	if !MSAA__rect(o, childID, left, right, top, bottom)
		return
	endIf
	rect = new collection
	rect.left = left
	rect.right = right
	rect.top = top
	rect.bottom = bottom
	return rect
endIf
var object el = roomtree___UIAFromMSAA(o, childID)
if !el
	return
endIf
rect = roomtree__UIA__boundingRectangle(el, True)
if !rect
	return
endIf
return rect
endFunction

object function roomtree___UIAFromMSAA(object o, int childID)
; Return the focused or indicated UIA element.
if getJFWVersion() < 1400000
	return
endIf
var object el = roomtree__UIA__elementFromMSAA(o, childID)
if !el
	if o.accState(childID) & State_System_Focused
		el = roomtree__UIA__focusedElement(True)
		if !el
			return
		endIf
	else
		return
	endIf
endIf
return el
endFunction


;---------------------------------- Module UIA ---------------------------------
; Code for accessing and manipulating the UIAScriptAPI in JAWS 14 and later.
;
; Globals not under namespace control: gc_cachedAPIObjects.
;
; Author: Doug Lee of Level Access

globals
; The main UIA interface, cached in case this speeds things.
	object roomtree__UIA___gUIA,
; An optional cache of API objects.
; This is a workaround for a JAWS (at least 14-18) issue whereby retrieval of an API object can start failing until the next JAWS restart.
; This global is deliberately JAWS-wide, not namespace-controlled, so that one script set building a cache will benefit all others.
	collection gc_cachedAPIObjects,
; An indicator of the version of the UIA support (see roomtree__UIA__version()).
; Set by the first roomtree__UIA__getScriptAPI() call.
	int roomtree__UIA___version,
; Lists of IDs from UIA.jsh / UI Automation.
	collection roomtree__UIA___props,
	collection roomtree__UIA___vals,
	collection roomtree__UIA___events,
	collection roomtree__UIA___attributes,
	collection roomtree__UIA___textUnits,
	collection roomtree__UIA___fullnames,
; A flag indicating when all those lists are completely built.
	int roomtree__UIA___built

; Messages for this code that should not be localized,
; because they are for scripters and should not be heard by users.
messages
@roomtree__UIA___msgInvalidScopeChar
Invalid search scope character: %1
@@
@roomtree__UIA___msgNullConditionString
Null condition string
@@
@roomtree__UIA___msgStringNotTerminated
String literal in condition spec was not terminated
@@
; %1 is a reference like %11 (only up to %10 allowed).
@roomtree__UIA___msgInvalidRef
Invalid reference in condition: %1
@@
; %1 is a list type (property, value, event, etc.).
; %2 is an invalid property/value/event/etc. name.
@roomtree__UIA___msgInvalidName
Module roomtree__UIA: Invalid %1 name %2
@@
; %1 is the number of unused stack values.
@roomtree__UIA___msgSpecIncomplete
%1 value(s) not used - missing trailing 'and'?
@@
@roomtree__UIA___msgUnknownError
Unknown error in condition specification
@@
endMessages

object function roomtree__UIA__getScriptAPI(optional int forceNew)
; Get the main UIAScriptAPI object.
; Also sets roomtree__UIA__version on first call.
; If forceNew is True, the object is a new reference.
; Otherwise it may (but may not) be a shared object.
; Shared objects may be shared by other scripts.
; You may want to pass True when hooking events.
if !roomtree__UIA___gUIA
	roomtree__UIA___gUIA = roomtree__UIA___rawCreate()
	if !roomtree__UIA___gUIA
		; Use the API object cache if somebody built one.
		roomtree__UIA___gUIA = roomtree__UIA___getCachedAPIObject(forceNew)
	endIf
	if !roomtree__UIA___gUIA
		return
	endIf
	if !roomtree__UIA___version
		; roomtree__UIA__version() sets roomtree__UIA___version to avoid recursion.
		roomtree__UIA___version = roomtree__UIA__version(roomtree__UIA___gUIA)
	endIf
endIf
scheduleFunction("roomtree__UIA___UIAClear", 50)
if forceNew
	var object newObj = roomtree__UIA___rawCreate()
	if !newObj
		; Use the API object cache if somebody built one.
		newObj = roomtree__UIA___getCachedAPIObject(forceNew)
	endIf
	return newObj
endIf
return roomtree__UIA___gUIA
endFunction

object function roomtree__UIA___rawCreate()
; Raw creator of the UIAScriptAPI object.
var object uia = createObjectEx("FreedomSci.UIA", FALSE, "UIAScriptAPI.x.manifest")
return uia
endFunction

int function roomtree__UIA__version(optional object oUIA)
; Indicate the version of the available or given UIA support:
;	-1 - Unknown.
;	0 - Not set yet.
;	1 - JAWS 14.
;	2 - JAWS 15 before update 1.
;	3 - JAWS 15 update 1 (15.0.5056).
; The version of the available UIA support is cached by roomtree__UIA__getScriptAPI().
var int ver = -1
if !oUIA
	if roomtree__UIA___version
		; Already figured out for this JAWS instance.
		return roomtree__UIA___version
	endIf
	; roomtree__UIA__getScriptAPI() passes a valid object,
	; so this will not cause recursion.
	; Set roomtree__UIA___version here to keep roomtree__UIA__getScriptAPI() from calling us though.
	roomtree__UIA___version = -1
	oUIA = roomtree__UIA__getScriptAPI()
	if !oUIA
		return -1
	endIf
endIf
var variant names = COMGetAvailableNames(oUIA)
names = stringLower(names)
if stringContains(names, "addtextedittextchangedeventhandler()")
	; Added in JAWS 15.0.5056.
	; Anything newer than that JAWS version must be detected here.
	ver = 3
elif stringContains(names, "events()")
	; Included in JAWS 14 but removed in JAWS 15.
	ver = 1
else
	; All we have left is the initial JAWS 15 release.
	ver = 2
endIf
return ver
endFunction

object function roomtree__UIA___getCachedAPIObject(int forceNew)
; Return (and remove from cache if not shared) a cached UIAScriptAPI object.
var collection cache = gc_cachedAPIObjects
if !cache
	return
endIf
var int i = 1
if forceNew
	i = cache.count
	if i <= 1
		; Out of new ones... (cache item 1 is reserved as the shared object).
		roomtree__UIA___reportAPICacheSize()
		return
	endIf
endIf
var object api = cache[intToString(i)]
if i > 1
	var variant null
	cache[intToString(i)] = null
	cache.count = cache.count -1
endIf
roomtree__UIA___reportAPICacheSize()
return api
endFunction

void function roomtree__UIA___reportAPICacheSize()
; Indicate to the user the API object cache size. This is called when the size changes.
var string fname = getJAWSSettingsDirectory() +"\\Sounds\\gauge" +intToString(gc_cachedAPIObjects.count) +".wav"
if !fileExists(fname)
	return
endIf
playSound(fname)
endFunction

object function roomtree__UIA__focusedElement(optional int deepest)
; Return the focused UIA element or null on error.
; If deepest is negative, returns the UIAScriptAPI.getFocusedElement() result without adjustment.
; For all other values of deepest, does the following:
;	- Accounts for anomalies in focus tracking with an active menu.
;	- Fixes a focus breakage that can occur in Silverlight windows inline while finding focus.
; If deepest is positive, also seeks the deepest focused element among alternatives when necessary.
; If deepest is specifically 2, this includes a search for a single selected item.
var object oUIA = roomtree__UIA__getScriptAPI()
if !oUIA
	return
endIf
if deepest < 0
	return oUIA.getFocusedElement().buildUpdatedCache()
endIf
var object el
var handle hwnd = getFocus()
if getWindowSubtypeCode(hwnd) == WT_ContextMenu
|| getObjectSubtypeCode() == WT_Menu
	; UIA can miss these and give application focus; fix this here.
	var handle hwnd1 = getWindowAtPoint(getCursorCol(), getCursorRow())
	if hwnd1 != getAppMainWindow(hwnd1)
		hwnd = hwnd1
	endIf
	el = roomtree__UIA__elementFromWindow(hwnd)
	if !el
		return
	endIf
	deepest = True
elif menusActive()
	; Happens in QT5 apps at least; MSAA can get the right focus when nothing else can.
	var object o, int childID
	o = getFocusObject(childID)
	if !o
		return
	endIf
	el = roomtree__UIA__elementFromMSAA(o, childID)
	if !el
		return
	endIf
	deepest = True
else
	roomtree__UIA___fixSilverlightFocus(hwnd)
	el = oUIA.getFocusedElement().buildUpdatedCache()
endIf
if !deepest
; 50003 is UIA_ComboBoxControlTypeId
|| el.controlType == 50003
; 50004 is UIA_EditControlTypeId
|| el.controlType == 50004
	; These can have descendants but we shouldn't need them.
	return el
endIf

; Before we try chasing focus the hard way, try a quick selection pattern.
; Does nothing in JAWS 14.
el = el.buildUpdatedCache()
var collection iter = roomtree__UIA__elementIterator(roomtree__UIA__selection(el, "af"))
if iter.count >= 1
	roomtree__UIA__nextElement(iter)
	var object sel = iter.currentElement
	; Allow first selection to be focus if it has focus.
	if sel.hasKeyboardFocus
		return sel
	endIf
endIf

var object cond = roomtree__UIA__condition("hasKeyboardFocus True")
if !cond
	beep()
	return el
endIf
var int moved = False
var int focusStopped = False
while !focusStopped && el
	; 4 is TreeScope_Descendants, 2 is Children.
	iter = roomtree__UIA__findAllIterator(el, 2, cond)
	if iter.count > 1
		; Ambiguous; just stop where we are.
		return el
	elif iter.count == 0
		; No more subitems focused at this level,
		; but maybe at lower ones...
		iter = roomtree__UIA__findAllIterator(el, 6, cond)
			; Focus searching will stop here since we just searched exhaustively.
			focusStopped = True
		if iter.count == 1
			; Deepest focus found unless there's selection beyond.
			roomtree__UIA__nextElement(iter)
			el = iter.currentElement
			moved = True
		elif iter.count > 1
			; Ambiguous; we have to stop here.
			return el
		else
			; No more focus; stop here.
		endIf
	else
		; Focus found at this level; keep going.
		roomtree__UIA__nextElement(iter)
		el = iter.currentElement
		moved = True
	endIf
endWhile
if focusStopped && deepest == 2
	var object el1 = roomtree__UIA__selection(el)
	if el1
		el = el1
		moved = True
	endIf
endIf
if moved
	el = el.buildUpdatedCache()
endIf
return el
endFunction

int function roomtree__UIA___fixSilverlightFocus(handle hwnd)
; When applicable, fix a problem with focus on Silverlight windows that breaks UIA's focusedElement result.
; Returns True if any change was attempted and False if not.
var object el = roomtree__UIA__elementFromWindow(hwnd)
if !el
	return False
endIf
if getWindowClass(hwnd) != "MicrosoftSilverlight"
&& el.name != "Silverlight Control"
	return False
elif el.hasKeyboardFocus
	; All is well already.
	return False
endIf
setFocus(hwnd)
return True
endFunction

object function roomtree__UIA__selection(object el, optional string flags)
; Return the selected item if there is just one, or optionally an array of them all.
; Flags:
;	a: Return an array of all selections.
;		Otherwise, return the single selection or null.
;	f: Only allow fast means of finding selections.
;		Otherwise, findAll is allowed for this.
;		Warning: f will always cause null returns on JAWS 14.
flags = stringLower(flags)
var int returnArray = stringContains(flags, "a")
var int requireFast = stringContains(flags, "f")
var object els
var object selpat = el.getSelectionPattern()
; Fails in JAWS 14 and maybe early 15, and when the pattern is not supported.
if selpat
	els = selpat.getSelection()
elif requireFast
	return
else
	; All we have left is findAll.
	; Note that this function regards findFirst as unsafe; no way to know if there are several.
	; 4 is TreeScope_Descendants.
	els = roomtree__UIA__findAll(el, 4, "selectionItemIsSelected True")
endIf

if returnArray
	return els
elif roomtree__UIA__elementCount(els) != 1
	return
endIf
return roomtree__UIA__elementByIndex(els, 0)
endFunction

object function roomtree__UIA__elementFromWindow(handle hwnd)
; Return the UIA element rooted at the given window or null on error.
var object oUIA = roomtree__UIA__getScriptAPI()
if !oUIA
	return
endIf
return oUIA.getElementFromHandle(hwnd)
endFunction

object function roomtree__UIA__elementFromMSAA(object o, int childID)
; Return the UIA element corresponding to the given MSAA object and childID, or null on error.
var object oUIA = roomtree__UIA__getScriptAPI()
if !oUIA
	return
endIf
return oUIA.getElementFromIAccessible(o, childID)
endFunction

variant function roomtree__UIA__boundingRectangle(object el, optional int adjustForJAWS)
; Return an object or collection with left/right/top/bottom properties for the given element.
; In JAWS 15.0.5056 and later, an object is returned.
; Otherwise a collection simulating that object is returned.
; If adjustForJAWS is passed and True, 1 is added to each coordinate,
; and a collection is always returned.
; Warning: When an object is returned, its properties are read-only.
var variant rect, object rect0
if roomtree__UIA__version() >= 3
	if !adjustForJAWS
		return el.boundingRectangle()
	endIf
	rect0 = el.boundingRectangle()
	rect = new collection
	rect.left = rect0.left +1
	rect.right = rect0.right +1
	rect.top = rect0.top +1
	rect.bottom = rect0.bottom +1
	return rect
endIf

rect = new collection
var int left, int right, int top, int bottom
if !el.boundingRectangle(intRef(left), intRef(top), intRef(right), intRef(bottom))
	return
endIf
if adjustForJAWS
rect.left = left +1
rect.right = right +1
rect.top = top +1
rect.bottom = bottom +1
else
rect.left = left
rect.right = right
rect.top = top
rect.bottom = bottom
endIf
return rect
endFunction

collection function roomtree__UIA__findAllIterator(object el, int treeScope, variant cond, optional
	variant v1, variant v2, variant v3, variant v4, variant v5,
	variant v6, variant v7, variant v8, variant v9, variant v10
)
; Return an iterator over matches to the given condition.
; The condition may be an object as expected by oUIA.FindAll(),
; or a string as expected by roomtree__UIA__condition().
; Useful when supporting JAWS 14.
return roomtree__UIA__elementIterator(roomtree__UIA__findAll(el, treeScope, cond, v1, v2, v3, v4, v5, v6, v7, v8, v9, v10))
endFunction

object function roomtree__UIA__findAll(object el, int treeScope, variant cond, optional
	variant v1, variant v2, variant v3, variant v4, variant v5,
	variant v6, variant v7, variant v8, variant v9, variant v10
)
; Return an array of matches to the given condition.
; The condition may be an object as expected by oUIA.FindAll(),
; or a string as expected by roomtree__UIA__condition().
if getVariantType(cond) == 8  ; VT_String
	cond = roomtree__UIA__condition(cond, v1, v2, v3, v4, v5, v6, v7, v8, v9, v10)
	if !cond
		return
	endIf
endIf
return el.findAll(treeScope, cond)
endFunction

int function roomtree__UIA__elementCount(object els)
; Return the number of elements in els, JAWS 14 or 15+.
return els.length +els.count  ; length before JAWS 15, count thereafter.
endFunction

object function roomtree__UIA__elementByIndex(object els, int idx)
; Return the idxth element, JAWS 14 or 15.
; idx is 0-based.
if roomtree__UIA__version() >= 3
	return els(idx)
else
	return els.getElement(idx)
endIf
endFunction

collection function roomtree__UIA__elementIterator(object els)
; Return an iterator for roomtree__UIA__nextElement.
; Includes .count for the element count, for convenience.
; At any instant, .index is the 1-based index of iteration.
; .elements is the actual set of elements.
; .currentElement is the current element during iteration.
; ._version is a (private) indicator of UIAScriptAPI version.
; Useful when supporting JAWS 14.
var collection iter iter = new collection
var variant null
iter.elements = els
iter.count = els.count +els.length
roomtree__UIA__resetIterator(iter)
iter._version = roomtree__UIA__version()
return iter
endFunction

void function roomtree__UIA__resetIterator(collection iter)
; Reset the iterator to start at the beginning.
iter.index = 0
var variant null
iter.currentElement = null
endFunction

int function roomtree__UIA__nextElement(collection iter)
; Moves to the next element in iter, as returned by roomtree__UIA__elementIterator().
; Returns True if there is another element and False otherwise.
; .currentElement becomes the next element.
var variant null
iter.index = iter.index +1
if iter.index > iter.count
	iter.currentElement = null
	return False
endIf
var object els = iter.elements
var int idx = iter.index -1
if iter._version >= 3
	iter.currentElement = els(idx)
else
	iter.currentElement = els.getElement(idx)
endIf
return True
endFunction

; Cache of conditions by spec, excluding those with %<n> parameters.
globals collection roomtree__UIA___conds

object function roomtree__UIA__condition(string spec, optional
	variant v1, variant v2, variant v3, variant v4, variant v5,
	variant v6, variant v7, variant v8, variant v9, variant v10
)
; Build and return a UIAScriptAPI Condition object from a string specification.
; Values and other Condition objects may be passed and referred to in the spec.
; The spec is in post-fix notation, case-insensitive.
; Items in the spec are separated by a single space or line ending.  Allowed items:
;	contentView, controlView, rawView: Require the given UIA tree view type.
;	focusedProcess: Require the process ID to match the focused application's process ID.
;	FPRaw/Control/ContentView: One of the above with focusedProcess.
;	not: Reverse the previous condition: "focusedProcess not."
;	and, or: Combine previous two conditions: "contentView focusedProcess and."
;	property name: The name of a property with a PropertyID.
;		These are from UIA.jsh without UIA_ prefix and PropertyId suffix.
;		Examples: Name, ControlType.
;	value name: The name of a value with an ID. Categories:
;		Control types like CheckBox, Edit, and Hyperlink.
;		Other constants without *_ prefix, like Horizontal and LargeIncrement.
;		Other raw constants like Next that contain no underscore.
;	False, True: Create a boolPropertyCondition with the previous item.
;		Example: "IsEnabled True."
;		If not after a property, creates a False or True condition
;		(directly in JAWS 15.0.5056+, simulated before that).
;	int like 1, -50, etc.: An integer value.
;		Makes an IntPropertyCondition with the previous item.
;		Example: "RangeValueMinimum -5"
;	String like 'hello': A string value.
;		Makes a StringPropertyCondition with the previous item.
;		Example: "Name 'My UIA Application'"
;		Apostrophes in the string are written as ''. Backslash escapes like \t also work.
;	%n (1-10): Use argument n, 1 being the first after the spec, etc.
;	May be a string or a Condition object. Ints are not supported here.
; Examples:
;	c1 = roomtree__UIA__condition("ControlType GroupControl Name %1 and", getObjectName(True, 1))
;		(Must be a group control named whatever getObjectName(True, 1) returns.)
;	c2 = roomtree__UIA__condition("ContentView %1 and", c1)
;		Same as c1 but must also be in the ContentView tree.
; Returns an IDispatch Condition object for further use with the UIAScriptAPI.
; Returns null on error and may speak a message explaining the problem.
if !roomtree__UIA___conds
	roomtree__UIA___conds = new collection
endIf
var object cnd, variant null
spec = stringReplaceChars(spec, "\10\13\160", " ")
spec = stringReplaceSubstrings(spec, "  ", " ")
if !stringContains(spec, "%")
	cnd = roomtree__UIA___conds[spec]
	if cnd
		return cnd
	endIf
endIf
var string spec0 = spec
var object oUIA = roomtree__UIA__getScriptAPI()
if !oUIA
	roomtree__UIA___sayError(roomtree__UIA___msgNoUIAObject)
	return
endIf
spec = stringTrimLeadingBlanks(stringTrimTrailingBlanks(spec))
if !spec
	roomtree__UIA___sayError(roomtree__UIA___msgNullConditionString)
	return
endIf
; Shortcuts allowed starting in JAWS 15.0.5056.
if roomtree__UIA__version() >= 3
	cnd = null
	if stringCompare(spec, "rawview", False) == 0
		cnd = oUIA.rawViewCondition
	elif stringCompare(spec, "controlview", False) == 0
		cnd = oUIA.controlViewCondition
	elif stringCompare(spec, "contentview", False) == 0
		cnd = oUIA.contentViewCondition
	endIf
	if cnd
		roomtree__UIA___conds[spec0] = cnd
		return cnd
	endIf
endIf
var variant stack = roomtree__UIA__stl__stack(stringLength(spec))
var string sItem, int pos
var object c1, object c2
roomtree__UIA___buildLists()
; This is used to disambiguate %1 etc. as condition versus value.
var int expectValue = False
while spec
	if stringLeft(spec, 1) == "'"
		; A string literal that may contain '' apostrophe escapes.
		; JAWS will translate the other escapes like \t, \n, etc.
		expectValue = False
		spec = stringChopLeft(spec, 1)
		sItem = ""
		pos = stringContains(spec, "'")
		if !pos
			roomtree__UIA___sayError(roomtree__UIA___msgStringNotTerminated, spec0)
			return
		endIf
		while substring(spec, pos+1, 1) == "'"
			; Include the first of the two apostrophes.
			sItem = formatString("%1%2", sItem, stringLeft(spec, pos))
			; Then chop both off the spec.
			spec = stringChopLeft(spec, pos+1)
			; And finally find the next one.
			pos = stringContains(spec, "'")
			if !pos
				roomtree__UIA___sayError(roomtree__UIA___msgStringNotTerminated, spec0)
				return
			endIf
		endWhile
		sItem = formatString("%1%2", sItem, stringLeft(spec, pos-1))
		spec = stringChopLeft(spec, pos)
		roomtree__UIA__stl__push(stack, oUIA.CreateStringPropertyCondition(
			roomtree__UIA__stl__pop(stack).value, sItem
		))
	else
		pos = stringContains(spec, " ")
		if pos
			sItem = stringLeft(spec, pos-1)
			spec = stringChopLeft(spec, pos)
		else
			sItem = spec
			spec = ""
		endIf
		; Some shortcuts for developer (and slight runtime) efficiency.
		if stringStartsWith(sItem, "fp") && stringContains("fprawview fpcontrolview fpcontentview", stringLower(sItem))
			var object c0
			if sItem == "FPRawView"
				c0 = oUIA.CreateRawViewCondition()
			elif sItem == "FPControlView"
				c0 = oUIA.CreateControlViewCondition()
			else  ; FPContentView
				c0 = oUIA.CreateContentViewCondition()
			endIf
			c0 = oUIA.CreateAndCondition(c0,
				oUIA.CreateIntPropertyCondition(roomtree__UIA__propertyID("processID"),
				oUIA.GetFocusedElement().ProcessID)
			)
			roomtree__UIA__stl__push(stack, c0)
			expectValue = False
		elif stringCompare(sItem, "RawView", False) == 0
			roomtree__UIA__stl__push(stack, oUIA.CreateRawViewCondition())
			expectValue = False
		elif stringCompare(sItem, "ControlView", False) == 0
			roomtree__UIA__stl__push(stack, oUIA.CreateControlViewCondition())
			expectValue = False
		elif stringCompare(sItem, "ContentView", False) == 0
			roomtree__UIA__stl__push(stack, oUIA.CreateContentViewCondition())
			expectValue = False
		elif stringCompare(sItem, "FocusedProcess", False) == 0
			roomtree__UIA__stl__push(stack,
				oUIA.CreateIntPropertyCondition(roomtree__UIA__propertyID("processID"),
				oUIA.GetFocusedElement().ProcessID)
			)
			expectValue = False
		elif stringCompare(sItem, "True", False) == 0
			roomtree__UIA___handleTrueAndFalse(stack, oUIA, True)
			expectValue = False
		elif stringCompare(sItem, "False", False) == 0
			roomtree__UIA___handleTrueAndFalse(stack, oUIA, False)
			expectValue = False
		elif stringCompare(sItem, "not", False) == 0
			roomtree__UIA__stl__push(stack,
				oUIA.CreateNotCondition(roomtree__UIA__stl__pop(stack).value)
			)
			expectValue = False
		elif stringCompare(sItem, "and", False) == 0
			c2 = roomtree__UIA__stl__pop(stack).value
			c1 = roomtree__UIA__stl__pop(stack).value
			roomtree__UIA__stl__push(stack, oUIA.CreateAndCondition(c1, c2))
			expectValue = False
		elif stringCompare(sItem, "or", False) == 0
			c2 = roomtree__UIA__stl__pop(stack).value
			c1 = roomtree__UIA__stl__pop(stack).value
			roomtree__UIA__stl__push(stack, oUIA.CreateOrCondition(c1, c2))
			expectValue = False
		elif stringToInt(sItem) || sItem == "0" || sItem == "-0"
				roomtree__UIA__stl__push(stack,
					oUIA.CreateIntPropertyCondition(roomtree__UIA__stl__pop(stack).value, stringToInt(sItem))
				)
			expectValue = False
		elif stringLeft(sItem, 1) == "%"
			var int idx = stringToInt(stringChopLeft(sItem, 1))
			var variant v
			if idx == 1 v = v1
			elif idx == 2 v = v2
			elif idx == 3 v = v3
			elif idx == 4 v = v4
			elif idx == 5 v = v5
			elif idx == 6 v = v6
			elif idx == 7 v = v7
			elif idx == 8 v = v8
			elif idx == 9 v = v9
			elif idx == 10 v = v10
			else
				roomtree__UIA___sayError(formatString(roomtree__UIA___msgInvalidRef, sItem), spec0)
				return
			endIf
			if expectValue
				; ToDo: Assuming references like this are strings.
				roomtree__UIA__stl__push(stack, oUIA.CreateStringPropertyCondition(
					roomtree__UIA__stl__pop(stack).value, v
				))
			else
				roomtree__UIA__stl__push(stack, v)
			endIf
			expectValue = False
		else
			; All that's left are property and value names.
			var int val
			if collectionItemExists(roomtree__UIA___props, sItem)
				; Save for the value to compare it to.
				val = roomtree__UIA__propertyID(sItem)
				roomtree__UIA__stl__push(stack, val)
				expectValue = True
			else
				val = roomtree__UIA__valueID(sItem)
				if val == roomtree__UIA__Error
					return
				endIf
				; Now we have one of those values.
				; All of these should be ints.
				roomtree__UIA__stl__push(stack, oUIA.CreateIntPropertyCondition(
					roomtree__UIA__stl__pop(stack).value, val
				))
				expectValue = False
			endIf  ; Property name or value name.
		endIf  ; list of possible keywords.
	endIf  ; leading ' or no leading '.
	spec = stringTrimLeadingBlanks(spec)
endWhile
if roomtree__UIA__stl__size(stack) != 1
	roomtree__UIA___sayError(formatString(roomtree__UIA___msgSpecIncomplete, intToString(roomtree__UIA__stl__size(stack)-1)), spec0)
	return
endIf
var object condition = roomtree__UIA__stl__pop(stack).value
if !condition
	roomtree__UIA___sayError(roomtree__UIA___msgUnknownError, spec0)
endIf
roomtree__UIA___conds[spec0] = condition
return condition
endFunction

void function roomtree__UIA___handleTrueAndFalse(variant stack, object oUIA, int val)
; Handle "false" and "true" in specs.
; Helper for roomtree__UIA__condition().
var int standAlone = False
var variant prevVal
var int size = roomtree__UIA__stl__size(stack)
if size < 1
	standAlone = True
else
	prevVal = roomtree__UIA__stl__pop(stack).value
	if getVariantType(prevVal) == 9  ; VT_Dispatch
		; This should be a Condition object.
		standAlone = True
	endIf
endIf
if standAlone
	var object newVal
	if val
		if roomtree__UIA__version() >= 3
			newVal = oUIA.CreateTrueCondition()
		else
			; Something that should always be True.
			newVal = oUIA.CreateRawViewCondition()
		endIf
	else
		if roomtree__UIA__version() >= 3
			newVal = oUIA.CreateFalseCondition()
		else
			; Something that should always be False.
			; This is controlType = <invalidValue>.
			newVal = oUIA.CreateIntPropertyCondition(30003, 0xFFF00000L)
		endIf
	endIf
	roomtree__UIA__stl__push(stack, newVal)
	return
endIf
roomtree__UIA__stl__push(stack,
	oUIA.CreateBoolPropertyCondition(prevVal, val)
)
endFunction

variant function roomtree__UIA__propertyValue(object el, variant id)
; Get a UIA property value by its ID or name.
; Returns null and speaks an error message if an invalid property name is passed.
if !stringToInt(id)
	id = roomtree__UIA__propertyId(id)
	if id == roomtree__UIA__Error
		return
	endIf
endIf
return el.getPropertyValue(id)
endFunction

int function roomtree__UIA__propertyID(string name, optional int quiet)
; Get a UIA property ID.
; Returns the constant roomtree__UIA__Error if an invalid property name is passed.
; Also speaks an error message if that happens unless quiet is True.
return roomtree__UIA___getID("property", name, quiet)
endFunction

int function roomtree__UIA__valueID(string name)
; Get a UIA value ID.
; Returns the constant roomtree__UIA__Error if an invalid value name is passed.
; Also speaks an error message if that happens.
return roomtree__UIA___getID("value", name)
endFunction

int function roomtree__UIA__eventID(string name)
; Get a UIA event ID.
; Returns the constant roomtree__UIA__Error if an invalid event name is passed.
; Also speaks an error message if that happens.
return roomtree__UIA___getID("event", name)
endFunction

int function roomtree__UIA__attributeID(string name)
; Get a UIA attribute ID.
; Returns the constant roomtree__UIA__Error if an invalid attribute name is passed.
; Also speaks an error message if that happens.
return roomtree__UIA___getID("attribute", name)
endFunction

int function roomtree__UIA__textUnitID(string name)
; Get a UIA text unit id.
; Returns the constant roomtree__UIA__Error if an invalid text unit name is passed.
; Also speaks an error message if that happens.
return roomtree__UIA___getID("textUnit", name)
endFunction

collection function roomtree__UIA__getList(string listType)
; Get a list by a string identifier.
; The list is returned as a collection, name to value.
roomtree__UIA___buildLists()
var collection lst
if listType == "property"
	lst = roomtree__UIA___props
elif listType == "value"
	lst = roomtree__UIA___vals
elif listType == "event"
	lst = roomtree__UIA___events
elif listType == "attribute"
	lst = roomtree__UIA___attributes
elif listType == "textUnit"
	lst = roomtree__UIA___textUnits
endIf
return lst
endFunction

int function roomtree__UIA___getID(string listType, string name, optional int quiet)
; Get a property, value, event, attribute, or text unit id.
; Returns the constant roomtree__UIA__Error if an invalid name is passed.
; Also speaks an error message if that happens unless quiet is True.
var int result = roomtree__UIA___getID0(listType, name, quiet)
if result != roomtree__UIA__Error
	return result
endIf
; Rebuild lists in case they were updated since the last JAWS restart.
; (This implies a script update and reload.)
roomtree__UIA___buildLists(True)
result = roomtree__UIA___getID0(listType, name, quiet)
if result != roomtree__UIA__Error
	return result
endIf
if !quiet
	roomtree__UIA___sayError(formatString(roomtree__UIA___msgInvalidName, listType, name))
endIf
return roomtree__UIA__Error
endFunction

int function roomtree__UIA___getID0(string listType, string name, optional int quiet)
; Helper for roomtree__UIA___getID().
var collection lst = roomtree__UIA__getList(listType)
if stringLeft(name, 4) == "UIA_"
	name = stringChopLeft(name, 4)
endIf
if collectionItemExists(lst, name)
	return lst[name]
elif collectionItemExists(roomtree__UIA___fullnames, name)
	return roomtree__UIA___fullnames[name]
endIf
return roomtree__UIA__Error
endFunction

void function roomtree__UIA___UIAClear()
; Release the UIAScriptAPI cached object.
var variant null
roomtree__UIA___gUIA = null
endFunction

void function roomtree__UIA___buildLists(optional int force)
; Build internal lists of property, value, event, attribute, and textUnit Ids.
if roomtree__UIA___built && !force
	return
endIf
roomtree__UIA___built = False
roomtree__UIA___props = new collection
roomtree__UIA___vals = new collection
roomtree__UIA___events = new collection
roomtree__UIA___attributes = new collection
roomtree__UIA___textUnits = new collection
roomtree__UIA___fullnames = new collection
roomtree__UIA___fullnames["UIATrue"] = 0xffffffff
roomtree__UIA___vals["UIATrue"] = 0xffffffff
roomtree__UIA___fullnames["UIAError"] = 0xfffe
roomtree__UIA___vals["UIAError"] = 0xfffe
roomtree__UIA___fullnames["ButtonControlTypeId"] = 50000
roomtree__UIA___vals["Button"] = 50000
roomtree__UIA___fullnames["CalendarControlTypeId"] = 50001
roomtree__UIA___vals["Calendar"] = 50001
roomtree__UIA___fullnames["CheckBoxControlTypeId"] = 50002
roomtree__UIA___vals["CheckBox"] = 50002
roomtree__UIA___fullnames["ComboBoxControlTypeId"] = 50003
roomtree__UIA___vals["ComboBox"] = 50003
roomtree__UIA___fullnames["EditControlTypeId"] = 50004
roomtree__UIA___vals["Edit"] = 50004
roomtree__UIA___fullnames["HyperlinkControlTypeId"] = 50005
roomtree__UIA___vals["Hyperlink"] = 50005
roomtree__UIA___fullnames["ImageControlTypeId"] = 50006
roomtree__UIA___vals["Image"] = 50006
roomtree__UIA___fullnames["ListItemControlTypeId"] = 50007
roomtree__UIA___vals["ListItem"] = 50007
roomtree__UIA___fullnames["ListControlTypeId"] = 50008
roomtree__UIA___vals["List"] = 50008
roomtree__UIA___fullnames["MenuControlTypeId"] = 50009
roomtree__UIA___vals["Menu"] = 50009
roomtree__UIA___fullnames["MenuBarControlTypeId"] = 50010
roomtree__UIA___vals["MenuBar"] = 50010
roomtree__UIA___fullnames["MenuItemControlTypeId"] = 50011
roomtree__UIA___vals["MenuItem"] = 50011
roomtree__UIA___fullnames["ProgressBarControlTypeId"] = 50012
roomtree__UIA___vals["ProgressBar"] = 50012
roomtree__UIA___fullnames["RadioButtonControlTypeId"] = 50013
roomtree__UIA___vals["RadioButton"] = 50013
roomtree__UIA___fullnames["ScrollBarControlTypeId"] = 50014
roomtree__UIA___vals["ScrollBar"] = 50014
roomtree__UIA___fullnames["SliderControlTypeId"] = 50015
roomtree__UIA___vals["Slider"] = 50015
roomtree__UIA___fullnames["SpinnerControlTypeId"] = 50016
roomtree__UIA___vals["Spinner"] = 50016
roomtree__UIA___fullnames["StatusBarControlTypeId"] = 50017
roomtree__UIA___vals["StatusBar"] = 50017
roomtree__UIA___fullnames["TabControlTypeId"] = 50018
roomtree__UIA___vals["Tab"] = 50018
roomtree__UIA___fullnames["TabItemControlTypeId"] = 50019
roomtree__UIA___vals["TabItem"] = 50019
roomtree__UIA___fullnames["TextControlTypeId"] = 50020
roomtree__UIA___vals["Text"] = 50020
roomtree__UIA___fullnames["ToolBarControlTypeId"] = 50021
roomtree__UIA___vals["ToolBar"] = 50021
roomtree__UIA___fullnames["ToolTipControlTypeId"] = 50022
roomtree__UIA___vals["ToolTip"] = 50022
roomtree__UIA___fullnames["TreeControlTypeId"] = 50023
roomtree__UIA___vals["Tree"] = 50023
roomtree__UIA___fullnames["TreeItemControlTypeId"] = 50024
roomtree__UIA___vals["TreeItem"] = 50024
roomtree__UIA___fullnames["CustomControlTypeId"] = 50025
roomtree__UIA___vals["Custom"] = 50025
roomtree__UIA___fullnames["GroupControlTypeId"] = 50026
roomtree__UIA___vals["Group"] = 50026
roomtree__UIA___fullnames["ThumbControlTypeId"] = 50027
roomtree__UIA___vals["Thumb"] = 50027
roomtree__UIA___fullnames["DataGridControlTypeId"] = 50028
roomtree__UIA___vals["DataGrid"] = 50028
roomtree__UIA___fullnames["DataItemControlTypeId"] = 50029
roomtree__UIA___vals["DataItem"] = 50029
roomtree__UIA___fullnames["DocumentControlTypeId"] = 50030
roomtree__UIA___vals["Document"] = 50030
roomtree__UIA___fullnames["SplitButtonControlTypeId"] = 50031
roomtree__UIA___vals["SplitButton"] = 50031
roomtree__UIA___fullnames["WindowControlTypeId"] = 50032
roomtree__UIA___vals["Window"] = 50032
roomtree__UIA___fullnames["PaneControlTypeId"] = 50033
roomtree__UIA___vals["Pane"] = 50033
roomtree__UIA___fullnames["HeaderControlTypeId"] = 50034
roomtree__UIA___vals["Header"] = 50034
roomtree__UIA___fullnames["HeaderItemControlTypeId"] = 50035
roomtree__UIA___vals["HeaderItem"] = 50035
roomtree__UIA___fullnames["TableControlTypeId"] = 50036
roomtree__UIA___vals["Table"] = 50036
roomtree__UIA___fullnames["TitleBarControlTypeId"] = 50037
roomtree__UIA___vals["TitleBar"] = 50037
roomtree__UIA___fullnames["SeparatorControlTypeId"] = 50038
roomtree__UIA___vals["Separator"] = 50038
roomtree__UIA___fullnames["SemanticZoomControlTypeId"] = 50039
roomtree__UIA___vals["SemanticZoom"] = 50039
roomtree__UIA___fullnames["AppBarControlTypeId"] = 50040
roomtree__UIA___vals["AppBar"] = 50040
roomtree__UIA___fullnames["RuntimeIdPropertyId"] = 30000
roomtree__UIA___props["RuntimeId"] = 30000
roomtree__UIA___fullnames["BoundingRectanglePropertyId"] = 30001
roomtree__UIA___props["BoundingRectangle"] = 30001
roomtree__UIA___fullnames["ProcessIdPropertyId"] = 30002
roomtree__UIA___props["ProcessId"] = 30002
roomtree__UIA___fullnames["ControlTypePropertyId"] = 30003
roomtree__UIA___props["ControlType"] = 30003
roomtree__UIA___fullnames["LocalizedControlTypePropertyId"] = 30004
roomtree__UIA___props["LocalizedControlType"] = 30004
roomtree__UIA___fullnames["NamePropertyId"] = 30005
roomtree__UIA___props["Name"] = 30005
roomtree__UIA___fullnames["AcceleratorKeyPropertyId"] = 30006
roomtree__UIA___props["AcceleratorKey"] = 30006
roomtree__UIA___fullnames["AccessKeyPropertyId"] = 30007
roomtree__UIA___props["AccessKey"] = 30007
roomtree__UIA___fullnames["HasKeyboardFocusPropertyId"] = 30008
roomtree__UIA___props["HasKeyboardFocus"] = 30008
roomtree__UIA___fullnames["IsKeyboardFocusablePropertyId"] = 30009
roomtree__UIA___props["IsKeyboardFocusable"] = 30009
roomtree__UIA___fullnames["IsEnabledPropertyId"] = 30010
roomtree__UIA___props["IsEnabled"] = 30010
roomtree__UIA___fullnames["AutomationIdPropertyId"] = 30011
roomtree__UIA___props["AutomationId"] = 30011
roomtree__UIA___fullnames["ClassNamePropertyId"] = 30012
roomtree__UIA___props["ClassName"] = 30012
roomtree__UIA___fullnames["HelpTextPropertyId"] = 30013
roomtree__UIA___props["HelpText"] = 30013
roomtree__UIA___fullnames["ClickablePointPropertyId"] = 30014
roomtree__UIA___props["ClickablePoint"] = 30014
roomtree__UIA___fullnames["CulturePropertyId"] = 30015
roomtree__UIA___props["Culture"] = 30015
roomtree__UIA___fullnames["IsControlElementPropertyId"] = 30016
roomtree__UIA___props["IsControlElement"] = 30016
roomtree__UIA___fullnames["IsContentElementPropertyId"] = 30017
roomtree__UIA___props["IsContentElement"] = 30017
roomtree__UIA___fullnames["LabeledByPropertyId"] = 30018
roomtree__UIA___props["LabeledBy"] = 30018
roomtree__UIA___fullnames["IsPasswordPropertyId"] = 30019
roomtree__UIA___props["IsPassword"] = 30019
roomtree__UIA___fullnames["NativeWindowHandlePropertyId"] = 30020
roomtree__UIA___props["NativeWindowHandle"] = 30020
roomtree__UIA___fullnames["ItemTypePropertyId"] = 30021
roomtree__UIA___props["ItemType"] = 30021
roomtree__UIA___fullnames["IsOffscreenPropertyId"] = 30022
roomtree__UIA___props["IsOffscreen"] = 30022
roomtree__UIA___fullnames["OrientationPropertyId"] = 30023
roomtree__UIA___props["Orientation"] = 30023
roomtree__UIA___fullnames["FrameworkIdPropertyId"] = 30024
roomtree__UIA___props["FrameworkId"] = 30024
roomtree__UIA___fullnames["IsRequiredForFormPropertyId"] = 30025
roomtree__UIA___props["IsRequiredForForm"] = 30025
roomtree__UIA___fullnames["ItemStatusPropertyId"] = 30026
roomtree__UIA___props["ItemStatus"] = 30026
roomtree__UIA___fullnames["IsDockPatternAvailablePropertyId"] = 30027
roomtree__UIA___props["IsDockPatternAvailable"] = 30027
roomtree__UIA___fullnames["IsExpandCollapsePatternAvailablePropertyId"] = 30028
roomtree__UIA___props["IsExpandCollapsePatternAvailable"] = 30028
roomtree__UIA___fullnames["IsGridItemPatternAvailablePropertyId"] = 30029
roomtree__UIA___props["IsGridItemPatternAvailable"] = 30029
roomtree__UIA___fullnames["IsGridPatternAvailablePropertyId"] = 30030
roomtree__UIA___props["IsGridPatternAvailable"] = 30030
roomtree__UIA___fullnames["IsInvokePatternAvailablePropertyId"] = 30031
roomtree__UIA___props["IsInvokePatternAvailable"] = 30031
roomtree__UIA___fullnames["IsMultipleViewPatternAvailablePropertyId"] = 30032
roomtree__UIA___props["IsMultipleViewPatternAvailable"] = 30032
roomtree__UIA___fullnames["IsRangeValuePatternAvailablePropertyId"] = 30033
roomtree__UIA___props["IsRangeValuePatternAvailable"] = 30033
roomtree__UIA___fullnames["IsScrollPatternAvailablePropertyId"] = 30034
roomtree__UIA___props["IsScrollPatternAvailable"] = 30034
roomtree__UIA___fullnames["IsScrollItemPatternAvailablePropertyId"] = 30035
roomtree__UIA___props["IsScrollItemPatternAvailable"] = 30035
roomtree__UIA___fullnames["IsSelectionItemPatternAvailablePropertyId"] = 30036
roomtree__UIA___props["IsSelectionItemPatternAvailable"] = 30036
roomtree__UIA___fullnames["IsSelectionPatternAvailablePropertyId"] = 30037
roomtree__UIA___props["IsSelectionPatternAvailable"] = 30037
roomtree__UIA___fullnames["IsTablePatternAvailablePropertyId"] = 30038
roomtree__UIA___props["IsTablePatternAvailable"] = 30038
roomtree__UIA___fullnames["IsTableItemPatternAvailablePropertyId"] = 30039
roomtree__UIA___props["IsTableItemPatternAvailable"] = 30039
roomtree__UIA___fullnames["IsTextPatternAvailablePropertyId"] = 30040
roomtree__UIA___props["IsTextPatternAvailable"] = 30040
roomtree__UIA___fullnames["IsTogglePatternAvailablePropertyId"] = 30041
roomtree__UIA___props["IsTogglePatternAvailable"] = 30041
roomtree__UIA___fullnames["IsTransformPatternAvailablePropertyId"] = 30042
roomtree__UIA___props["IsTransformPatternAvailable"] = 30042
roomtree__UIA___fullnames["IsValuePatternAvailablePropertyId"] = 30043
roomtree__UIA___props["IsValuePatternAvailable"] = 30043
roomtree__UIA___fullnames["IsWindowPatternAvailablePropertyId"] = 30044
roomtree__UIA___props["IsWindowPatternAvailable"] = 30044
roomtree__UIA___fullnames["ValueValuePropertyId"] = 30045
roomtree__UIA___props["ValueValue"] = 30045
roomtree__UIA___fullnames["ValueIsReadOnlyPropertyId"] = 30046
roomtree__UIA___props["ValueIsReadOnly"] = 30046
roomtree__UIA___fullnames["RangeValueValuePropertyId"] = 30047
roomtree__UIA___props["RangeValueValue"] = 30047
roomtree__UIA___fullnames["RangeValueIsReadOnlyPropertyId"] = 30048
roomtree__UIA___props["RangeValueIsReadOnly"] = 30048
roomtree__UIA___fullnames["RangeValueMinimumPropertyId"] = 30049
roomtree__UIA___props["RangeValueMinimum"] = 30049
roomtree__UIA___fullnames["RangeValueMaximumPropertyId"] = 30050
roomtree__UIA___props["RangeValueMaximum"] = 30050
roomtree__UIA___fullnames["RangeValueLargeChangePropertyId"] = 30051
roomtree__UIA___props["RangeValueLargeChange"] = 30051
roomtree__UIA___fullnames["RangeValueSmallChangePropertyId"] = 30052
roomtree__UIA___props["RangeValueSmallChange"] = 30052
roomtree__UIA___fullnames["ScrollHorizontalScrollPercentPropertyId"] = 30053
roomtree__UIA___props["ScrollHorizontalScrollPercent"] = 30053
roomtree__UIA___fullnames["ScrollHorizontalViewSizePropertyId"] = 30054
roomtree__UIA___props["ScrollHorizontalViewSize"] = 30054
roomtree__UIA___fullnames["ScrollVerticalScrollPercentPropertyId"] = 30055
roomtree__UIA___props["ScrollVerticalScrollPercent"] = 30055
roomtree__UIA___fullnames["ScrollVerticalViewSizePropertyId"] = 30056
roomtree__UIA___props["ScrollVerticalViewSize"] = 30056
roomtree__UIA___fullnames["ScrollHorizontallyScrollablePropertyId"] = 30057
roomtree__UIA___props["ScrollHorizontallyScrollable"] = 30057
roomtree__UIA___fullnames["ScrollVerticallyScrollablePropertyId"] = 30058
roomtree__UIA___props["ScrollVerticallyScrollable"] = 30058
roomtree__UIA___fullnames["SelectionSelectionPropertyId"] = 30059
roomtree__UIA___props["SelectionSelection"] = 30059
roomtree__UIA___fullnames["SelectionCanSelectMultiplePropertyId"] = 30060
roomtree__UIA___props["SelectionCanSelectMultiple"] = 30060
roomtree__UIA___fullnames["SelectionIsSelectionRequiredPropertyId"] = 30061
roomtree__UIA___props["SelectionIsSelectionRequired"] = 30061
roomtree__UIA___fullnames["GridRowCountPropertyId"] = 30062
roomtree__UIA___props["GridRowCount"] = 30062
roomtree__UIA___fullnames["GridColumnCountPropertyId"] = 30063
roomtree__UIA___props["GridColumnCount"] = 30063
roomtree__UIA___fullnames["GridItemRowPropertyId"] = 30064
roomtree__UIA___props["GridItemRow"] = 30064
roomtree__UIA___fullnames["GridItemColumnPropertyId"] = 30065
roomtree__UIA___props["GridItemColumn"] = 30065
roomtree__UIA___fullnames["GridItemRowSpanPropertyId"] = 30066
roomtree__UIA___props["GridItemRowSpan"] = 30066
roomtree__UIA___fullnames["GridItemColumnSpanPropertyId"] = 30067
roomtree__UIA___props["GridItemColumnSpan"] = 30067
roomtree__UIA___fullnames["GridItemContainingGridPropertyId"] = 30068
roomtree__UIA___props["GridItemContainingGrid"] = 30068
roomtree__UIA___fullnames["DockDockPositionPropertyId"] = 30069
roomtree__UIA___props["DockDockPosition"] = 30069
roomtree__UIA___fullnames["ExpandCollapseExpandCollapseStatePropertyId"] = 30070
roomtree__UIA___props["ExpandCollapseExpandCollapseState"] = 30070
roomtree__UIA___fullnames["MultipleViewCurrentViewPropertyId"] = 30071
roomtree__UIA___props["MultipleViewCurrentView"] = 30071
roomtree__UIA___fullnames["MultipleViewSupportedViewsPropertyId"] = 30072
roomtree__UIA___props["MultipleViewSupportedViews"] = 30072
roomtree__UIA___fullnames["WindowCanMaximizePropertyId"] = 30073
roomtree__UIA___props["WindowCanMaximize"] = 30073
roomtree__UIA___fullnames["WindowCanMinimizePropertyId"] = 30074
roomtree__UIA___props["WindowCanMinimize"] = 30074
roomtree__UIA___fullnames["WindowWindowVisualStatePropertyId"] = 30075
roomtree__UIA___props["WindowWindowVisualState"] = 30075
roomtree__UIA___fullnames["WindowWindowInteractionStatePropertyId"] = 30076
roomtree__UIA___props["WindowWindowInteractionState"] = 30076
roomtree__UIA___fullnames["WindowIsModalPropertyId"] = 30077
roomtree__UIA___props["WindowIsModal"] = 30077
roomtree__UIA___fullnames["WindowIsTopmostPropertyId"] = 30078
roomtree__UIA___props["WindowIsTopmost"] = 30078
roomtree__UIA___fullnames["SelectionItemIsSelectedPropertyId"] = 30079
roomtree__UIA___props["SelectionItemIsSelected"] = 30079
roomtree__UIA___fullnames["SelectionItemSelectionContainerPropertyId"] = 30080
roomtree__UIA___props["SelectionItemSelectionContainer"] = 30080
roomtree__UIA___fullnames["TableRowHeadersPropertyId"] = 30081
roomtree__UIA___props["TableRowHeaders"] = 30081
roomtree__UIA___fullnames["TableColumnHeadersPropertyId"] = 30082
roomtree__UIA___props["TableColumnHeaders"] = 30082
roomtree__UIA___fullnames["TableRowOrColumnMajorPropertyId"] = 30083
roomtree__UIA___props["TableRowOrColumnMajor"] = 30083
roomtree__UIA___fullnames["TableItemRowHeaderItemsPropertyId"] = 30084
roomtree__UIA___props["TableItemRowHeaderItems"] = 30084
roomtree__UIA___fullnames["TableItemColumnHeaderItemsPropertyId"] = 30085
roomtree__UIA___props["TableItemColumnHeaderItems"] = 30085
roomtree__UIA___fullnames["ToggleToggleStatePropertyId"] = 30086
roomtree__UIA___props["ToggleToggleState"] = 30086
roomtree__UIA___fullnames["TransformCanMovePropertyId"] = 30087
roomtree__UIA___props["TransformCanMove"] = 30087
roomtree__UIA___fullnames["TransformCanResizePropertyId"] = 30088
roomtree__UIA___props["TransformCanResize"] = 30088
roomtree__UIA___fullnames["TransformCanRotatePropertyId"] = 30089
roomtree__UIA___props["TransformCanRotate"] = 30089
roomtree__UIA___fullnames["IsLegacyIAccessiblePatternAvailablePropertyId"] = 30090
roomtree__UIA___props["IsLegacyIAccessiblePatternAvailable"] = 30090
roomtree__UIA___fullnames["LegacyIAccessibleChildIdPropertyId"] = 30091
roomtree__UIA___props["LegacyIAccessibleChildId"] = 30091
roomtree__UIA___fullnames["LegacyIAccessibleNamePropertyId"] = 30092
roomtree__UIA___props["LegacyIAccessibleName"] = 30092
roomtree__UIA___fullnames["LegacyIAccessibleValuePropertyId"] = 30093
roomtree__UIA___props["LegacyIAccessibleValue"] = 30093
roomtree__UIA___fullnames["LegacyIAccessibleDescriptionPropertyId"] = 30094
roomtree__UIA___props["LegacyIAccessibleDescription"] = 30094
roomtree__UIA___fullnames["LegacyIAccessibleRolePropertyId"] = 30095
roomtree__UIA___props["LegacyIAccessibleRole"] = 30095
roomtree__UIA___fullnames["LegacyIAccessibleStatePropertyId"] = 30096
roomtree__UIA___props["LegacyIAccessibleState"] = 30096
roomtree__UIA___fullnames["LegacyIAccessibleHelpPropertyId"] = 30097
roomtree__UIA___props["LegacyIAccessibleHelp"] = 30097
roomtree__UIA___fullnames["LegacyIAccessibleKeyboardShortcutPropertyId"] = 30098
roomtree__UIA___props["LegacyIAccessibleKeyboardShortcut"] = 30098
roomtree__UIA___fullnames["LegacyIAccessibleSelectionPropertyId"] = 30099
roomtree__UIA___props["LegacyIAccessibleSelection"] = 30099
roomtree__UIA___fullnames["LegacyIAccessibleDefaultActionPropertyId"] = 30100
roomtree__UIA___props["LegacyIAccessibleDefaultAction"] = 30100
roomtree__UIA___fullnames["AriaRolePropertyId"] = 30101
roomtree__UIA___props["AriaRole"] = 30101
roomtree__UIA___fullnames["AriaPropertiesPropertyId"] = 30102
roomtree__UIA___props["AriaProperties"] = 30102
roomtree__UIA___fullnames["IsDataValidForFormPropertyId"] = 30103
roomtree__UIA___props["IsDataValidForForm"] = 30103
roomtree__UIA___fullnames["ControllerForPropertyId"] = 30104
roomtree__UIA___props["ControllerFor"] = 30104
roomtree__UIA___fullnames["DescribedByPropertyId"] = 30105
roomtree__UIA___props["DescribedBy"] = 30105
roomtree__UIA___fullnames["FlowsToPropertyId"] = 30106
roomtree__UIA___props["FlowsTo"] = 30106
roomtree__UIA___fullnames["ProviderDescriptionPropertyId"] = 30107
roomtree__UIA___props["ProviderDescription"] = 30107
roomtree__UIA___fullnames["IsItemContainerPatternAvailablePropertyId"] = 30108
roomtree__UIA___props["IsItemContainerPatternAvailable"] = 30108
roomtree__UIA___fullnames["IsVirtualizedItemPatternAvailablePropertyId"] = 30109
roomtree__UIA___props["IsVirtualizedItemPatternAvailable"] = 30109
roomtree__UIA___fullnames["IsSynchronizedInputPatternAvailablePropertyId"] = 30110
roomtree__UIA___props["IsSynchronizedInputPatternAvailable"] = 30110
roomtree__UIA___fullnames["OptimizeForVisualContentPropertyId"] = 30111
roomtree__UIA___props["OptimizeForVisualContent"] = 30111
roomtree__UIA___fullnames["IsObjectModelPatternAvailablePropertyId"] = 30112
roomtree__UIA___props["IsObjectModelPatternAvailable"] = 30112
roomtree__UIA___fullnames["AnnotationAnnotationTypeIdPropertyId"] = 30113
roomtree__UIA___props["AnnotationAnnotationTypeId"] = 30113
roomtree__UIA___fullnames["AnnotationAnnotationTypeNamePropertyId"] = 30114
roomtree__UIA___props["AnnotationAnnotationTypeName"] = 30114
roomtree__UIA___fullnames["AnnotationAuthorPropertyId"] = 30115
roomtree__UIA___props["AnnotationAuthor"] = 30115
roomtree__UIA___fullnames["AnnotationDateTimePropertyId"] = 30116
roomtree__UIA___props["AnnotationDateTime"] = 30116
roomtree__UIA___fullnames["AnnotationTargetPropertyId"] = 30117
roomtree__UIA___props["AnnotationTarget"] = 30117
roomtree__UIA___fullnames["IsAnnotationPatternAvailablePropertyId"] = 30118
roomtree__UIA___props["IsAnnotationPatternAvailable"] = 30118
roomtree__UIA___fullnames["IsTextPattern2AvailablePropertyId"] = 30119
roomtree__UIA___props["IsTextPattern2Available"] = 30119
roomtree__UIA___fullnames["StylesStyleIdPropertyId"] = 30120
roomtree__UIA___props["StylesStyleId"] = 30120
roomtree__UIA___fullnames["StylesStyleNamePropertyId"] = 30121
roomtree__UIA___props["StylesStyleName"] = 30121
roomtree__UIA___fullnames["StylesFillColorPropertyId"] = 30122
roomtree__UIA___props["StylesFillColor"] = 30122
roomtree__UIA___fullnames["StylesFillPatternStylePropertyId"] = 30123
roomtree__UIA___props["StylesFillPatternStyle"] = 30123
roomtree__UIA___fullnames["StylesShapePropertyId"] = 30124
roomtree__UIA___props["StylesShape"] = 30124
roomtree__UIA___fullnames["StylesFillPatternColorPropertyId"] = 30125
roomtree__UIA___props["StylesFillPatternColor"] = 30125
roomtree__UIA___fullnames["StylesExtendedPropertiesPropertyId"] = 30126
roomtree__UIA___props["StylesExtendedProperties"] = 30126
roomtree__UIA___fullnames["IsStylesPatternAvailablePropertyId"] = 30127
roomtree__UIA___props["IsStylesPatternAvailable"] = 30127
roomtree__UIA___fullnames["IsSpreadsheetPatternAvailablePropertyId"] = 30128
roomtree__UIA___props["IsSpreadsheetPatternAvailable"] = 30128
roomtree__UIA___fullnames["SpreadsheetItemFormulaPropertyId"] = 30129
roomtree__UIA___props["SpreadsheetItemFormula"] = 30129
roomtree__UIA___fullnames["SpreadsheetItemAnnotationObjectsPropertyId"] = 30130
roomtree__UIA___props["SpreadsheetItemAnnotationObjects"] = 30130
roomtree__UIA___fullnames["SpreadsheetItemAnnotationTypesPropertyId"] = 30131
roomtree__UIA___props["SpreadsheetItemAnnotationTypes"] = 30131
roomtree__UIA___fullnames["IsSpreadsheetItemPatternAvailablePropertyId"] = 30132
roomtree__UIA___props["IsSpreadsheetItemPatternAvailable"] = 30132
roomtree__UIA___fullnames["Transform2CanZoomPropertyId"] = 30133
roomtree__UIA___props["Transform2CanZoom"] = 30133
roomtree__UIA___fullnames["IsTransformPattern2AvailablePropertyId"] = 30134
roomtree__UIA___props["IsTransformPattern2Available"] = 30134
roomtree__UIA___fullnames["LiveSettingPropertyId"] = 30135
roomtree__UIA___props["LiveSetting"] = 30135
roomtree__UIA___fullnames["IsTextChildPatternAvailablePropertyId"] = 30136
roomtree__UIA___props["IsTextChildPatternAvailable"] = 30136
roomtree__UIA___fullnames["IsDragPatternAvailablePropertyId"] = 30137
roomtree__UIA___props["IsDragPatternAvailable"] = 30137
roomtree__UIA___fullnames["DragIsGrabbedPropertyId"] = 30138
roomtree__UIA___props["DragIsGrabbed"] = 30138
roomtree__UIA___fullnames["DragDropEffectPropertyId"] = 30139
roomtree__UIA___props["DragDropEffect"] = 30139
roomtree__UIA___fullnames["DragDropEffectsPropertyId"] = 30140
roomtree__UIA___props["DragDropEffects"] = 30140
roomtree__UIA___fullnames["IsDropTargetPatternAvailablePropertyId"] = 30141
roomtree__UIA___props["IsDropTargetPatternAvailable"] = 30141
roomtree__UIA___fullnames["DropTargetDropTargetEffectPropertyId"] = 30142
roomtree__UIA___props["DropTargetDropTargetEffect"] = 30142
roomtree__UIA___fullnames["DropTargetDropTargetEffectsPropertyId"] = 30143
roomtree__UIA___props["DropTargetDropTargetEffects"] = 30143
roomtree__UIA___fullnames["DragGrabbedItemsPropertyId"] = 30144
roomtree__UIA___props["DragGrabbedItems"] = 30144
roomtree__UIA___fullnames["Transform2ZoomLevelPropertyId"] = 30145
roomtree__UIA___props["Transform2ZoomLevel"] = 30145
roomtree__UIA___fullnames["Transform2ZoomMinimumPropertyId"] = 30146
roomtree__UIA___props["Transform2ZoomMinimum"] = 30146
roomtree__UIA___fullnames["Transform2ZoomMaximumPropertyId"] = 30147
roomtree__UIA___props["Transform2ZoomMaximum"] = 30147
roomtree__UIA___fullnames["FlowsFromPropertyId"] = 30148
roomtree__UIA___props["FlowsFrom"] = 30148
roomtree__UIA___fullnames["IsTextEditPatternAvailablePropertyId"] = 30149
roomtree__UIA___props["IsTextEditPatternAvailable"] = 30149
roomtree__UIA___fullnames["IsPeripheralPropertyId"] = 30150
roomtree__UIA___props["IsPeripheral"] = 30150
roomtree__UIA___fullnames["PositionInSetPropertyId"] = 30152
roomtree__UIA___props["PositionInSet"] = 30152
roomtree__UIA___fullnames["SizeOfSetPropertyId"] = 30153
roomtree__UIA___props["SizeOfSet"] = 30153
roomtree__UIA___fullnames["LevelPropertyId"] = 30154
roomtree__UIA___props["Level"] = 30154
roomtree__UIA___fullnames["AnnotationTypesPropertyId"] = 30155
roomtree__UIA___props["AnnotationTypes"] = 30155
roomtree__UIA___fullnames["AnnotationObjectsPropertyId"] = 30156
roomtree__UIA___props["AnnotationObjects"] = 30156
roomtree__UIA___fullnames["LandmarkTypePropertyId"] = 30157
roomtree__UIA___props["LandmarkType"] = 30157
roomtree__UIA___fullnames["LocalizedLandmarkTypePropertyId"] = 30158
roomtree__UIA___props["LocalizedLandmarkType"] = 30158
roomtree__UIA___fullnames["FullDescriptionPropertyId"] = 30159
roomtree__UIA___props["FullDescription"] = 30159
roomtree__UIA___fullnames["IsCustomNavigationPatternAvailablePropertyId"] = 30151
roomtree__UIA___props["IsCustomNavigationPatternAvailable"] = 30151
roomtree__UIA___fullnames["FillColorPropertyId"] = 30160
roomtree__UIA___props["FillColor"] = 30160
roomtree__UIA___fullnames["OutlineColorPropertyId"] = 30161
roomtree__UIA___props["OutlineColor"] = 30161
roomtree__UIA___fullnames["FillTypePropertyId"] = 30162
roomtree__UIA___props["FillType"] = 30162
roomtree__UIA___fullnames["VisualEffectsPropertyId"] = 30163
roomtree__UIA___props["VisualEffects"] = 30163
roomtree__UIA___fullnames["OutlineThicknessPropertyId"] = 30164
roomtree__UIA___props["OutlineThickness"] = 30164
roomtree__UIA___fullnames["CenterPointPropertyId"] = 30165
roomtree__UIA___props["CenterPoint"] = 30165
roomtree__UIA___fullnames["RotationPropertyId"] = 30166
roomtree__UIA___props["Rotation"] = 30166
roomtree__UIA___fullnames["SizePropertyId"] = 30167
roomtree__UIA___props["Size"] = 30167
roomtree__UIA___fullnames["UIA_AnnotationType_Unknown"] = 60000
roomtree__UIA___vals["UIA_AnnotationType_Unknown"] = 60000
roomtree__UIA___fullnames["UIA_AnnotationType_SpellingError"] = 60001
roomtree__UIA___vals["UIA_AnnotationType_SpellingError"] = 60001
roomtree__UIA___fullnames["UIA_AnnotationType_GrammarError"] = 60002
roomtree__UIA___vals["UIA_AnnotationType_GrammarError"] = 60002
roomtree__UIA___fullnames["UIA_AnnotationType_Comment"] = 60003
roomtree__UIA___vals["UIA_AnnotationType_Comment"] = 60003
roomtree__UIA___fullnames["UIA_AnnotationType_FormulaError"] = 60004
roomtree__UIA___vals["UIA_AnnotationType_FormulaError"] = 60004
roomtree__UIA___fullnames["UIA_AnnotationType_TrackChanges"] = 60005
roomtree__UIA___vals["UIA_AnnotationType_TrackChanges"] = 60005
roomtree__UIA___fullnames["UIA_AnnotationType_Header"] = 60006
roomtree__UIA___vals["UIA_AnnotationType_Header"] = 60006
roomtree__UIA___fullnames["UIA_AnnotationType_Footer"] = 60007
roomtree__UIA___vals["UIA_AnnotationType_Footer"] = 60007
roomtree__UIA___fullnames["UIA_AnnotationType_Highlighted"] = 60008
roomtree__UIA___vals["UIA_AnnotationType_Highlighted"] = 60008
roomtree__UIA___fullnames["UIA_AnnotationType_Endnote"] = 60009
roomtree__UIA___vals["UIA_AnnotationType_Endnote"] = 60009
roomtree__UIA___fullnames["UIA_AnnotationType_Footnote"] = 60010
roomtree__UIA___vals["UIA_AnnotationType_Footnote"] = 60010
roomtree__UIA___fullnames["UIA_AnnotationType_InsertionChange"] = 60011
roomtree__UIA___vals["UIA_AnnotationType_InsertionChange"] = 60011
roomtree__UIA___fullnames["UIA_AnnotationType_DeletionChange"] = 60012
roomtree__UIA___vals["UIA_AnnotationType_DeletionChange"] = 60012
roomtree__UIA___fullnames["UIA_AnnotationType_MoveChange"] = 60013
roomtree__UIA___vals["UIA_AnnotationType_MoveChange"] = 60013
roomtree__UIA___fullnames["UIA_AnnotationType_FormatChange"] = 60014
roomtree__UIA___vals["UIA_AnnotationType_FormatChange"] = 60014
roomtree__UIA___fullnames["UIA_AnnotationType_UnsyncedChange"] = 60015
roomtree__UIA___vals["UIA_AnnotationType_UnsyncedChange"] = 60015
roomtree__UIA___fullnames["UIA_AnnotationType_EditingLockedChange"] = 60016
roomtree__UIA___vals["UIA_AnnotationType_EditingLockedChange"] = 60016
roomtree__UIA___fullnames["UIA_AnnotationType_ExternalChange"] = 60017
roomtree__UIA___vals["UIA_AnnotationType_ExternalChange"] = 60017
roomtree__UIA___fullnames["UIA_AnnotationType_ConflictingChange"] = 60018
roomtree__UIA___vals["UIA_AnnotationType_ConflictingChange"] = 60018
roomtree__UIA___fullnames["UIA_AnnotationType_Author"] = 60019
roomtree__UIA___vals["UIA_AnnotationType_Author"] = 60019
roomtree__UIA___fullnames["UIA_AnnotationType_AdvancedProofingIssue"] = 60020
roomtree__UIA___vals["UIA_AnnotationType_AdvancedProofingIssue"] = 60020
roomtree__UIA___fullnames["UIA_AnnotationType_DataValidationError"] = 60021
roomtree__UIA___vals["UIA_AnnotationType_DataValidationError"] = 60021
roomtree__UIA___fullnames["UIA_AnnotationType_CircularReferenceError"] = 60022
roomtree__UIA___vals["UIA_AnnotationType_CircularReferenceError"] = 60022
roomtree__UIA___fullnames["ToolTipOpenedEventId"] = 20000
roomtree__UIA___events["ToolTipOpened"] = 20000
roomtree__UIA___fullnames["ToolTipClosedEventId"] = 20001
roomtree__UIA___events["ToolTipClosed"] = 20001
roomtree__UIA___fullnames["StructureChangedEventId"] = 20002
roomtree__UIA___events["StructureChanged"] = 20002
roomtree__UIA___fullnames["MenuOpenedEventId"] = 20003
roomtree__UIA___events["MenuOpened"] = 20003
roomtree__UIA___fullnames["AutomationPropertyChangedEventId"] = 20004
roomtree__UIA___events["AutomationPropertyChanged"] = 20004
roomtree__UIA___fullnames["AutomationFocusChangedEventId"] = 20005
roomtree__UIA___events["AutomationFocusChanged"] = 20005
roomtree__UIA___fullnames["AsyncContentLoadedEventId"] = 20006
roomtree__UIA___events["AsyncContentLoaded"] = 20006
roomtree__UIA___fullnames["MenuClosedEventId"] = 20007
roomtree__UIA___events["MenuClosed"] = 20007
roomtree__UIA___fullnames["LayoutInvalidatedEventId"] = 20008
roomtree__UIA___events["LayoutInvalidated"] = 20008
roomtree__UIA___fullnames["Invoke_InvokedEventId"] = 20009
roomtree__UIA___events["Invoke_Invoked"] = 20009
roomtree__UIA___fullnames["SelectionItem_ElementAddedToSelectionEventId"] = 20010
roomtree__UIA___events["SelectionItem_ElementAddedToSelection"] = 20010
roomtree__UIA___fullnames["SelectionItem_ElementRemovedFromSelectionEventId"] = 20011
roomtree__UIA___events["SelectionItem_ElementRemovedFromSelection"] = 20011
roomtree__UIA___fullnames["SelectionItem_ElementSelectedEventId"] = 20012
roomtree__UIA___events["SelectionItem_ElementSelected"] = 20012
roomtree__UIA___fullnames["Selection_InvalidatedEventId"] = 20013
roomtree__UIA___events["Selection_Invalidated"] = 20013
roomtree__UIA___fullnames["Text_TextSelectionChangedEventId"] = 20014
roomtree__UIA___events["Text_TextSelectionChanged"] = 20014
roomtree__UIA___fullnames["Text_TextChangedEventId"] = 20015
roomtree__UIA___events["Text_TextChanged"] = 20015
roomtree__UIA___fullnames["Window_WindowOpenedEventId"] = 20016
roomtree__UIA___events["Window_WindowOpened"] = 20016
roomtree__UIA___fullnames["Window_WindowClosedEventId"] = 20017
roomtree__UIA___events["Window_WindowClosed"] = 20017
roomtree__UIA___fullnames["MenuModeStartEventId"] = 20018
roomtree__UIA___events["MenuModeStart"] = 20018
roomtree__UIA___fullnames["MenuModeEndEventId"] = 20019
roomtree__UIA___events["MenuModeEnd"] = 20019
roomtree__UIA___fullnames["InputReachedTargetEventId"] = 20020
roomtree__UIA___events["InputReachedTarget"] = 20020
roomtree__UIA___fullnames["InputReachedOtherElementEventId"] = 20021
roomtree__UIA___events["InputReachedOtherElement"] = 20021
roomtree__UIA___fullnames["InputDiscardedEventId"] = 20022
roomtree__UIA___events["InputDiscarded"] = 20022
roomtree__UIA___fullnames["SystemAlertEventId"] = 20023
roomtree__UIA___events["SystemAlert"] = 20023
roomtree__UIA___fullnames["LiveRegionChangedEventId"] = 20024
roomtree__UIA___events["LiveRegionChanged"] = 20024
roomtree__UIA___fullnames["HostedFragmentRootsInvalidatedEventId"] = 20025
roomtree__UIA___events["HostedFragmentRootsInvalidated"] = 20025
roomtree__UIA___fullnames["DragStartEventId"] = 20026
roomtree__UIA___events["DragStart"] = 20026
roomtree__UIA___fullnames["DragCancelEventId"] = 20027
roomtree__UIA___events["DragCancel"] = 20027
roomtree__UIA___fullnames["Drag_CompleteEventId"] = 20028
roomtree__UIA___events["Drag_Complete"] = 20028
roomtree__UIA___fullnames["DragEnterEventId"] = 20029
roomtree__UIA___events["DragEnter"] = 20029
roomtree__UIA___fullnames["DragLeaveEventId"] = 20030
roomtree__UIA___events["DragLeave"] = 20030
roomtree__UIA___fullnames["DroppedEventId"] = 20031
roomtree__UIA___events["Dropped"] = 20031
roomtree__UIA___fullnames["TextEdit_TextChangedEventId"] = 20032
roomtree__UIA___events["TextEdit_TextChanged"] = 20032
roomtree__UIA___fullnames["TextEdit_ConversionTargetChangedEventId"] = 20033
roomtree__UIA___events["TextEdit_ConversionTargetChanged"] = 20033
roomtree__UIA___fullnames["OrientationType_None"] = 0
roomtree__UIA___vals["OrientationType_None"] = 0
roomtree__UIA___fullnames["OrientationType_Horizontal"] = 1
roomtree__UIA___vals["OrientationType_Horizontal"] = 1
roomtree__UIA___fullnames["OrientationType_Vertical"] = 2
roomtree__UIA___vals["OrientationType_Vertical"] = 2
roomtree__UIA___fullnames["TreeScope_Element"] = 1
roomtree__UIA___vals["TreeScope_Element"] = 1
roomtree__UIA___fullnames["TreeScope_Children"] = 2
roomtree__UIA___vals["TreeScope_Children"] = 2
roomtree__UIA___fullnames["TreeScope_Descendants"] = 4
roomtree__UIA___vals["TreeScope_Descendants"] = 4
roomtree__UIA___fullnames["TreeScope_Parent"] = 8
roomtree__UIA___vals["TreeScope_Parent"] = 8
roomtree__UIA___fullnames["TreeScope_Ancestors"] = 16
roomtree__UIA___vals["TreeScope_Ancestors"] = 16
roomtree__UIA___fullnames["TreeScope_Subtree"] = 7
roomtree__UIA___vals["TreeScope_Subtree"] = 7
roomtree__UIA___fullnames["next"] = 1
roomtree__UIA___vals["next"] = 1
roomtree__UIA___fullnames["prior"] = 0
roomtree__UIA___vals["prior"] = 0
roomtree__UIA___fullnames["ScrollAmount_LargeDecrement"] = 0
roomtree__UIA___vals["ScrollAmount_LargeDecrement"] = 0
roomtree__UIA___fullnames["ScrollAmount_SmallDecrement"] = 1
roomtree__UIA___vals["ScrollAmount_SmallDecrement"] = 1
roomtree__UIA___fullnames["ScrollAmount_NoAmount"] = 2
roomtree__UIA___vals["ScrollAmount_NoAmount"] = 2
roomtree__UIA___fullnames["ScrollAmount_LargeIncrement"] = 3
roomtree__UIA___vals["ScrollAmount_LargeIncrement"] = 3
roomtree__UIA___fullnames["ScrollAmount_SmallIncrement"] = 4
roomtree__UIA___vals["ScrollAmount_SmallIncrement"] = 4
roomtree__UIA___fullnames["StructureChangeType_ChildAdded"] = 0
roomtree__UIA___vals["StructureChangeType_ChildAdded"] = 0
roomtree__UIA___fullnames["StructureChangeType_ChildRemoved"] = 1
roomtree__UIA___vals["StructureChangeType_ChildRemoved"] = 1
roomtree__UIA___fullnames["StructureChangeType_ChildrenInvalidated"] = 2
roomtree__UIA___vals["StructureChangeType_ChildrenInvalidated"] = 2
roomtree__UIA___fullnames["StructureChangeType_ChildrenBulkAdded"] = 3
roomtree__UIA___vals["StructureChangeType_ChildrenBulkAdded"] = 3
roomtree__UIA___fullnames["StructureChangeType_ChildrenBulkRemoved"] = 4
roomtree__UIA___vals["StructureChangeType_ChildrenBulkRemoved"] = 4
roomtree__UIA___fullnames["StructureChangeType_ChildrenReordered"] = 5
roomtree__UIA___vals["StructureChangeType_ChildrenReordered"] = 5
roomtree__UIA___fullnames["AnimationStyleAttributeId"] = 40000
roomtree__UIA___attributes["AnimationStyle"] = 40000
roomtree__UIA___fullnames["AnnotationObjectsAttributeId"] = 40032
roomtree__UIA___attributes["AnnotationObjects"] = 40032
roomtree__UIA___fullnames["AnnotationTypesAttributeId"] = 40031
roomtree__UIA___attributes["AnnotationTypes"] = 40031
roomtree__UIA___fullnames["BackgroundColorAttributeId"] = 40001
roomtree__UIA___attributes["BackgroundColor"] = 40001
roomtree__UIA___fullnames["BulletStyleAttributeId"] = 40002
roomtree__UIA___attributes["BulletStyle"] = 40002
roomtree__UIA___fullnames["CapStyleAttributeId"] = 40003
roomtree__UIA___attributes["CapStyle"] = 40003
roomtree__UIA___fullnames["CaretBidiModeAttributeId"] = 40039
roomtree__UIA___attributes["CaretBidiMode"] = 40039
roomtree__UIA___fullnames["CaretPositionAttributeId"] = 40038
roomtree__UIA___attributes["CaretPosition"] = 40038
roomtree__UIA___fullnames["CultureAttributeId"] = 40004
roomtree__UIA___attributes["Culture"] = 40004
roomtree__UIA___fullnames["FontNameAttributeId"] = 40005
roomtree__UIA___attributes["FontName"] = 40005
roomtree__UIA___fullnames["FontSizeAttributeId"] = 40006
roomtree__UIA___attributes["FontSize"] = 40006
roomtree__UIA___fullnames["FontWeightAttributeId"] = 400007
roomtree__UIA___attributes["FontWeight"] = 400007
roomtree__UIA___fullnames["ForegroundColorAttributeId"] = 40008
roomtree__UIA___attributes["ForegroundColor"] = 40008
roomtree__UIA___fullnames["HorizontalTextAlignmentAttributeId"] = 40009
roomtree__UIA___attributes["HorizontalTextAlignment"] = 40009
roomtree__UIA___fullnames["IndentationFirstLineAttributeId"] = 40010
roomtree__UIA___attributes["IndentationFirstLine"] = 40010
roomtree__UIA___fullnames["IndentationLeadingAttributeId"] = 40011
roomtree__UIA___attributes["IndentationLeading"] = 40011
roomtree__UIA___fullnames["IndentationTrailingAttributeId"] = 40012
roomtree__UIA___attributes["IndentationTrailing"] = 40012
roomtree__UIA___fullnames["IsActiveAttributeId"] = 40036
roomtree__UIA___attributes["IsActive"] = 40036
roomtree__UIA___fullnames["IsHiddenAttributeId"] = 40013
roomtree__UIA___attributes["IsHidden"] = 40013
roomtree__UIA___fullnames["IsItalicAttributeId"] = 40014
roomtree__UIA___attributes["IsItalic"] = 40014
roomtree__UIA___fullnames["IsReadOnlyAttributeId"] = 40015
roomtree__UIA___attributes["IsReadOnly"] = 40015
roomtree__UIA___fullnames["IsSubscriptAttributeId"] = 40016
roomtree__UIA___attributes["IsSubscript"] = 40016
roomtree__UIA___fullnames["IsSuperscriptAttributeId"] = 40017
roomtree__UIA___attributes["IsSuperscript"] = 40017
roomtree__UIA___fullnames["LinkAttributeId"] = 40035
roomtree__UIA___attributes["Link"] = 40035
roomtree__UIA___fullnames["MarginBottomAttributeId"] = 40018
roomtree__UIA___attributes["MarginBottom"] = 40018
roomtree__UIA___fullnames["MarginLeadingAttributeId"] = 40019
roomtree__UIA___attributes["MarginLeading"] = 40019
roomtree__UIA___fullnames["MarginTopAttributeId"] = 40020
roomtree__UIA___attributes["MarginTop"] = 40020
roomtree__UIA___fullnames["MarginTrailingAttributeId"] = 40021
roomtree__UIA___attributes["MarginTrailing"] = 40021
roomtree__UIA___fullnames["OutlineStylesAttributeId"] = 40022
roomtree__UIA___attributes["OutlineStyles"] = 40022
roomtree__UIA___fullnames["OverlineColorAttributeId"] = 40023
roomtree__UIA___attributes["OverlineColor"] = 40023
roomtree__UIA___fullnames["OverlineStyleAttributeId"] = 40024
roomtree__UIA___attributes["OverlineStyle"] = 40024
roomtree__UIA___fullnames["SelectionActiveEndAttributeId"] = 40037
roomtree__UIA___attributes["SelectionActiveEnd"] = 40037
roomtree__UIA___fullnames["StrikethroughColorAttributeId"] = 40025
roomtree__UIA___attributes["StrikethroughColor"] = 40025
roomtree__UIA___fullnames["StrikethroughStyleAttributeId"] = 40026
roomtree__UIA___attributes["StrikethroughStyle"] = 40026
roomtree__UIA___fullnames["StyleIdAttributeId"] = 40034
roomtree__UIA___attributes["StyleId"] = 40034
roomtree__UIA___fullnames["StyleNameAttributeId"] = 40033
roomtree__UIA___attributes["StyleName"] = 40033
roomtree__UIA___fullnames["TabsAttributeId"] = 40027
roomtree__UIA___attributes["Tabs"] = 40027
roomtree__UIA___fullnames["TextFlowDirectionsAttributeId"] = 40028
roomtree__UIA___attributes["TextFlowDirections"] = 40028
roomtree__UIA___fullnames["UnderlineColorAttributeId"] = 40029
roomtree__UIA___attributes["UnderlineColor"] = 40029
roomtree__UIA___fullnames["UnderlineStyleAttributeId"] = 40030
roomtree__UIA___attributes["UnderlineStyle"] = 40030
roomtree__UIA___fullnames["TextUnit_Character"] = 0
roomtree__UIA___textUnits["Character"] = 0
roomtree__UIA___fullnames["TextUnit_Format"] = 1
roomtree__UIA___textUnits["Format"] = 1
roomtree__UIA___fullnames["TextUnit_Word"] = 2
roomtree__UIA___textUnits["Word"] = 2
roomtree__UIA___fullnames["TextUnit_Line"] = 3
roomtree__UIA___textUnits["Line"] = 3
roomtree__UIA___fullnames["TextUnit_Paragraph"] = 4
roomtree__UIA___textUnits["Paragraph"] = 4
roomtree__UIA___fullnames["TextUnit_Page"] = 5
roomtree__UIA___textUnits["Page"] = 5
roomtree__UIA___fullnames["TextUnit_Document"] = 6
roomtree__UIA___textUnits["Document"] = 6
roomtree__UIA___fullnames["TextPatternRangeEndpoint_Start"] = 0
roomtree__UIA___vals["TextPatternRangeEndpoint_Start"] = 0
roomtree__UIA___fullnames["TextPatternRangeEndpoint_End"] = 1
roomtree__UIA___vals["TextPatternRangeEndpoint_End"] = 1
roomtree__UIA___fullnames["SupportedTextSelection_None"] = 0
roomtree__UIA___vals["SupportedTextSelection_None"] = 0
roomtree__UIA___fullnames["SupportedTextSelection_Single"] = 1
roomtree__UIA___vals["SupportedTextSelection_Single"] = 1
roomtree__UIA___fullnames["SupportedTextSelection_Multiple"] = 2
roomtree__UIA___vals["SupportedTextSelection_Multiple"] = 2
roomtree__UIA___fullnames["DockPosition_Top"] = 0
roomtree__UIA___vals["DockPosition_Top"] = 0
roomtree__UIA___fullnames["DockPosition_Left"] = 1
roomtree__UIA___vals["DockPosition_Left"] = 1
roomtree__UIA___fullnames["DockPosition_Bottom"] = 2
roomtree__UIA___vals["DockPosition_Bottom"] = 2
roomtree__UIA___fullnames["DockPosition_Right"] = 3
roomtree__UIA___vals["DockPosition_Right"] = 3
roomtree__UIA___fullnames["DockPosition_Fill"] = 4
roomtree__UIA___vals["DockPosition_Fill"] = 4
roomtree__UIA___fullnames["DockPosition_None"] = 5
roomtree__UIA___vals["DockPosition_None"] = 5
roomtree__UIA___fullnames["SynchronizedInputType_KeyUp"] = 0
roomtree__UIA___vals["SynchronizedInputType_KeyUp"] = 0
roomtree__UIA___fullnames["SynchronizedInputType_KeyDown"] = 1
roomtree__UIA___vals["SynchronizedInputType_KeyDown"] = 1
roomtree__UIA___fullnames["SynchronizedInputType_LeftMouseUp"] = 2
roomtree__UIA___vals["SynchronizedInputType_LeftMouseUp"] = 2
roomtree__UIA___fullnames["SynchronizedInputType_LeftMouseDown"] = 3
roomtree__UIA___vals["SynchronizedInputType_LeftMouseDown"] = 3
roomtree__UIA___fullnames["SynchronizedInputType_RightMouseUp"] = 4
roomtree__UIA___vals["SynchronizedInputType_RightMouseUp"] = 4
roomtree__UIA___fullnames["SynchronizedInputType_RightMouseDown"] = 5
roomtree__UIA___vals["SynchronizedInputType_RightMouseDown"] = 5
roomtree__UIA___fullnames["RowOrColumnMajor_RowMajor"] = 0
roomtree__UIA___vals["RowOrColumnMajor_RowMajor"] = 0
roomtree__UIA___fullnames["RowOrColumnMajor_ColumnMajor"] = 1
roomtree__UIA___vals["RowOrColumnMajor_ColumnMajor"] = 1
roomtree__UIA___fullnames["RowOrColumnMajor_Indeterminate"] = 2
roomtree__UIA___vals["RowOrColumnMajor_Indeterminate"] = 2
roomtree__UIA___fullnames["ZoomUnit_NoAmount"] = 0
roomtree__UIA___vals["ZoomUnit_NoAmount"] = 0
roomtree__UIA___fullnames["ZoomUnit_LargeDecrement"] = 1
roomtree__UIA___vals["ZoomUnit_LargeDecrement"] = 1
roomtree__UIA___fullnames["ZoomUnit_SmallDecrement"] = 2
roomtree__UIA___vals["ZoomUnit_SmallDecrement"] = 2
roomtree__UIA___fullnames["ZoomUnit_LargeIncrement"] = 3
roomtree__UIA___vals["ZoomUnit_LargeIncrement"] = 3
roomtree__UIA___fullnames["ZoomUnit_SmallIncrement"] = 4
roomtree__UIA___vals["ZoomUnit_SmallIncrement"] = 4
roomtree__UIA___fullnames["WindowVisualState_Normal"] = 0
roomtree__UIA___vals["WindowVisualState_Normal"] = 0
roomtree__UIA___fullnames["WindowVisualState_Maximized"] = 1
roomtree__UIA___vals["WindowVisualState_Maximized"] = 1
roomtree__UIA___fullnames["WindowVisualState_Minimized"] = 2
roomtree__UIA___vals["WindowVisualState_Minimized"] = 2
roomtree__UIA___fullnames["WindowInteractionState_Running"] = 0
roomtree__UIA___vals["WindowInteractionState_Running"] = 0
roomtree__UIA___fullnames["WindowInteractionState_Closing"] = 1
roomtree__UIA___vals["WindowInteractionState_Closing"] = 1
roomtree__UIA___fullnames["WindowInteractionState_ReadyForUserInteraction"] = 2
roomtree__UIA___vals["WindowInteractionState_ReadyForUserInteraction"] = 2
roomtree__UIA___fullnames["WindowInteractionState_BlockedByModalWindow"] = 3
roomtree__UIA___vals["WindowInteractionState_BlockedByModalWindow"] = 3
roomtree__UIA___fullnames["WindowInteractionState_NotResponding"] = 4
roomtree__UIA___vals["WindowInteractionState_NotResponding"] = 4
roomtree__UIA___fullnames["TextEditChangeType_None"] = 0
roomtree__UIA___vals["TextEditChangeType_None"] = 0
roomtree__UIA___fullnames["TextEditChangeType_AutoCorrect"] = 1
roomtree__UIA___vals["TextEditChangeType_AutoCorrect"] = 1
roomtree__UIA___fullnames["TextEditChangeType_Composition"] = 2
roomtree__UIA___vals["TextEditChangeType_Composition"] = 2
roomtree__UIA___fullnames["TextEditChangeType_CompositionFinalized"] = 3
roomtree__UIA___vals["TextEditChangeType_CompositionFinalized"] = 3
roomtree__UIA___fullnames["ExpandCollapseState_Collapsed"] = 0
roomtree__UIA___vals["ExpandCollapseState_Collapsed"] = 0
roomtree__UIA___fullnames["ExpandCollapseState_Expanded"] = 1
roomtree__UIA___vals["ExpandCollapseState_Expanded"] = 1
roomtree__UIA___fullnames["ExpandCollapseState_PartiallyExpanded"] = 2
roomtree__UIA___vals["ExpandCollapseState_PartiallyExpanded"] = 2
roomtree__UIA___fullnames["ExpandCollapseState_LeafNode"] = 3
roomtree__UIA___vals["ExpandCollapseState_LeafNode"] = 3
roomtree__UIA___fullnames["StyleId_BulletedList"] = 70015
roomtree__UIA___vals["StyleId_BulletedList"] = 70015
roomtree__UIA___fullnames["StyleId_Custom"] = 70000
roomtree__UIA___vals["StyleId_Custom"] = 70000
roomtree__UIA___fullnames["StyleId_Emphasis"] = 70013
roomtree__UIA___vals["StyleId_Emphasis"] = 70013
roomtree__UIA___fullnames["StyleId_Heading1"] = 70001
roomtree__UIA___vals["StyleId_Heading1"] = 70001
roomtree__UIA___fullnames["StyleId_Heading2"] = 70002
roomtree__UIA___vals["StyleId_Heading2"] = 70002
roomtree__UIA___fullnames["StyleId_Heading3"] = 70003
roomtree__UIA___vals["StyleId_Heading3"] = 70003
roomtree__UIA___fullnames["StyleId_Heading4"] = 70004
roomtree__UIA___vals["StyleId_Heading4"] = 70004
roomtree__UIA___fullnames["StyleId_Heading5"] = 70005
roomtree__UIA___vals["StyleId_Heading5"] = 70005
roomtree__UIA___fullnames["StyleId_Heading6"] = 70006
roomtree__UIA___vals["StyleId_Heading6"] = 70006
roomtree__UIA___fullnames["StyleId_Heading7"] = 70007
roomtree__UIA___vals["StyleId_Heading7"] = 70007
roomtree__UIA___fullnames["StyleId_Heading8"] = 70008
roomtree__UIA___vals["StyleId_Heading8"] = 70008
roomtree__UIA___fullnames["StyleId_Heading9"] = 70009
roomtree__UIA___vals["StyleId_Heading9"] = 70009
roomtree__UIA___fullnames["StyleId_Normal"] = 70012
roomtree__UIA___vals["StyleId_Normal"] = 70012
roomtree__UIA___fullnames["StyleId_NumberedList"] = 70016
roomtree__UIA___vals["StyleId_NumberedList"] = 70016
roomtree__UIA___fullnames["StyleId_Quote"] = 70014
roomtree__UIA___vals["StyleId_Quote"] = 70014
roomtree__UIA___fullnames["StyleId_Subtitle"] = 70011
roomtree__UIA___vals["StyleId_Subtitle"] = 70011
roomtree__UIA___fullnames["StyleId_Title"] = 70010
roomtree__UIA___vals["StyleId_Title"] = 70010
roomtree__UIA___fullnames["UIA_CustomLandmarkTypeId"] = 80000
roomtree__UIA___vals["UIA_CustomLandmarkTypeId"] = 80000
roomtree__UIA___fullnames["UIA_FormLandmarkTypeId"] = 80001
roomtree__UIA___vals["UIA_FormLandmarkTypeId"] = 80001
roomtree__UIA___fullnames["UIA_MainLandmarkTypeId"] = 80002
roomtree__UIA___vals["UIA_MainLandmarkTypeId"] = 80002
roomtree__UIA___fullnames["UIA_NavigationLandmarkTypeId"] = 80003
roomtree__UIA___vals["UIA_NavigationLandmarkTypeId"] = 80003
roomtree__UIA___fullnames["UIA_SearchLandmarkTypeId"] = 80004
roomtree__UIA___vals["UIA_SearchLandmarkTypeId"] = 80004
roomtree__UIA___fullnames["ToggleState_Off"] = 0
roomtree__UIA___vals["ToggleState_Off"] = 0
roomtree__UIA___fullnames["ToggleState_On"] = 1
roomtree__UIA___vals["ToggleState_On"] = 1
roomtree__UIA___fullnames["ToggleState_Indeterminate"] = 2
roomtree__UIA___vals["ToggleState_Indeterminate"] = 2
roomtree__UIA___fullnames["Off"] = 0
roomtree__UIA___vals["Off"] = 0
roomtree__UIA___fullnames["Polite"] = 1
roomtree__UIA___vals["Polite"] = 1
roomtree__UIA___fullnames["Assertive"] = 2
roomtree__UIA___vals["Assertive"] = 2
roomtree__UIA___fullnames["SELFLAG_NONE"] = 0
roomtree__UIA___vals["SELFLAG_NONE"] = 0
roomtree__UIA___fullnames["SELFLAG_TAKEFOCUS"] = 1
roomtree__UIA___vals["SELFLAG_TAKEFOCUS"] = 1
roomtree__UIA___fullnames["SELFLAG_TAKESELECTION"] = 2
roomtree__UIA___vals["SELFLAG_TAKESELECTION"] = 2
roomtree__UIA___fullnames["SELFLAG_EXTENDSELECTION"] = 4
roomtree__UIA___vals["SELFLAG_EXTENDSELECTION"] = 4
roomtree__UIA___fullnames["SELFLAG_ADDSELECTION"] = 8
roomtree__UIA___vals["SELFLAG_ADDSELECTION"] = 8
roomtree__UIA___fullnames["SELFLAG_REMOVESELECTION"] = 16
roomtree__UIA___vals["SELFLAG_REMOVESELECTION"] = 16
roomtree__UIA___fullnames["TextRange_NoMaxLength"] = 0xffffffff
roomtree__UIA___vals["TextRange_NoMaxLength"] = 0xffffffff
roomtree__UIA___fullnames["CaretPosition_Unknown"] = 0
roomtree__UIA___vals["CaretPosition_Unknown"] = 0
roomtree__UIA___fullnames["CaretPosition_EndOfLine"] = 1
roomtree__UIA___vals["CaretPosition_EndOfLine"] = 1
roomtree__UIA___fullnames["CaretPosition_BeginningOfLine"] = 2
roomtree__UIA___vals["CaretPosition_BeginningOfLine"] = 2
roomtree__UIA___fullnames["ActiveEnd_None"] = 0
roomtree__UIA___vals["ActiveEnd_None"] = 0
roomtree__UIA___fullnames["ActiveEnd_Start"] = 1
roomtree__UIA___vals["ActiveEnd_Start"] = 1
roomtree__UIA___fullnames["ActiveEnd_End"] = 2
roomtree__UIA___vals["ActiveEnd_End"] = 2

roomtree__UIA___built = True
endFunction

collection function roomtree__UIA__propList()
; Return the list of supported property names.
; This is a collection with names as keys and integer property Ids as values.
roomtree__UIA___buildLists()
return roomtree__UIA___props
endFunction

void function roomtree__UIA___sayError(string s, optional string spec)
; Speak an error message.
if spec
	s = s +": " +spec
endIf
sayMessage(OT_Error, s)
endFunction


;__________________________________ Module stl _________________________________
; Standard data structures for JAWS.
;
; Author: Doug Lee of Level Access

const
; List types supported.
	roomtree__UIA__stl___LstStack = 1,
	roomtree__UIA__stl___LstQueue = 2

collection function roomtree__UIA__stl__stack(optional int maxSize)
; Return a stack for use via roomtree__UIA__stl__push/roomtree__UIA__stl__pop etc.
return roomtree__UIA__stl___new(roomtree__UIA__stl___LstStack, maxSize)
endFunction

collection function roomtree__UIA__stl__queue(optional int maxSize)
; Return a queue for use via roomtree__UIA__stl__push/roomtree__UIA__stl__pop etc.
return roomtree__UIA__stl___new(roomtree__UIA__stl___LstQueue, maxSize)
endFunction

collection function roomtree__UIA__stl___new(int ltype, optional int maxSize)
; Return a structure of the indicated type.
var collection lst lst = new collection
lst.type = ltype
if ltype == roomtree__UIA__stl___LstQueue
	; ToDo: maxSize ignored.
	lst.size = 0
	return lst
endIf

lst.maxSize = 0
lst.top = 0
if maxSize
	; More efficient.
	lst.maxSize = maxSize
	var variantArray ar ar = new variantArray[lst.maxSize]
	lst.content = ar
else
	; A bit less efficient but more flexible.
	var collection nc nc = new collection
	lst.content = nc
endIf
return lst
endFunction

int function roomtree__UIA__stl__push(collection struct, variant item)
; Push an item onto whatever structure is passed.
; Returns True on success and False on failure.
if struct.type == roomtree__UIA__stl___LstQueue
	var collection node node = new collection
	node.value = item
	var collection node0 = struct.tail
	if node0
		node0.next = node
		struct.tail = node
	else
		struct.head = node
		struct.tail = node
	endIf
	struct.size = struct.size +1
	return True
endIf

struct.top = struct.top +1
if struct.maxSize
	if struct.top > struct.maxSize
		struct.top = struct.top -1
		return False
	endIf
	var variantArray ar = struct.content
	ar[struct.top] = item
	;sayString("Push " +intToString(struct.top))
else
	var collection col = struct.content
	col[intToString(struct.top)] = item
	;sayString("Dynamic push " +intToString(struct.top))
endIf
return True
endFunction

collection function roomtree__UIA__stl__top(collection struct, optional int alsoPop)
; Return the top element of whatever structure is passed.
; Also pops the element off if alsoPop is True.
; The item is returned as the .value property of a simple collection.
; On error, null is returned rather than a collection.
var collection item item = new collection
if struct.type == roomtree__UIA__stl___LstQueue
	var collection node = struct.head
	if !node
		return
	elif alsoPop
		struct.head = node.next
		if !struct.head
			struct.tail = struct.head
		endIf
		struct.size = struct.size -1
	endIf
	return node
endIf

if struct.top < 1
	return
endIf
if struct.type == roomtree__UIA__stl___LstStack
	if struct.maxSize
		var variantArray ar = struct.content
		item.value = ar[struct.top]
	else
		var collection col = struct.content
		item.value = col[intToString(struct.top)]
	endIf
	if alsoPop
		;sayString("Pop " +intToString(struct.top))
		struct.top = struct.top -1
	endIf
	return item
endIf
return
endFunction

collection function roomtree__UIA__stl__pop(collection struct)
; Pop an item from whatever structure is passed.
; The item comes back as the .value property of a collection.
; On error, returns null instead of a collection.
return roomtree__UIA__stl__top(struct, True)
endFunction

int function roomtree__UIA__stl__size(collection struct)
; Return the size (item count) of whatever structure is passed.
if struct.type == roomtree__UIA__stl___LstQueue
	return struct.size
endIf
return struct.top
endFunction

;________________________________ End module stl _______________________________

;-------------------------------- End module UIA -------------------------------

;============================= End module roomtree =============================


;================================ Module ttutil ================================
; TeamTalk utility functions.
;
; Author:  Doug Lee

handle function ttutil__mainWindow(optional handle hwnd, int allowChats)
; Return the main top-level TeamTalk window for the focused or given window.
; If allowChats is True, returns the top-level window for a chat;
; otherwise returns null for such a case.
if !hwnd
	hwnd = getFocus()
endIf
hwnd = getTopLevelWindow(hwnd)
if !stringStartsWith(getOwningAppName(hwnd), "TeamTalk")
	return
endIf
if getParent(hwnd)
	; Happens when a dialog (e.g., volume adjustment) is active.
	hwnd = getParent(hwnd)
endIf
if ttutil__isMainWindow(hwnd)
	return hwnd
elif !allowChats
	return
endIf
; We already ruled out non-TeamTalk windows and non top-level windows.
; Dialogs are handled by the getParent check.
; Main windows returned earlier; so chats should be all we have left.
return hwnd
endFunction

int function ttutil__isMainWindow(handle hwnd)
; Returns True if and only if hwnd is the main top-level TeamTalk window.
if getWindowClass(hwnd) == "#32770"
&& (stringContains(getWindowName(hwnd), ttutil__scMainWindowName5)
|| stringContains(getWindowName(hwnd), ttutil__scMainWindowName4))
	return True
endIf
return False
endFunction

string function ttutil__statusText(handle hwnd)
; Return the status window text for the given TeamTalk instance.
hwnd = getTopLevelWindow(hwnd)
hwnd = findWindow(hwnd, "msctls_statusbar32", "")
if !hwnd
	return ""
endIf
; The status text in this app is all under childID 1.
var string txt = MSAA__windowName(hwnd, 1)
return txt
endFunction

string function ttutil__version()
; Return the full dotted version number for the current TeamTalk instance.
var string appPath = getAppFilePath()
if !appPath
	var handle hwnd = getAppMainWindow(getFocus())
	if !hwnd
		return ""
	endIf
	appPath = getWindowOwner(hwnd)
	if !appPath
		return ""
	endIf
endIf
var int major = getProgramVersion(appPath)
if !major
	return ""
endIf
var string appFolder = stringLeft(appPath, stringContainsFromRight(appPath, "\\"))
; Leaves the trailing backslash in place.
if !appFolder
	return ""
endIf
var string checkPath = formatString("%1TeamTalk%2.dll",
	appFolder, intToString(major)
)
if !fileExists(checkPath)
	return ""
endIf
var string sVer = getVersionInfoString(checkPath, "ProductVersion")
if !sVer
	return ""
endIf
sVer = stringReplaceSubstrings(sVer, ", ", ".")
return sVer
endFunction

;============================== End module ttutil ==============================


;================================ Module toolbar ===============================
; Module for reporting and tracking states of TeamTalk Classic toolbar checkbox states.
;
; Author:  Doug Lee
;

globals
; Structures by top-level TeamTalk window. Fields:
;	hTop: The top-level window (same as the key for this record).
;	hToolbar, oToolbar: The handle and MSAA client object for the toolbar window.
;	sIconNames: The names of the toolbar items (none appear in MSAA).
;	sIconStates: The states of the toolbar items (see toolbar___getStates()).
	collection toolbar___toolbars,
; How often toolbar__watchForChanges() watches for state changes,
; in tenths of a second.
	int toolbar___tenths,
; The schedule handle for toolbar__watchForChanges().
	int toolbar___sch

collection function toolbar__record()
; Get or make the record for the currently active client's toolbar.
var handle hwnd = getAppMainWindow(getFocus())
if !ttutil__isMainWindow(hwnd)
	return
endIf
var collection rec = chat4__winrecs__record(toolbar___toolbars, hwnd)
if !rec
	return
endIf

var handle hTop = hwnd
if !collectionItemExists(rec, "hTop")
|| !getWindowStyleBits(rec.hTop)
|| !rec.oToolbar.accRole(0)
	; New record.
	hwnd = findWindow(hwnd, "ToolbarWindow32", "")
	if !hwnd
		return
	endIf
	var object o = MSAA__objectFromWindow(hwnd)
	if !o
		; This happens sometimes, and UIA returns an element with no patterns. [DGL, 2016-11-27]
		return
	endIf
	var int n = o.accChildCount+0
	var string names
	if n == 12
		names = toolbar___msgNames45
	elif n == 10
		names = toolbar___msgNames43
	elif n == 9
		names = toolbar___msgNames42
	elif n == 8
		names = toolbar___msgNames41
	else
		; TODO:  Probably should report this.
		return
	endIf
	rec.hToolbar = hwnd
	rec.oToolbar = o
	rec.sIconNames = names
	rec.sIconStates = toolbar___getStates(rec)
	; Add this last as it signals record completion.
	rec.hTop = hTop
endIf

return rec
endFunction

string function toolbar___getStates(collection rec)
; Get the states of all toolbar items as a string.
var string states = ""
var object o = rec.oToolbar
var int n = o.accChildCount+0
var int i = 1
while i <= n
	states = states +"|" + intToString(o.accState(i))
	i = i +1
endWhile
return stringChopLeft(states, 1)
endFunction

int function toolbar__checkStates()
; Check and update the states of toolbar items,
; announcing any changes as needed.
; Returns True if there were any changes and False if not.
; Meant to be called on a schedule and/or after specific key presses.
var collection rec = toolbar__record()
if !rec
	return False
endIf
var string oldStates = rec.sIconStates
var string newStates = toolbar___getStates(rec)
if stringCompare(oldStates, newStates, True) == 0
	return False
endIf
var int n = stringSegmentCount(oldStates, "|")
if n != stringSegmentCount(newStates, "|")
	beep()
	rec.sIconStates = newStates
	return True
endIf
var int i = 1
var int oldState, int newState
while i <= n
	oldState = stringToInt(stringSegment(oldStates, "|", i)) & State_System_Checked
	newState = stringToInt(stringSegment(newStates, "|", i)) & State_System_Checked
	if newState != oldState
		toolbar___speakState(newState, i)
	endIf
	i = i +1
endWhile
rec.sIconStates = newStates
return True
endFunction

void function toolbar___speakState(int state, int i)
; Speak the given state.
var collection rec = toolbar__record()
var string name = stringSegment(rec.sIconNames, "|", i)
if stringCompare(name, "sep", False) == 0
	return
endIf
if stringContains(name, "/")
	if state
		name = stringSegment(name, "/", 2)
	else
		name = stringSegment(name, "/", 1)
	endIf
elif state
	name = formatString(toolbar___msgActive, name)
else
	name = formatString(toolbar___msgInactive, name)
endIf
sayUsingVoice(VCTX_Message, name, OT_Message)
endFunction

int function toolbar__watchForChanges(int tenths)
; Watch for changes in toolbar item states.
; Pass -1 to disable.
; Calls itself to do the watching.
if tenths
	; New schedule being established.
	if toolbar___sch
		unscheduleFunction(toolbar___sch)
		toolbar___sch = 0
	endIf
	toolbar___tenths = tenths
	if tenths > 0
		scheduleFunction("toolbar__watchForChanges", tenths)
	endIf
	return
endIf
; Internal call to do the watching.
toolbar___sch = scheduleFunction("toolbar__watchForChanges", toolbar___tenths)
toolbar__checkStates()
endFunction

void function toolbar__sayStatuses(int verbose)
; Speak the statuses of toolbar icons.
; If verbose is False, abbreviate the results.
var collection rec = toolbar__record()
if !rec
	; This can happen: The toolbar window can refuse to provide an MSAA client object.
	; Cause unknown. [DGL, 2016-11-27]
	beep()
	return
endIf
var string states = toolbar___getStates(rec)
var int n = stringSegmentCount(states, "|")
var int i = 1
var int state, string name
var string buf = ""
var string sep = ", "
while i <= n
	state = stringToInt(stringSegment(states, "|", i)) & State_System_Checked
	if verbose
		toolbar___speakState(state, i)
	else
		name = stringSegment(rec.sIconNames, "|", i)
		if i == 1
			if state
				buf = buf +sep +toolbar___msgConnected
			else
				buf = buf +sep +toolbar___msgDisconnected
			endIf
		elif name != "sep" && state
			buf = buf +sep +name
		endIf
	endIf
	i = i +1
endWhile
if !verbose
	buf = stringChopLeft(buf, stringLength(sep))
	sayMessage(OT_User_Requested_Information, buf)
endIf
endFunction

;============================== End module toolbar =============================


;================================ Module tooltip ===============================
; Functions related to the management of tooltips.
;
; Author:  Doug Lee

string function tooltip__getText(/*?*/ handle hwnd, int x, int y)
; Fire the tooltip associated with the current or given screen location if possible, and return its text.
; If x and y are not passed or passed as 0, the active cursor's location is used.
; If hwnd is not passed or passed as 0, the window at the effective point is used.
hwnd = tooltip__getHandle(hwnd, x, y)
if !hwnd
	return ""
endIf
return getWindowText(hwnd, Read_Everything)
endFunction

handle function tooltip__getHandle(/*?*/ handle hwnd, int x, int y)
; Create and obtain the tooltip handle associated with the current or given screen location if possible.
; If x and y are not passed or passed as 0, the active cursor's location is used.
; If hwnd is not passed or passed as 0, the window at the effective point is used.
if !x && !y
	x = getCursorCol()
	y = getCursorRow()
endIf
if !hwnd
	hwnd = getWindowAtPoint(x, y)
endIf
var handle hwndDesk = util__getDesktopWindow()
if !hwndDesk
	return 0
endIf
var handle hwndBefore = getFirstChild(hwndDesk)
if !hwndBefore
	return 0
endIf
saveCursor()  JAWSCursor()  saveCursor()
moveTo(x, y)
winclick__send(hwnd, "hover", x, y)
pause()
var int cnt = 0
while getFirstChild(hwndDesk) == hwndBefore
	cnt = cnt +1
	pause()
	if cnt > 10
		return 0
	endIf
endWhile
var handle hwndAfter = getFirstChild(hwndDesk)
sayInteger(cnt)
return hwndAfter
endFunction

;============================== End module tooltip =============================


;================================== Module opt =================================
; User option code.
;
; Usage:
;	- Call opt__init() from autoStartEvent to set file and section names for options.
;	  This is optional; iff not done, getActiveConfiguration().ini and Options are used.
;	- Call opt__get() to get option values anywhere they are needed.
;	- Call opt__set() to alter option values.
;	- Call opt__cycle() from verbosity functions to handle the verbosity UI.
;
; Author:  Doug Lee

const
; A constant used to detect missing, as opposed to set-to-empty, .ini keys.
	opt___NoVal = 12345678

globals
; Information maintained by this module.
	collection opt___data

void function opt__init(/*?*/ string sectname, string filename)
; Set the section and file names used by this module for storing options.
if !opt___data
	opt___data = new collection
endIf
opt___setVal("sectname", sectname, "NonJCFOptions")
opt___setVal("filename", filename, getActiveConfiguration()+".jcf")
endFunction

void function opt___setVal(string name, string val, string dfl)
; Sets an internal configuration option to a given or default value.
; Helper for opt__init().
var string curval = opt___data[name]
if !val && !curval
	val = dfl
endIf
if val
	opt___data[name] = val
endIf
endFunction

variant function opt__get(string varName)
; Retrieves and returns the value of the indicated user-settable variable from the ini file for these scripts.
; This function makes sure to return integer 0 for the integer value 0, rather than returning the string "0" which shows up as True in a boolean test.
; varName: The name of the option to retrieve.
var
	string val
opt__init()
; Chop off descriptions.
varname = stringSegment(varname, ":", 1)
val = iniReadString(opt___data.sectname, varname, intToString(opt___NoVal), opt___data.filename)
; Favor integer 0 (False) rather than string "0" (True).
if stringContains(stringLeft(stringTrimLeadingBlanks(val), 1), "0")
	return ""
endIf
if val == opt___NoVal
	; All default values for options go here.
	if varname == SV_ChatMessageTimes
		val = "1"
	elif varname == SV_AnalyzeTreeIcons
		val = "1"
	else
		val = ""
	endIf
endIf
return val
endFunction

int function opt__set(string varName, variant val)
; Sets the indicated user-settable variable in ttclassic.ini to the given value.
opt__init()
; Chop off descriptions.
varname = stringSegment(varname, ":", 1)
return iniWriteString(opt___data.sectname, varname, val, opt___data.filename)
endFunction

variant function opt__cycle(string varname, string choices, int iRetCurVal)
; Used by DlgSelectFunctionToRun cycler/toggler functions to cycle/toggle variable values.
; Varname should be the option name being handled (SV_* constant).
; Choices should be a list of text values, separated by vertical bars (|).
; iRetCurVal should be the integer taken by the calling toggler/cycler function.
; The return value of this function behaves according to specifications for DlgSelectFunctionToRun functions and should be returned directly by the toggler/cycler function.
; The integer value saved in ttclassic.ini will be the 0-based index of the chosen option.
var
	int i,
	int n
opt__init()
i = opt__get(varname)
n = string__segmentCount(choices, "|")
if !iRetCurVal
	; Update it.
	i = i +1
	if i >= n
		i = 0
	endIf
	opt__set(varname, i)
endIf
; i+1 rather than just i because i is 0-based and stringSegment is 1-based.
return stringSegment(choices, "|", i+1)
endFunction

;================================ End module opt ===============================


;================================ Module appOpt ================================
; User option code.

String Function jvoChatMessageTimes (int RetCurValue)
return opt__cycle(SV_ChatMessageTimes, opt__msgOffOnChoices, RetCurValue)
EndFunction

String Function jvoAnalyzeTreeIcons (int RetCurValue)
return opt__cycle(SV_AnalyzeTreeIcons, opt__msgOffOnChoices, RetCurValue)
EndFunction

void function appOpt___addVerbosityOption(string byRef lst, string optionData)
; Adds an option to the given option list.
var
	string varAndDesc
varAndDesc = formatString("jvo%1:%2",
	stringSegment(optionData, ":", 1),
	stringSegment(optionData, ":", 2))
if lst
	lst = lst +"|"
endIf
lst = lst +varAndDesc
endFunction

Script AdjustJawsOptions ()
var
	string lst

lst = ""
appOpt___addVerbosityOption(lst, SV_ChatMessageTimes)
appOpt___addVerbosityOption(lst, SV_AnalyzeTreeIcons)
OptionsTreeCore(lst)
EndScript

String Function NodeHLP (string NodeName)
	if NodeName == GetActiveConfiguration ()
return appOpt__msgNodeHlp
endIf
return nodeHlp(nodeName)
EndFunction

String Function jvoChatMessageTimesHlp()
return stringSegment(SV_ChatMessageTimes, ":", 3)
EndFunction

String Function jvoAnalyzeTreeIconsHlp()
return stringSegment(SV_AnalyzeTreeIcons, ":", 3)
EndFunction

;============================== End module appOpt ==============================


;=============================== Module settings ===============================
; Code for handling user settings. Requires JAWS 11.0 or later.
;
; Usage:
;	- For JAWS 13 or later, set up application qs and qsm files for Quick Settings.
;	- For older JAWS versions, write Options or Verbosity Settings code using settings__write() to write settings.
;	- From autoStartEvent, call settings__default(...) to set defaults for each setting.
;	- Use settings__get() to get settings as needed throughout code.
;
; Author: Doug Lee of Level Access

const
; The name of the section for settings unless otherwise specified by callers.
	settings___DflSect = "NonJCFOptions"

int function settings__default(string settingName, variant val, optional string sect)
; Establish a default JCF setting value.
; Call from AutoStartEvent.
; Settings are located in the NonJCFOptions section unless sect is given and not null.
var string dflsect = settings___DflSect
if !sect
	sect = settings___DflSect
endIf
var string keys = iniReadSectionKeys(sect, settings___filename())
if stringSegmentIndex(keys, "|", settingName, True)
	; There is already a value for this one.
	return True
endIf
return settings__write(settingName, val)
endFunction

variant function settings__get(string settingName, optional string sect)
; Return a setting value.
; Settings are located in the NonJCFOptions section unless sect is given and not null.
if getJFWVersion() >= 1300000 && (!sect || sect == settings___DflSect)
	return getNonJCFOption(settingName)
endIf
if !sect
	sect = settings___DflSect
endIf
var string cur = iniReadString(sect, settingName, "", settings___filename())
return cur
endFunction

collection function settings__getAll(optional string sect)
; Return all setting names and values as a collection.
; Settings are located in the NonJCFOptions section unless sect is given and not null.
; Warning: If multiple entries for one setting name exist, only one is returned; and which one is not defined.
if !sect
	sect = settings___DflSect
endIf
var collection c c = new collection
var string keys = iniReadSectionKeys(sect, settings___filename())
if !keys
	return c
endIf
var string sep = "|"
var int n = stringSegmentCount(keys, sep)
var int i
for i = 1 to n
	var string k = stringSegment(keys, sep, i)
	var string v = settings__get(k, sect)
	c[k] = v
endFor
return c
endFunction

int function settings__write(string settingName, variant val, optional string sect)
; Write a setting value.
; Settings are located in the NonJCFOptions section unless sect is given and not null.
if !sect
	sect = settings___DflSect
endIf
if !iniWriteString(sect, settingName, val, settings___filename())
		/*
		spellString(formatString("Default fail %1, %2, %3, %4",
			sect, settingName, val, getActiveConfiguration()
		))
		*/
	beep()
	return False
endIf
return True
endFunction

string function settings___filename()
; Return the file name to use for settings.
return getJAWSSettingsDirectory() +"\\" +getActiveConfiguration() +".jcf"
endFunction

;============================= End module settings =============================


;================================= Module stats ================================
; A module for keeping and reporting statistics, often for debugging purposes.
;
; Author:  Doug Lee of Level Access

globals
	collection stats___stats,
	int stats___enabled

void function stats__enable()
; Enable the keeping of stats.
stats___enabled = True
endFunction

void function stats__disable()
; Disable the keeping of stats for runtime efficiency.
stats___enabled = False
endFunction

void function stats__inc(string name, optional collection byRef stats)
; Add 1 to the indicated value, creating it at 1 if necessary.
stats__delta(name, 1, stats)
endFunction

void function stats__dec(string name, optional collection byRef stats)
; Subtract 1 from the indicated value, creating it at -1 if necessary.
stats__delta(name, -1, stats)
endFunction

void function stats__delta(string name, int delta, optional collection byRef stats)
; Apply a change to the indicated value.
if !stats___enabled return endIf
if !stats___stats
	stats___stats = new collection
endIf
if !stats
	stats = stats___stats
endIf
stats[name] = stats[name] +delta
endFunction

void function stats__set(string name, int val, optional collection byRef stats)
; Set the indicated value.
if !stats___enabled return endIf
if !stats___stats
	stats___stats = new collection
endIf
if !stats
	stats = stats___stats
endIf
stats[name] = val
endFunction

void function stats__clear(optional collection byRef stats)
if !stats___enabled return endIf
if stats
	stats = new collection
else
	stats___stats = new collection
endIf
endFunction

void function stats__say(optional collection byRef stats)
stats__dump(False, stats)
endFunction

void function stats__show(optional collection byRef stats)
stats__dump(True, stats)
endFunction

void function stats__dump(int show, optional collection byRef stats)
var string statname, int val
var int customStats = True
if !stats
	stats = stats___stats
	customStats = False
endIf
var int ot = OT_Message
if show
	ot = OT_User_Buffer
endIf
var int found = False
forEach statname in stats
	val = stats[statname]
	if val
		sayMessage(ot, statname +" " +intToString(val))
		found = True
	endIf
endForEach
if !found
	beep()
endIf
var int enabled = stats___enabled
stats___enabled = True
if customStats
stats__clear(stats)
else
	stats__clear()
endIf
stats___enabled = enabled
endFunction

;=============================== End module stats ==============================


;================================ Module apphelp ===============================
; Automatic access to custom application help (HTM) file.
;
; The help file should be named <configName>.htm, <configName> being the result of getActiveConfiguration().
; Alternatively, the file can be named by a HelpDocument key in a Script Options section of the active configuration's .jcf file.
; As a final alternative, the name can be supplied from autoStartEvent by calling apphelp__setDocName().
;
; Author:  Doug Lee
;
; This is public code.

globals
; Custom doc name if set by a call to apphelp__setDocName().
	string apphelp___docName,
; App in focus when the above name was set.
	string apphelp___docApp,
; Number of quick-succession calls to screenSensitiveHelp seen so far.
; apphelp___doHelp() also sets this to -1 to tell the screenSensitiveHelp script in this file to bypass itself and call up the chain.
	int apphelp___callCount

string function apphelp___getCustomHelpDoc()
; Gets the full path to the custom help document for the running scripts.
; Returns null if no such file is found.
var string docname = ""
var string appname = getActiveConfiguration()
if stringLength(apphelp___docApp)
&& stringCompare(appname, apphelp___docApp, True) == 0
	docname = apphelp___docName
endIf
if !docName
	var string fname = getActiveConfiguration() +".jcf"
	fname = findJAWSSettingsFile(fname)
	if stringLength(fname) && fileExists(fname)
		docname = iniReadString("Script Options", "HelpDocument", "", fname)
	endIf
endIf
if !docname
	docname = getActiveConfiguration()
	if !stringLength(docname) || docname == "default"
		return ""
	endIf
	docname = docname +".htm"
endIf
docname = findJAWSSettingsFile(docname)
if !fileExists(docname)
	return ""
endIf
return docname
endFunction

void function apphelp___doHelp()
; Implements the following behavior based on the number of JAWSKey+F1 presses and the presence or absence of a custom (.htm) help file:
;	One press: Normal ScreenSensitiveHelp in 0.6 seconds.
;	Two quick presses: The custom help document opens if present, or if not, JAWS app-specific (chm) help opens.
;	Three or more quick presses: JAWS-specific (chm) help opens if present.
var
	int callCount,
		string docname
callCount = apphelp___callCount
apphelp___callCount = 0
if callCount != 2
	; 1 = screenSensitiveHelp, 3+ = app-specific help.
	; The standard screenSensitiveHelp script will use isSameScript() to decide what to do.
	apphelp___callCount = -1
	performScript screenSensitiveHelp()
	apphelp___callCount = 0
	return
endIf

docname = apphelp___getCustomHelpDoc()
if docname
	if userBufferIsActive()
		userBufferDeactivate()
	endIf
;	sayMessage(OT_Message, apphelp___msgOpening)
	run("\"" +docname +"\"")
	return
endIf

; No custom help document, so revert to standard behavior.
apphelp___callCount = -1
performScript screenSensitiveHelp()
apphelp___callCount = 0
endFunction

script screenSensitiveHelp()
; Intercept that makes the apphelp___doHelp() function take effect.
if apphelp___callCount == -1
	; apphelp___doHelp sets apphelp___callCount to -1 to indicate it wants to call the original screenSensitiveHelp script.
	performScript screenSensitiveHelp()
	return
endIf

apphelp___callCount = apphelp___callCount +1
scheduleFunction("apphelp___doHelp", 6)
endScript

Script HotKeyHelp ()
If UserBufferIsActive () then
	UserBufferDeactivate ()
EndIf
SayFormattedMessage (OT_USER_BUFFER, msgHotKeyHelp, msgHotKeyHelp)
EndScript

void function apphelp__setDocName(string docname)
; Set the name of the custom help doc to load.
; Call from AutoStartEvent if necessary.
apphelp___docApp = getActiveConfiguration()
apphelp___docname = docname
endFunction

;============================== End module apphelp =============================


;================================ Module string ================================
; String manipulation functions.
;
; Author:  Doug Lee

string function string__segmentNext(string byRef s, string sep)
; Gets, returns, and removes from s the next string segment and the trailing separator if any.
; sep may be multi-character and is case sensitive.
; If sep is null, the entire string is treated as one segment.
var
	int pos,
	string s1
pos = 0
if sep
	pos = stringContains(s, sep)
endIf
if !pos || !sep
	s1 = stringChopLeft(" " +s, 1)  ; avoids pointer duplication
	s = ""
	return s1
endIf
s1 = stringLeft(s, pos-1)
s = stringChopLeft(s, pos +stringLength(sep) -1)
return s1
endFunction

int function string__segmentRemove(string byRef s, string sep, string r)
; Removes the first segment in s that exactly matches r, including case.
; Removes or leaves segment separators as appropriate.
; TODO: string__segmentRemove(s, "|", "val") is identical for s = "val" or "|val" or "val|":  the null string.
; Returns True if a segment is removed and False otherwise.
; Tested thoroughly for all segmentation and null cases 09/16/03.
var
	int slen, int seplen, int rlen, int pos
if !s || !sep || !r
	return False
endIf

; Cases of s: (1) no r, (2) r, (3) r+sep+right, (4) left+sep+r, (5) left+sep1+r+sep2+right.
slen = StringLength(s)
seplen = StringLength(sep)
rlen = stringLength(r)
pos = stringContains(s, sep+r+sep)
if pos
	; (5):  Most common case: r is in the middle of s somewhere, not at beginning or end.
	; This changes left+sep1+r+sep2+right to left+sep2+right.
	; pos currently points to the first character of sep1.
	s = stringLeft(s, pos-1) +stringChopLeft(s, pos+seplen+rlen-1)
	return True
elif stringContains(s, r) == 1
	if slen == rlen
		; (2) s is exactly r.
		s = ""
		return True
	elif slen >= rlen+seplen && substring(s, rlen+1, seplen) == sep
		; (3) s is r+sep+right (right may be null).
		s = stringChopLeft(s, rlen+seplen)
		return True
	endIf
elif stringContains(stringRight(s, seplen+rlen), sep+r)
	; (4):  s is left+sep+r (left may be null).
	s = stringChopRight(s, seplen+rlen)
	return True
endif

; (1):  s does not contain r.
return False
endFunction

int function string__segmentCount(string s, string sep)
; Returns the number of segments in the given string.
; If sep is null, the result is 0.
; As of 03/15/03, the null string is defined to have zero segments (previously one).
; Any non-null string containing no separators has one segment.
; WARNING: Unlike the built-in (as of JAWS 7.0) stringSegmentCount function,
; this function counts the null string after a trailing delimiter;
; thus, string__segmentCount("a|b|", "|") is 3, where JAWS would say 2.
; Also, this function considers a multi-character sep as a multi-character delimiter,
; whereas stringSegmentCount allows each character of sep to be a delimiter.
; This function requires JAWS 7.0 or later.
var
	int len,
	int seplen,
	int l,
	int n
seplen = stringLength(sep)
if seplen == 1
	n = stringSegmentCount(s, sep)
	if stringRight(s, 1) == sep
		; JAWS does not count the null string after a trailing delimiter, but I do.
		n = n +1
	endIf
	return n
endIf
len = stringLength(s)
if !len
	return 0
endIf
if !seplen
	return 0
endIf

; A little optimization made possible by JAWS 6.0's StringReplaceSubstrings function...
; Count segments by removing all separators and comparing string lengths.
l = stringLength(string__ref(stringReplaceSubstrings(s, sep, "")))
l = len -l
return (l / seplen) +1
endFunction

int function string__replace(string byRef s, string r1, string r2, string whatToDo)
; Replace r1 in s with r2 if r1 is found where whatToDo indicates.
; Note:  A space is sometimes prepended to and then removed from results in this function to keep JAWS from adding
; numeric strings arithmetically, which can happen in JAWS versions at least through 3.70.47 [DGL, 12/20/2000]
; whatToDo: l=left, r=right, a=anywhere, e=everywhere.
; Add t (for Trim) if whitespace should be removed at the point where s is split.
; Add c (for ConsiderCase) if case-sensitive matching is desired.
; For compatibility with older scripts, WhatToDo can also be an integer sum of flags:
; 0=l, 1=r, 2=a, 3=e, 8=t, 16=c.
; Returns the number of changes made to s, which must be passed by reference.
; If s is null, no changes are made and 0 is returned.
; If r1 is null, the following happens based on WhatToDo:
;	"e" - r2 is inserted between each character pair in s.  All blanks are removed first if "t" is included.
;	"l" or "a" - r2 is inserted after the first character of s, after blanks around it are removed if "t" is included.
;	"r" - r2 is inserted before the last character of s, after blanks around it are removed if "t" is included.
var
	string fromWhere,
	int trimEnds,
	int considerCase,
	int pos,
	int l1,
	string sa,
	string sb,
	int changed,
	int changeCount,
	int done

if !stringLength(s)
	return 0
endIf

; Compatibility with older integer whatToDo parameter.
if stringLength(whatToDo) == 1 && stringContains(whatToDo, "0")
	whatToDo = "l"
	elif stringToInt(whatToDo)
	whatToDo = stringSegment("r|a|e|||||lt|rt|at|et|||||lc|rc|ac|ec|||||ltc|rtc|atc|etc", "|", stringToInt(whatToDo) & 31)
endIf

whatToDo = stringLower(whatToDo)
trimEnds = stringContains(whatToDo, "t")
considerCase = stringContains(whatToDo, "c")
fromWhere = whatToDo

; Special cases for null r1.
if !stringLength(r1) && fromWhere == "e"
	if trimEnds
		sa = stringStripAllBlanks(s)
	else
		sa = string__ref(s)
	endIf
	s = ""
	l1 = stringLength(sa)
	pos = 1
	while pos < l1
		s = string__cat(s, string__cat(substring(sa, pos, 1), r2))
		pos = pos +1
	endWhile
	s = string__cat(s, stringRight(sa, 1))
	return stringLength(sa)
elif !stringLength(r1)
	if fromWhere == "l" || fromWhere == "a"
		if trimEnds
			sb = stringTrimLeadingBlanks(s)
		else
			sb = s +""
		endIf
		sa = stringLeft(sb, 1)
		sb = stringChopLeft(sb, 1)
		if trimEnds
			sb = stringTrimLeadingBlanks(sb)
		endIf
	else  ; fromWhere == "r"
		if trimEnds
			sb = stringTrimTrailingBlanks(s)
		else
			sb = s +""
		endIf
		sa = stringChopRight(sb, 1)
		if trimEnds
			sa = stringTrimTrailingBlanks(sa)
		endIf
		sb = stringRight(sb, 1)
	endIf
	if !stringLength(sa) || !stringLength(sb)
		return 0
	endIf
	s = string__cat(sa, string__cat(r2, sb))
	return 1
endIf

changeCount = 0
while !done
	l1 = stringLength(r1)
	changed = false
	if fromWhere == "l"
		if !stringCompare(stringLeft(s, l1), r1, considerCase)
			if trimEnds
				s = stringTrimLeadingBlanks(stringChopLeft(s, l1))
				s = r2 + s
			else
				s = r2 + stringChopLeft(s, l1)
			endIf
			changed = true
		endIf
	elif fromWhere == "r"
		if !stringCompare(stringRight(s, l1), r1, considerCase)
			if trimEnds
				s = stringTrimTrailingBlanks(stringChopRight(s, l1))
				s = s + r2
			else
				s = stringChopRight(s, l1) + r2
			endIf
			changed = true
		endIf
	else  ; fromWhere == "a" or "e"
		if considerCase
			pos = StringContains(s, r1)
		else
			pos = StringContains(stringLower(s), stringLower(r1))
		endIf
		if pos
			if trimEnds
				sa = stringTrimTrailingBlanks(stringLeft(s, pos-1))
				sb = stringTrimLeadingBlanks(stringChopLeft(s, pos+l1-1))
				s = " " +sa +r2 +sb
				s = stringChopLeft(s, 1)
			else
				s = " " +stringLeft(s, pos-1) +r2 +stringChopLeft(s, pos+l1-1)
				s = stringChopLeft(s, 1)
			endIf
			changed = True
		endIf
	endIf
	if fromWhere == "e"
		done = !changed
	else
		done = True
	endIf
	if changed
		changeCount = changeCount +1
	endIf
endWhile
return changeCount
endFunction

string function string__join(string s1, string sep, string s2)
; Returns the two given strings joined with the given separator.
; If s1 or s2 is empty, the separator is not included.
if stringIsBlank(s1)
	return s2
elif stringIsBlank(s2)
	return s1
else
	return s1 +sep +s2
endIf
endFunction

string function string__ref(variant v)
; Type cast from anything to string.
; Useful for allowing object results to be included in string concatenations without a compile-time error.
return v
endFunction

string function string__cat(string s1, string s2)
; Safe string concatenator.
return stringChopLeft(" " +s1 +s2, 1)
endFunction

;============================== End module string ==============================


;================================= Module util =================================
; Miscellaneous utility functions
;
; Author:  Doug Lee

int function util__compare(variant v1, variant v2, string how)
; Comparison that returns 1 if v1 is greater than v2, 0 if equal to v2, and -1 if less than v2.
; Comparison types (values of how):
;	n - Numeric comparison (non numbers sort as 0).
;	i - Case-insensitive string comparison.
;	s (default if how is null) - Case-sensitive string comparison.
;	<funcName> - Returns funcname(v1, v2). Uses CallFunctionByName() in JAWS 14 and later.
;	<funcCall> - Evaluates funcCall substituting %1 with "v1" and %2 with "v2."
; TODO:  Beeps and returns 0 in the following cases:
;	- String comparison (how == i or s) for JAWS versions older than 6.20.
;	- funcName when v1 or v2 contains a quote and JAWS is older than 14.0.
;	- funcCall) when v1 or v2 contains a quote (").
; This function requires JAWS 6.20 or later.
if !how
	how = "s"
endIf
if stringLength(how) == 1
	if how == "n"
		if stringToInt(v1) == stringToInt(v2)
			return 0
		elif stringToInt(v1) < stringToInt(v2)
			return -1
		else
			return 1
		endIf
	elif how == "s" || how == "i"
		return stringCompare(v1, v2, (how == "s"))
	endIf
endIf
var int isCall = (stringContains(how, "(") > 0)
var int isQuote = (stringContains(v1, "\"") || stringContains(v2, "\""))
var int jver = getJFWVersion()
if isQuote && (isCall || jver < 1400000)
	; Incompatible with formatStringWithEmbeddedFunctions().
	beep()
	return 0
endIf
if isCall
	string__replace(how, "%1", "\""+string__ref(v1)+"\"", "e")
	string__replace(how, "%2", "\""+string__ref(v2)+"\"", "e")
	return stringToInt(formatStringWithEmbeddedFunctions("<" +how +">"))
endIf
if jver < 1400000
	return stringToInt(formatStringWithEmbeddedFunctions("<" +how +"(\"" +string__ref(v1) +"\", \"" +string__ref(v2) +"\")>"))
endIf
return callFunctionByName(how, v1, v2)
endFunction

int function util__abs(int n)
; Return the absolute value of n.
if n < 0  return 0-n
else return n
endIf
endFunction

int function util__isSpecialFocus()
; Returns True if there is a special condition for which normal focus handling should not apply.
; Currently, the conditions that are considered special are menusActive, userBufferIsActive, inHJDialog, and switchingTasks (Alt+Tab is active).
return menusActive() ||userBufferIsActive() ||inHJDialog() || util__switchingTasks()
endFunction

int function util__switchingTasks()
; Returns True if the user is currently Alt+Tabbing between tasks.

; This approach is adapted from JAWS 9.0.2169's default.jss FocusChangedEventEx function.
if getWindowClass(getFocus()) == "#32771"
&& getObjectTypeCode()== WT_LISTBOXITEM
	return True
endIf
return False
endFunction

string function util__stringFromFile(string fname, /*?*/ string fmt)
; Returns the file contents as a string.
; fmt: null for system default, "a" for ASCII, "u" for Unicode.
var
 object fso, object f,
	int iFmt,
	string s
iFmt = -2  ; TristateUseDefault
if fmt == "a"
	iFmt = 0  ; TristateFalse
elif fmt == "u"
	iFmt = -1  ; TristateTrue
endIf
fso = createObjectEx("Scripting.FilesystemObject", False)
if !fso
fso = createObjectEx("Scripting.FilesystemObject", True)
endIf
if !fso
	;sayMessage(OT_Error, "Unable to get a fileSystemObject")
	return ""
endIf
f = fso.openTextFile(fname, 1, 0, iFmt)  ; 1 is ForReading, 0 means don't create
if !f
	;sayString("No code file")
	return ""
endIf
s = f.readAll()
f.close()
if !stringLength(s)
	;sayString("No code in the file")
endIf
return s
endFunction

int function util__stringToFile(string fname, string s, /*?*/ string fmt)
; Writes the string to the indicated file, creating or replacing the file.
; fmt: null for system default, "a" for ASCII, "u" for Unicode.
var
 object fso, object f,
	int iFmt
iFmt = -2  ; TristateUseDefault
if fmt == "a"
	iFmt = 0  ; TristateFalse
elif fmt == "u"
	iFmt = -1  ; TristateTrue
endIf
fso = createObjectEx("Scripting.FilesystemObject", False)
if !fso
fso = createObjectEx("Scripting.FilesystemObject", True)
endIf
if !fso
	;sayMessage(OT_Error, "Unable to get a fileSystemObject")
	return ""
endIf
f = fso.openTextFile(fname, 2, 1, iFmt)  ; 2 is ForWriting, 1 means create
if !f
	;sayString("Can't create file")
	return ""
endIf
f.write(s)
f.close()
return True
endFunction

handle function util__getDesktopWindow()
; Analogous to the Windows call of the same name.
var
	handle hwnd,
	string hlist, int hcount, int i

; Switching to the invisible cursor keeps GetWindowTypeCode from returning 3 (WT_Edit) for all windows when the virtual cursor is active.
saveCursor() invisibleCursor()

; First try handles known to be the Desktop window in various versions of Windows.
; 128=95/98, 65566=NT, 65556=XP, 65546=2k.
; Thanks to Victor Tsaran for helping collect these handles. [DGL, 04/28/03]
; 65552 for Windows 8.1.
hlist = "128|65566|65556|65552|65546|65536|65526|65516|65506"
hcount = 8
i = 1
while i <= hcount
	hwnd = stringToHandle(stringSegment(hlist, "|", i))
	if getWindowTypeCode(hwnd) == WT_Desktop
		return hwnd
	endIf
	i = i +1
endWhile

; The hard way...

; Start at the focused window, or if there isn't one, from an arbitrary window on the screen.
hwnd = getFocus()
if !hwnd
	hwnd = getWindowAtPoint(300, 300)  ; A valid point regardless of screen resolution
endIf
hwnd = getFirstWindow(getTopLevelWindow(hwnd))
while hwnd
	if getWindowTypeCode(getParent(hwnd)) == WT_Desktop
		return getParent(hwnd)
	endIf
	hwnd = getNextWindow(hwnd)
endWhile
return 0
endFunction

string function util__callStackInfo(int nlevels, string flags, /*?*/ int skip)
; Produce a representation of the current call stack formatted according to the given parameters.
; The following shortenings are applied to each script/function call in the return value:
;	- The shared folder path is replaced with "shared."
;	- The user folder path is removed.
;	- The word "function" is removed.
;	- The word "script" is replaced with "$."
; Requires JAWS 10 or later because the JAWS 10+ getScriptCallStack() is required.
; nlevels: Number of levels from starting position up the stack to print.  1 means just current function.  0 or less means everything.
; flags: f for forward order (default is reverse), s for short (one line and no parameters).
; skip: How many calls to skip over besides this function itself, default 0.
; Returns the requested call stack information or null if the JAWS version is too old.
var
	string userFolder, string sharedFolder,
	int short, int forward,
	int stacksize
if getJFWVersion() < 1000000
	return ""
endIf
flags = stringLower(flags)
short = stringContains(flags, "s")
forward = stringContains(flags, "f")
userFolder = getJAWSSettingsDirectory()
sharedFolder = getSharedSettingsDirectory()
var collection stack = util__callStack()
var int i = 2  ; skip this function's entry.
i = i +skip
stacksize = stack.count
if nlevels > 0
	nlevels = min(nlevels, stacksize-i+1)
else
	nlevels = stacksize-i+1
endIf
if nlevels <= 0
	return ""
endIf
var int delta = 1
if forward
	delta = -1
	i = i +nlevels -1
endIf
var string buf = ""
while nlevels
	if buf
		if forward
			buf = buf +" -> "
		else
			buf = buf +" <- "
		endIf
	endIf
	var collection  entry = stack[intToString(i)]
	var string fname = entry.file
	fname = stringReplaceSubstrings(fname, sharedFolder, "")
	fname = stringReplaceSubstrings(fname, userFolder, "user")
	if stringRight(fname, 4) == ".jsb"
		fname = stringChopRight(fname, 4)
	endIf
	buf = buf +fname +" "
	if entry.type == "script"
		buf = buf +"$"
	endIf
	if short
		buf = buf +stringSegment(entry.call, "(", 1)
	else
		buf = buf +string__ref(entry.call)
	endIf
	i = i +delta
	nlevels = nlevels -1
endWhile
return buf
endFunction

globals object util___csRE
collection function util__callStack()
; Return a collection of collections representing the script call stack.
; Requires JAWS 11 update 2 or later and the VBScript.RegExp object.
; Index the result by stringified 1-based integers.
; The most recent call is item 1.
; Structure of an item:
;	file: The file from which the script/function came, either "builtin" or a full path.
;	type: The type ("script" or "function", lower case).
;	call: The full call with parameters.
if !util___csRE
	util___csRE = util__createObject("VBScript.RegExp")
	if !util___csRE
		return
	endIf
	util___csRE.Global = -1
	util___csRE.IgnoreCase = 0
	util___csRE.Pattern = "\10([^\10\t]*)\t(script |function )"
endIf
var string sStack = "\10" +getScriptCallStack()
var object matches = util___csRE.execute(sStack)
var object match
var int n = matches.count
var int i = 1  ; Skip this function itself.
var string tmp, int pos1, int pos2
var collection stack stack = new collection
var collection entry
stack.count = 0
while i < n
	entry = new collection
	stack.count = stack.count +1
	stack[intToString(stack.count)] = entry
	match = matches(i)
	; Skip the \10 here.
	; Also switch from 0-base to 1-base.
	pos1 = match.firstIndex +2
	if i == n-1
		pos2 = stringLength(sStack) +1
	else
		pos2 = matches(i+1).firstIndex +1
	endIf
	tmp = substring(sStack, pos1, pos2-pos1)
	entry.file = string__segmentNext(tmp, "\t")
	entry.type = string__segmentNext(tmp, " ")
	entry.call = tmp
	i = i +1
endWhile
return stack
endFunction

string function util__getJAWSInfo()
; Returns a list of fields, separated by "|", as follows:
; 1.  The JAWS version number from the JFWUI2 window, in the form [M]M.mm.bbb; example, 5.00.584.
; 2.  The JAWS serial number from the JFWUI2 window.
; 3.  The JAWS serial number from GetJFWSerialNumber().
; The build number returned by this function is accurate, whereas it is "000" in GetJFWVersion in older JAWS versions.
var
	handle hwnd0, handle hwnd,
	int pos,
	string sn, string ver

; Find the JAWS dialog, which exists even when invisible or when JAWS is on the system tray.
hwnd0 = getFirstChild(findTopLevelWindow("JFWUI2", ""))
if !hwnd0
	return ""
endIf

; Get the serial number from the status bar's window name.
hwnd = getNextWindow(hwnd0)
sn = stringLower(util__getFullWindowName(hwnd))
pos = stringContains(sn, "ber")
if pos
	sn = intToString(stringToInt(substring(sn, pos+4, 10)))
else
	sn = ""
endIf

; Find the static containing the version number and get it.
hwnd = getFirstChild(hwnd0)
ver = ""
while hwnd && !ver
	ver = stringTrimTrailingBlanks(util__getFullWindowName(hwnd))
	pos = stringContainsFromRight(ver, " ")
	if pos
		ver = stringChopLeft(ver, pos)
	else
		ver = ""
	endIf
	hwnd = getNextWindow(hwnd)
endWhile

; The above version retrieval sometimes fails, so try the normal way.
if !ver
	ver = intToString(getJFWVersion())
	var int verlen = stringLength(ver)
	ver = formatString("%1.%2.%3",
		stringChopRight(ver, 5),
		substring(ver, verlen-4, 1),
		stringRight(ver, 4)
	)
endIf

if !sn && !ver
	return ""
else
	return ver +"|" +sn +"|" +getJFWSerialNumber()
endIf
endFunction

string function util__getFullWindowName(handle hWnd)
; Gets the full name of the given window if possible, using MSAA when it seems advisable.
; The JAWS GetWindowName() function seems to cut the name at around 255 characters.
; JAWS can get up to 4 K (maybe more sometimes) from MSAA though.
; Some JAWS versions (15?) can also return null for getWindowName when MSAA can get it.
var
	string wName,
	string mName,
	int cutoffLen

; Window names can reach 255 characters, but we choose an earlier cutoff for safety.
cutoffLen = 240

; Zero-length names can mean getWindowName missed it entirely. [DGL, 2014-09-29, JAWS 15.0]
wName = GetWindowName(HWnd)
var int len = stringLength(wName)
if !len || len > cutoffLen
	var int childID, object o
	o = GetObjectFromEvent(HWnd, -4, 0, childID)  ; -4 = ObjID_Client
	mName = o.accName(0)
	if stringLength(mName) > len
		return mName
	endIf
endIf
return wName
EndFunction

int function util__WMScroll(handle hwnd, string cmd)
; Scroll the given window or get info about its scroll position according to cmd.
; This function uses WinAPI messages to accomplish scrolling.
; Commands allowed (18):
;	LineLeft/Right/Up/Down, PageLeft/Right/Up/Down.
;	Left/Top/Right/Bottom.
;	H/VThumbPosition/Track.
;	H/VEndScroll.
var
	int msg, int wParam
; Translate cmd in place into something more directly parseable - a two-character code:
;	h/v:  Horizontal or vertical operation.
;	-+<>^$pte:
;		-+<>:  Line or page forward/back.
;		^$:  Fully left/top or right/bottom.
;		pte:  ThumbPosition, ThumbTrack, or EndScroll.
; This system avoids up to 18 string comparisons and guarantees that cmd is valid.
if !util__stringTran(cmd, False,
		" LineLeft LineRight LineUp LineDown PageLeft PageRight PageUp PageDown Left Top Right Bottom HThumbPosition VThumbPosition HThumbTrack VThumbTrack HEndScroll VEndScroll",
		" h- h+ v- v+ h< h> v< v> h^ v^ h$ v$ hp vp ht vt he ve"
)
	return False
endIf

; Now set msg and wParam.
if stringLeft(cmd, 1) == "h"
	msg = 0x0114  ; WM_HScroll
else
	msg = 0x0115  ; WM_VScroll
endIf
; This produces an SB_* constant from winuser.h.
wParam = stringContains("-+<>pt^$e", stringRight(cmd, 1)) -1
return sendMessage(hwnd, msg, wParam, 0)
endFunction

int function util__stringTran(string byRef val0, int caseSensitive, string s1, string s2)
; Translate val in place by finding it in s1 and replacing it with the corresponding s2 value.
; Returns True on success and False on failure.
; TODO: This function probably belongs in the String module.
var
	string val,
	string sep1, string sep2,
	int idx
val = val0
if !caseSensitive
	val = stringLower(val)
	s1 = stringLower(s1)
endIf
sep1 = stringLeft(s1, 1)
sep2 = stringLeft(s2, 1)
idx = stringSegmentIndex(s1, sep1, val, True)
if !idx || idx > stringSegmentCount(s2, sep2)
	return False
endIf
val0 = stringSegment(s2, sep2, idx)
return True
endFunction

variant function util__vcast(variant v)
; Cast any type to variant.
return v
endFunction

void function util__activateWindow(handle hwnd)
; Activate the given window's application.
; This function does nothing if focus is already on hwnd's application.
var handle hwndFocus = getTopLevelWindow(getFocus())
; That can be null in some cases, but WM_ACTIVATE allows for this.
var handle hwndDest = getTopLevelWindow(hwnd)
if hwndFocus == hwndDest
	return
endIf
; 6 is WM_Activate.
postMessage(hwndFocus, 6, 0, hwndDest)
postMessage(hwndDest, 6, 1, hwndFocus)
endFunction

object function util__createObject(string objname)
; A CreateObject wrapper that avoids causing 5-7 second delays in DOS boxes when called if JAWS is running as a service.
; As of June 11, 2011, uses the CoCreateInstance method of creating an object
; unless that doesn't work, whereupon the GetObjectHelper method is used.
; Previously the CoCreateInstance method was only used if JAWS was running as a service.
var object o = createObjectEx(objname, False)
if !o
	o = createObject(objname)
endIf
return o
endFunction

const
; The WM_SysCommand message.
	util__WM_SysCommand = 0x0112,
; System Menu Command Values straight out of WinUser.h (Nov 9, 2009).
	util__SC_Size = 0xF000,
	util__SC_Move = 0xF010,
	util__SC_Minimize = 0xF020,
	util__SC_Maximize = 0xF030,
	util__SC_NextWindow = 0xF040,
	util__SC_PrevWindow = 0xF050,
	util__SC_Close = 0xF060,
	util__SC_VScroll = 0xF070,
	util__SC_HScroll = 0xF080,
	util__SC_MouseMenu = 0xF090,
	util__SC_KeyMenu = 0xF100,
	util__SC_Arrange = 0xF110,
	util__SC_Restore = 0xF120,
	util__SC_TaskList = 0xF130,
	util__SC_ScreenSave = 0xF140,
	util__SC_Hotkey = 0xF150,
	util__SC_Default = 0xF160,
	util__SC_MonitorPower = 0xF170,
	util__SC_ContextHelp = 0xF180,
	util__SC_Separator = 0xF00F

int function util__sysCommand(handle hwnd, string which)
; Send a WM_SysCommand message to the given window. Values of which supported:
;	close, min[imize], max[imize], restore: Change among the given window states.
var string which0 = which
var int cmd = util__WM_SysCommand
var int wParam = 0
var int lParam = 0L
which = stringLower(stringLeft(which, 3))
if which == "clo"
	wParam = util__SC_Close
elif which == "min"
	wParam = util__SC_Minimize
elif which == "max"
	wParam = util__SC_Maximize
elif which == "res"
	wParam = util__SC_Restore
elif which == "hsc"
	wParam = util__SC_HScroll
elif which == "vsc"
	wParam = util__SC_VScroll
elif which == "mov"
	wParam = util__SC_Move
elif which == "siz"
	wParam = util__SC_Size
elif which == "nex"
	wParam = util__SC_NextWindow
elif which == "pre"
	wParam = util__SC_PrevWindow
elif which == "con"
	wParam = util__SC_ContextHelp
elif which == "def"
	wParam = util__SC_Default
; Special exceptions
elif which == "mon"
	; MonitorPower states.
	wParam = util__SC_MonitorPower
	which = stringLower(substring(which0, 4, 3))
	if which == "on"
		lParam = -1
	elif which == "low"
		lParam = 1
	elif which == "off"
		lParam = 2
	else
		return False
	endIf
else
	return False
endIf
return sendMessage(hwnd, cmd, wParam, lParam)
endFunction

;=============================== End module util ===============================


;=============================== Module winclick ===============================
; Support for sending mouse clicks of various sorts directly to a window.
; This permits clicking on obscured windows
; or clicking without checking for what is in the foreground.
;
; Author:  Doug Lee, with credit to Jim Bauer for the idea and some initial research.
;
; TODO:
;	- No support for mouse wheel events.
;	- Alt+clicks not supported.
;	- Non-client clicks not supported, such as for scrollbar elements.


int function winclick__send(handle hwnd, string which, /*?*/ int x, int y)
; Send a click to the given window using a sendMessage() call.
; hwnd: The window to which to send an event.
; which: Which event and keys to include.  Click specifiers:
;	Left, Right, or Middle:  Which button to use for a click.
;	dbl (optional):  Include if you want a double click.
;	Click (optional):  Not required.
;	Hover: Use in place of the above to cause a hover event instead of a click.
;	Leave: Use in place of the above to cause a leave event.
; Examples: leftDblClick, leftClick, leftDbl, left, hover.
; Key specifiers (include zero or more):
;	ctrl or control: A control key is down.
;	shift: A shift key is down.
; x: The x coordinate to click or hover over (optional).
; y: The y coordinate to click or hover over (optional).
; If x and y are not passed or are 0,
; the top left corner of the Window's client area is used.

; Returns the sendMessage() result.
return winclick___doClick(True, hwnd, which, x, y)
endFunction

void function winclick__post(handle hwnd, string which, /*?*/ int x, int y)
; Send a click to the given window using a sendMessage() call.
; hwnd: The window to which to send an event.
; which: Which event and keys to include.  Click specifiers:
;	Left, Right, or Middle:  Which button to use for a click.
;	dbl (optional):  Include if you want a double click.
;	Click (optional):  Not required.
;	Hover: Use in place of the above to cause a hover event instead of a click.
;	Leave: Use in place of the above to cause a leave event.
; Examples: leftDblClick, leftClick, leftDbl, left, hover.
; Key specifiers (include zero or more):
;	ctrl or control: A control key is down.
;	shift: A shift key is down.
; x: The x coordinate to click or hover over (optional).
; y: The y coordinate to click or hover over (optional).
; If x and y are not passed or are 0,
; the top left corner of the Window's client area is used.

winclick___doClick(False, hwnd, which, x, y)
endFunction

int function winclick___doClick(int useSend, handle hwnd, string which, /*?*/ int x, int y)
; Send a click to the given window using a sendMessage() or postMessage call.
; See winclick__sendClick() or winclick__postClick() for details on usage.
; Returns the sendMessage() result if sendMessage() is used, or nothing if postMessage() is used.
; Fine print: This function sends two messages, ; one for buttonDown and one for buttonUp.
; The sendMessage result comes from the first of these calls when sendMessage is used.
; Only one message is sent for a hover or leave.
which = stringLower(which)
var int msgDown, int msgUp
var int wParam = 0
if stringContains(which, "hover")
	msgDown = 0x02A1  ; WM_MOUSEHOVER
	msgUp = 0  ; skip that one.
elif stringContains(which, "leave")
	msgDown = 0x02A3  ; WM_MOUSELEAVE
	msgUp = 0  ; skip that one.
elif stringContains(which, "left")
	if stringContains(which, "dbl")
		msgDown = 0x0203  ; WM_LBUTTONDBLCLK
	else
		msgDown = 0x0201  ; WM_LBUTTONDOWN
	endIf
	msgUp = 0x0202  ; WM_LBUTTONUP
	wParam = wParam | 1  ; MK_LBUTTON
elif stringContains(which, "right")
	if stringContains(which, "dbl")
		msgDown = 0x0206  ; WM_RBUTTONDBLCLK
	else
		msgDown = 0x0204  ; WM_RBUTTONDOWN
	endIf
	msgUp = 0x0205  ; WM_RBUTTONUP
	wParam = wParam | 2  ; MK_RBUTTON
elif stringContains(which, "middle")
	if stringContains(which, "dbl")
		msgDown = 0x0209  ; WM_MBUTTONDBLCLK
	else
		msgDown = 0x0207  ; WM_MBUTTONDOWN
	endIf
	msgUp = 0x0208  ; WM_MBUTTONUP
	wParam = wParam | 0x10  ; MK_MBUTTON
else
	beep()
	return False
endIf
if stringContains(which, "shift")
	wParam = wParam | 4  ; MK_SHIFT
endIf
if stringContains(which, "ctrl") || stringContains(which, "control")
	wParam = wParam | 8  ; MK_CONTROL
endIf
var int lParam = 0
if x || y
	if !msgUp
		saveCursor()  JAWSCursor()  saveCursor()
		moveTo(x, y)
	endIf
	var int left, int right, int top, int bottom
	; -4 is ObjID_Client
	if MSAA__windowRect(hwnd, -4, left, right, top, bottom)
		x = x -left
		y = y -top
	else
		x = x -getWindowLeft(hwnd)
		y = y -getWindowTop(hwnd)
	endIf
	lParam = makeLong(x, y)
endIf
var int result
if useSend
	result = sendMessage(hwnd, msgDown, wParam, lParam)
	if msgUp
		sendMessage(hwnd, msgUp, 0, lParam)
	endIf
else
	postMessage(hwnd, msgDown, wParam, lParam)
	if msgUp
		postMessage(hwnd, msgUp, 0, lParam)
	endIf
endIf
if !msgUp
	pause()
endIf
return result
endFunction

;============================= End module winclick =============================


;================================= Module MSAA =================================
; Miscellaneous functions for MSAA access
;
; Author:  Doug Lee

string function MSAA__windowName(handle hwnd, /*?*/ int childID, int objID)
; Get the MSAA name for the given window.
return MSAA__windowProp(hwnd, "name", childID, objID)
endFunction

int function MSAA__windowState(handle hwnd, /*?*/ int childID, int objID)
; Get the MSAA state flags for the given window.
return MSAA__windowProp(hwnd, "state", childID, objID)
endFunction

collection function MSAA__windowRect(handle hwnd, int objID, optional int byRef left, int byRef right, int byRef top, int byRef bottom)
; Return by reference and as a collection the rectangle for the given object ID in the given window.
; Returns a collection on success and null on failure.
left = 0
right = 0
top = 0
bottom = 0
var object o, int childID
o = getObjectFromEvent(hwnd, objID, 0, childID)
if !o
	return
endIf
return MSAA__rect(o, 0, left, right, top, bottom)
endFunction

variant function MSAA__windowProp(handle hwnd, string whichProp, /*?*/ int childID, int objID)
; Returns any one of several possible MSAA properties for the given window.
; Supported properties:
;	Name, Role, State, Value.
;	Description, KeyboardShortcut, Help, ChildCount.
var
	variant result,
	object o
if whichProp == "childCount" || whichProp == "role" || whichProp == "state"
	result = util__vcast(0)
else
	result = util__vcast("")
endIf
if !hwnd
	return result
endIf
o = MSAA__objectFromWindow(hwnd, objID)
if !o
	return result
endIf
childID = childID +0
if whichProp == "name"
	return o.accName(childID)
elif whichProp == "state"
	return o.accState(childID)
elif whichProp == "role"
	return o.accRole(childID)
elif whichProp == "value"
	return o.accValue(childID)
elif whichProp == "description"
	return o.accDescription(childID)
elif whichProp == "help"
	return o.accHelp(childID)
elif whichProp == "keyboardShortcut"
	return o.accKeyboardShortcut(childID)
elif whichProp == "childCount"
	if childID
		return o.accChild(childID).accChildCount
	else
		return o.accChildCount
	endIf
endIf
return result
endFunction

collection function MSAA__rect(object o, int childID, optional int byRef left, int byRef right, int byRef top, int byRef bottom)
; Get the bounding rectangle for the given MSAA object and childID.
; The rectangle is returned 1-based, as JAWS likes its rectangles.
; MSAA provides 0-based left/top/width/height.
; The rectangle is returned by reference and also as a collection. Returns null on failure.
if !o
	left = 0
	right = 0
	top = 0
	bottom = 0
	return
endIf
if !childID
	childID = 0
endIf
o.accLocation(intRef(left), intRef(top), intRef(right), intRef(bottom), childID)
if !left && !right && !top && !bottom
	return
endIf
; but we want 1-based left/right/top/bottom.
left = left +1
top = top +1
right = left +right
bottom = top +bottom
var collection c c = new collection
c.left = left
c.right = right
c.top = top
c.bottom = bottom
return c
endFunction

globals
	handle MSAA___lastHwnd,
	int MSAA___lastObjID,
	object MSAA___lastObj,
	int MSAA___lastTC

object function MSAA__objectFromWindow(handle hwnd, /*?*/ int objID)
; Return the requested (default ObjID_Client) MSAA object from the given window.
; Caches results a bit for better performance.
if !stringLength(util__vcast(objID))
	objID = -4  ; ObjID_Client
endIf
var int tc = getTickCount()
if hwnd == MSAA___lastHwnd && objID == MSAA___lastObjID && MSAA___lastObj
&& MSAA___lastTC && tc - MSAA___lastTC < 500
	return MSAA___lastObj
endIf
var object o, int childID
o = getObjectFromEvent(hwnd, objID, 0, childID)
if !o
	return o
endIf
MSAA___lastHwnd = hwnd
MSAA___lastObjID = objID
MSAA___lastObj = o
MSAA___lastTC = tc
return o
endFunction

;=============================== End module MSAA ===============================


;================================ Module compat ================================
; Compatibility shim to make modern scripts run on older JAWS versions.
; When this module supports a function, it supports it back through JAWS 11.0 unless otherwise indicated.
; Functions begin with comment lines indicating the function's classification:
;	Override:  An override (Unknown() should be the only one).
;	Support:  Support code for other functions.
;	NativeStart:  A JAWS function replacement for JAWS 11.0 through the version just before the one indicated on the comment line.
;
; To add a new function catcher:
;	- Add it in the If block in Unknown().
;	- Write compat_<funcName> after the fashion of the others in this module.
;
; Author:  Doug Lee


globals
	int compat___JAWSVer

int function compat___getJFWVersion()
; Support
; Used internally to speed recognition of the current JAWS version by caching the getJFWVersion() result.
if !compat___JAWSVer
	compat___JAWSVer = getJFWVersion()
endIf
return compat___JAWSVer
endFunction

string function compat___StringRef(variant v)
; Support
; Type cast from anything to string.
; Useful for allowing object results to be included in string concatenations without a compile-time error.
return v
endFunction

variant function unknown(string theName, int isScript,
	variant byRef v1, variant byRef v2, variant byRef v3, variant byRef v4, variant byRef v5,
	variant byRef v6, variant byRef v7, variant byRef v8, variant byRef v9, variant byRef v10
)
; Override
; Hander-offer for functions that are missing in the running JAWS.
if isScript
	return unknown(theName, isScript, v1, v2, v3, v4, v5, v6, v7, v8, v9, v10)
endIf
var int jver = compat___getJFWVersion()
if jver >= 1400000 && v1
	; Indicates recursion; don't mess with those.
	; This prevents, for instance, a recursive stringStartsWith() call
	; from getting sent incorrect arguments below in JAWS 14+.
	return unknown(theName, isScript, v1, v2, v3, v4, v5, v6, v7, v8, v9, v10)
endIf
if stringCompare(theName, "stringStartsWith", False) == 0
	return compat___stringStartsWith(v1, v2)
elif stringCompare(theName, "collectionItemCount", False) == 0
	return compat___collectionItemCount(v1)
elif stringCompare(theName, "collectionItemExists", False) == 0
	return compat___collectionItemExists(v1, v2)
elif stringCompare(theName, "collectionRemoveItem", False) == 0
	return compat___collectionRemoveItem(v1, v2)
elif stringCompare(theName, "collectionRemoveAll", False) == 0
	return compat___collectionRemoveAll(v1)
elif stringCompare(theName, "collectionCopy", False) == 0
	return compat___collectionCopy(v1)
elif stringCompare(theName, "isScreenShadeOn", False) == 0
	; If the function isn't here, the feature isn't active either.
	; This one was introduced in one of the JAWS 2018 updates.
	return False
endIf
return unknown(theName, isScript, v1, v2, v3, v4, v5, v6, v7, v8, v9, v10)
endFunction

int function compat___stringStartsWith(string s1, string s2)
; NativeStart: 13.00
var int len1 = stringLength(s1)
var int len2 = stringLength(s2)
if len1 < len2
	return False
endIf
s1 = stringLeft(s1, len2)
return (stringCompare(s1, s2, False) == 0)
endFunction

int function compat___collectionItemCount(collection c)
; NativeStart: 14.00
var string k
var int cnt = 0
forEach k in c
	cnt = cnt +1
endForEach
return cnt
endFunction

int function compat___collectionItemExists(collection c, string k0)
; NativeStart: 12.00
var variant null
if c[k0]
	; Catches non-empty values.
	return True
endIf
; Now to catch empty values.
var string k
forEach k in c
	if stringCompare(k, k0, False) == 0
	&& c[k] != null
		return True
	endIf
endForEach
return False
endFunction

void function compat___collectionRemoveAll(collection c)
; NativeStart: 12.00
var variant null
var string k
forEach k in c
	c[k] = null
endForEach
endFunction

int function compat___collectionRemoveItem(collection c, string k)
; NativeStart: 12.00
var variant null
if !compat___collectionItemExists(c, k)
	return False
endIf
c[k] = null
return True
endFunction

int function compat___collectionCopy(collection c)
; NativeStart: 12.00
; ToDo: This is not a deep copy and so will not always work as expected.
var collection nc nc = new collection
var string k
forEach k in c
	nc[k] = c[k]
endForEach
return nc
endFunction

;============================== End module compat ==============================

